/*
 *  tstream.c -- typed stream I/O.
 *
 *  See TypedOutputStream.java for a description of the file format.
 *
 *  20-feb-2001/gmt  initial version
 *  11-may-2001/gmt  fix missing includes, potential portability problem
 *  11-aug-2001/gmt  reorder codes to group all int formats together
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tstream.h"



/*
 *  peekValue(file) -- return type of next value in file.
 *
 *  returns one of:
 *	'n'   null value is next  (can be read as a string or integer)
 *	'i'   integer value is next
 *	'f'   floating value is next
 *	's'   string value is next
 *	'?'   unrecognized value is next
 *	'e'   EOF is next
 */
char peekValue(FILE *f) {
    int c = getc(f);			/* peek at next byte */
    if (c < 0) {
	return 'e';
    }
    ungetc(c, f);			/* push back byte if not EOF */
    if (c < 0xE0) {
	return 'i';
    } else if (c < 0xF0) {
	return 's';
    } else if (c < 0xFB) {
	return 'f';
    } else if (c == 0xFE) {
	return 'n';
    } else {
	return '?';
    }
}



/*
 *  readInteger(file, &value) -- read integral value and store in *v.
 *
 *  returns 0 for success, -1 for EOF or failure.
 */
int readInteger(FILE *f, long *value) {
    int b, n;
    long v, diff;
    static long prevInt = 0;

    b = getc(f);			/* read header byte */
    if (b == EOF)
	return -1;

    if (b < 0x40) {
	*value = prevInt = b;		/* 6-bit positive value */
	return 0;
    }
    if (b < 0x80) {
	*value = prevInt = b - 0x80;	/* 6-bit negative value */
	return 0;
    }
    if (b < 0xC0) {
	diff = -(b & 0x20) | (b & 0x1F);
	*value = prevInt += diff;	/* 6-bit signed difference */
	return 0;
    }

    if (b == 0xFE) {			/* null */
	*value = 0;
	return 0;
    }

    if ((b & 0xF8) == 0xC0) {		/* 11-bit signed value */
	v = ((- (b & 4)) | (b & 3)) << 8;
	b = getc(f);
	if (b == EOF)
	    return -1;
	*value = prevInt = v | b;
	return 0;
    }

    if ((b & 0xF8) == 0xC8) {		/* 11-bit signed difference */
	v = ((- (b & 4)) | (b & 3)) << 8;
	b = getc(f);
	if (b == EOF)
	    return -1;
	*value = prevInt += v | b;
	return 0;
    }

    if ((b & 0xF0) != 0xD0)
	return -1;			/* if not an integer */
    if ((b & 8) == 0)
	prevInt = 0;			/* if value follows, not difference */

    n = b & 7;				/* get bytecount-1 */
    b = getc(f);
    if (b == EOF)
	return -1;
    v = (signed char) b;		/* sign extend first byte */

    while (n-- > 0) {			/* while more bits */
	b = getc(f);
	if (b == EOF)
	    return -1;
	v = (v << 8) | b;		/* squeeze them in */
    }
    *value = prevInt += v;		/* combine with previous value */
    return 0;
}



/*
 * writeInteger(file, value) -- pack and write integer value.
 */
void writeInteger(FILE *f, long v) {
    int bbb, h;
    long diff, x;
    static long prevInt = 0;

    diff = v - prevInt;			/* difference from previous value */
    prevInt = v;			/* save curr value as new "previous" */

    if (-0x40 <= v && v < 0x40) {
	putc(v & 0x7F, f);
    } else if (-0x20 <= diff && diff < 0x20) {
	putc(0x80 | (diff & 0x3F), f);
    } else if (-0x400 <= v && v < 0x400) {
	putc(0xC0 + ((v >> 8) & 0x07), f);
	putc((int) v, f);
    } else if (-0x400 <= diff && diff < 0x400) {
	putc(0xC8 + ((diff >> 8) & 0x07), f);
	putc(diff, f);
    } else {
	if (labs(v) - 1 <= labs(diff) - 1) {	/* -1 handles 0x80...00 */
	    h = 0xD0;				/* encode value itself */
	} else {
	    h = 0xD8;				/* encode diff from previous */
	    v = diff;
	}
	bbb = 1;
	for (x = v >> 15; x != 0 && x != -1; x >>= 8) {
	    bbb++;
	}
	putc(h + bbb, f);
	while (bbb >= 0) {
	    putc(v >> (8 * bbb--), f);
	}
    }
}



/*
 *  readString(file, buf, len) -- read string into buffer.
 *
 *  returns actual length for success, -1 for EOF or failure.
 */
int readString(FILE *f, char *buf, size_t len) {
    size_t i;
    long n, tsp;

    int h = getc(f);			/* read header byte */
    if (h == EOF)
	return -1;
    else if (h == 0xFE)			/* treat null as empty string */
	h = 0XE0;
    else if ((h & 0xF0) != 0xE0)
	return -1;			/* invalid header byte */

    n = h & 0x0F;			/* extract length */
    tsp = 0;

    if (n == 15) {
	if (readInteger(f, &n) < 0)	/* read first length value */
	    return -1;
	if (n < 0) {			/* if negative, counts trailg spaces */
	    tsp = -n;
	    if (readInteger(f, &n) < 0 || n < 0)      /* read trimmed strlen */
		return -1;
	}
    }

    for (i = 0; i < n; i++) {		/* read characters into buffer */
	int c = getc(f);
	if (c == EOF)
	    return -1;
	if (i < len)
	    buf[i] = c;
    }

    n += tsp;				/* set total len including trailers */

    while (i < len && tsp-- > 0)	/* add trailing spaces that fit */
	buf[i++] = ' ';

    if (n < len)
	buf[n] = '\0';			/* terminate string */
    else
	buf[len - 1] = '\0';

    return n;				/* return total length */
}



/*
 *  writeString(file, s) -- write packed string.
 */

void writeString(FILE *f, char *s) {

    size_t len = strlen(s);		/* length of string */
    size_t tsp = 0;			/* trailing spaces */

    while (len > 0 && s[len - 1] == ' ') {
	tsp++;
	len--;
    }

    if (len < 15 && tsp < 3) {
	len += tsp;
	putc(0xE0 + len, f);
	while (len-- > 0)
	    putc(*s++, f);
    } else {
	putc(0xEF, f);
	if (tsp > 0)
	    writeInteger(f, -(int)tsp);
	writeInteger(f, len);
	while (len-- > 0)
	    putc(*s++, f);
    }
}
