/*
 * client.c --
 *
 *	This file implements the user interface functions of the
 *	token client.
 *
 *	The environment variable TSVC_TRACE may be set to trace
 *	incoming messages on stderr.
 */

static char rcsid[] =
"$Id: client.c,v 1.8 2001/08/31 20:51:27 gmt Exp $";

#include "tokInt.h"

/*
 * Private (per-service) data structure:
 */
typedef struct TokClientData {
    int ready;			/* have we completed initialization? */
    int active;			/* are we still active? */
    int tracing;		/* are we tracing? */
    long nextmsg;		/* next message sequence number to assign */
    TokMutex mnMutex;		/* mutex for that */
} *ClientData;

/*
 * Prototypes for procedures referenced only in this file:
 */
static void *Heartbeat(void *arg);
static void *Listener(void *arg);

static void RcvConfig(Tok_Service svc, TokPickle p, int from, int to);
static void RcvGrant(Tok_Service svc, TokPickle p);
static void RcvRevoke(Tok_Service svc, int from, char *tkname);
static void RcvConfirm(Tok_Service svc, long msgnum);
static void *DoCallback(void *arg);

static void ClearToken(Tok_Token t);
static void NewMsgNum(Tok_Token t);

/*
 * Tok_Config -- read token service configuration from file.
 *
 * Returns a null-terminated allocated list of pointers to allocated strings.
 * Returns null in case of errors, including nonexistent file or empty list.
 */

char **
Tok_Config(char *cfgfile) {

    char lbuf[200], tbuf[100];
    char **list = NULL;
    int n = 0;
    FILE *f;

    if ((f = fopen(cfgfile, "r")) == NULL)
	return NULL;

    while (fgets(lbuf, sizeof(lbuf), f) != NULL) {
	if (sscanf(lbuf, " server %99s ", tbuf) == 1) {
	    n++;
	    list = TokRealloc(list, (n + 1) * sizeof(char *));
	    list[n-1] = TokSalloc(tbuf);
	}
    }

    fclose(f);
    list[n] = NULL;	/* terminate list */
    return list;
}

/*
 * Tok_Open(servers) -- initialize client using specified configuration file
 *
 * Initializes one client by contacting a token service running on
 * the given set of servers.
 * 
 * Tok_Open does not return until an adequate set of servers is
 * available.  If the servers are down, Tok_Open periodically rechecks
 * them until they are again functioning.
 * 
 * A token service handle, for use with subsequent calls, is returned.
 */

Tok_Service
Tok_Open(char *servers[]) {
    Tok_Service svc;
    ClientData cd;
    int leader = 0;

    svc = TokAlloc(sizeof(struct TokService));
    cd = svc->clientData = TokAlloc(sizeof(struct TokClientData));
    cd->ready = FALSE;
    cd->active = TRUE;
    cd->tracing = (getenv("TSVC_TRACE") != NULL);
    cd->nextmsg = 1;
    TokInitMutex(&cd->mnMutex);
    svc->net = TokSetConfig(svc, servers);
    TokSpawn(Listener, svc);

    TokInitReissue(svc);
    TokInitHoldings(svc);

    /* initial handshake */
    for (;;) {				/* repeat until successful */
	TokSendVals(svc, TSVC_LOGIN, leader, "s", TokHostPort(svc->net));
	sleep(LOGIN_INTERVAL);
	if (cd->ready)
	    break;
	++leader;
	if (!servers[leader])
	    leader = 0;
    }

    TokSpawn(Heartbeat, svc);

    return svc;
}

/*
 * Tok_Close(svc) -- shut down client.
 * 
 * Disconnects a client from the token service.  This is the inverse
 * of Tok_Open.
 */
void
Tok_Close(Tok_Service svc) {
    ClientData cd = svc->clientData;
    cd->active = FALSE;
    TokTermReissue(svc);
    TokTermHoldings(svc);
    TokSendVals(svc, TSVC_LOGOUT, -1, "");
    TokTermNetw(svc->net);
    TokFree(svc->clientData);
    svc->clientData = NULL;
}

/*
 * Tok_Request(svc, name, how, callback, arg) -- acquire a token.
 * 
 * Acquires access to a token, blocking until the token is available.
 * "how" is TOK_SHARED or TOK_EXCLUSIVE to select the access mode.
 * An opaque token pointer is returned.
 * 
 * If a conflicting request is received while the token is held,
 *	callback(Tok_Token t, ClientData arg)
 * is invoked to notify the client of this situation.  The client
 * should generally release the token as soon as possible.
 * 
 * The callback is invoked at most once and is run in a separate
 * thread.  If a null callback address is supplied, the client cannot
 * be notified.  This is reasonable if the client intends to hold the
 * token only momentarily.
 * 
 * If a client requests a token it already holds, this is a fatal error.
 */
Tok_Token
Tok_Request(Tok_Service svc, char *name, int how,
	void (*callback)(Tok_Token, void *), void *arg) {
    Tok_Token t;

    t = TokLookup(svc, name, TRUE);
    TokLockMutex(&t->mutex);

    if (t->state != TOK_FREE) {
	TokUnlockMutex(&t->mutex);
	TokAbort("requesting token already held");
    }
    ClearToken(t);
    NewMsgNum(t);

    TokSendVals(svc, TSVC_REQUEST, TokServer(svc, name), "ti",
	t, (how == TOK_SHARED) ? FLAG_SHARED : FLAG_EXCLUSIVE);
    t->state = TOK_REQUESTED;
    t->access = how;
    t->callback = callback;
    t->clientData = arg;

    while (t->state != TOK_RECEIVED) {
	TokWaitCond(&t->cv, &t->mutex);
    }
    t->state = TOK_HELD;
    TokUnlockMutex(&t->mutex);
    return t;
}

/*
 * Tok_SetData(t, data, length) -- redefine token data.
 * 
 * Replaces the local copy of the data associated with a currently
 * held token.  The data is copied, so the caller need not preserve
 * it beyond the call.  Any previous value returned by Tok_GetData()
 * is rendered invalid by this call.  The new data is not forwarded
 * to the server.
 */
void
Tok_SetData(Tok_Token t, void *data, int length) {
    TokLockMutex(&t->mutex);
    ClearToken(t);
    if (data != NULL && length > 0) {
	memcpy(t->data = TokAlloc(length), data, length);
	t->datalen = length;
    }
    t->dirty = TRUE;
    TokUnlockMutex(&t->mutex);
}

/*
 *  ClearToken(t) -- clear token data.  Called while holding the mutex.
 */
void ClearToken(Tok_Token t) {
    TokFree(t->data);
    t->data = NULL;
    t->datalen = 0;
    t->dirty = FALSE;
}

/*
 * Tok_Update(t) -- send updated data to token server.
 *
 * Sends updated data to the token server without releasing the token.
 * Requests granted subsequently for this token will see the new data,
 * but the new data is not forwarded to any other current holders.
 *
 * Tok_Update has no effect:
 * --   if Tok_SetData has not been called since the token was granted
 * --   if Tok_Update has already been called since the last Tok_SetData
 * --   if the server has requested revocation of this token
 *	(in which case the expected Tok_Release call will update it)
 */
void
Tok_Update(Tok_Token t) {

    if (! t->dirty) {
	return;
    }

    TokLockMutex(&t->mutex);

    switch (t->state) {

	case TOK_REVISED:
	    while (t->state == TOK_REVISED) {
		TokWaitCond(&t->cv, &t->mutex);
	    }
	    if (t->state != TOK_HELD) {	/* if revoked or released, bail out */
		TokUnlockMutex(&t->mutex);
		return;
	    }
	    t->dirty = TRUE;	/* may have been reset by CONFIRM */
	    break;

	case TOK_HELD:
	    break;

	case TOK_LEAVING:
	    /* nothing to do: will be updated when revoke is answered */
	    TokUnlockMutex(&t->mutex);
	    return;

	default:
	    TokUnlockMutex(&t->mutex);
	    TokAbort("attempt to update released token");
    }

    NewMsgNum(t);
    TokSendVals(t->service, TSVC_RETURN, TokServer(t->service, t->name),
	"ti", t, FLAG_UPDATE);

    t->state = TOK_REVISED;
    TokUnlockMutex(&t->mutex);
}

/*
 * Tok_Release(t) -- return a token.
 * 
 * Returns a token to the token service and returns changed data
 * to the server.  This is the inverse of Tok_Request.
 * 
 * After a token has been released, the pointer t becomes invalid,
 * and no further actions are allowed on the token.  Additionally,
 * pointers returned by Tok_GetName and Tok_GetData for this token
 * are also rendered invalid.
 */
void
Tok_Release(Tok_Token t) {
    int f;

    TokDelete(t);				/* remove from holdings */
    TokLockMutex(&t->mutex);

    if (t->dirty) {
	f = FLAG_DEASSIGN + FLAG_UPDATE;
    } else {
	f = FLAG_DEASSIGN;
	ClearToken(t);
    }

    switch (t->state) {
	case TOK_HELD:
	case TOK_REVISED:
	case TOK_LEAVING:
	    NewMsgNum(t);
	    TokSendVals(t->service, TSVC_RETURN, TokServer(t->service, t->name),
		"ti", t, f);
	    t->state = TOK_RELEASED;
	    break;
	default:
	    TokUnlockMutex(&t->mutex);
	    TokAbort("attempt to release token not held");
    }
    TokSignalCond(&t->cv);
    TokUnlockMutex(&t->mutex);
}

/*
 * Tok_GetData(t) -- get token data.
 * 
 * Returns a pointer to the data inside a token.  The pointer is
 * guaranteed to be properly aligned for use with any datatype.
 * The return value is valid only until the next Tok_SetData or
 * Tok_Release call for this token.
 */
void *
Tok_GetData(Tok_Token t) {
    if (t->data == NULL)
	return "";
    else
	return t->data;
}

/*
 * Tok_GetLength(t) -- get length of token data.
 * 
 * Returns the length of the data held in a token and returned
 * by Tok_GetData.
 */
int
Tok_GetLength(Tok_Token t) {
    return t->datalen;
}

/*
 * Tok_GetName(t) -- get name of held token.
 * 
 * Returns the name of a token.  The returned pointer remains valid
 * only until the token is released by a call to Tok_Release.
 */
char *
Tok_GetName(Tok_Token t) {
    return t->name;
}

/*
 * Tok_GetAccess(t) -- get access mode of held token.
 * 
 * Returns TOK_SHARED or TOK_EXCLUSIVE indicating the access mode
 * of a token.
 */
int
Tok_GetAccess(Tok_Token t) {
    return t->access;
}

/*
 * Tok_GetCallback(t) -- get callback associated with a token.
 * 
 * Returns the callback function, which may be null, associated with 
 * a token.
 */
void (*Tok_GetCallback(Tok_Token t))(Tok_Token, void *) {
    return t->callback;
}

/*
 * Tok_GetArgument(t) -- get argument of token callback.
 * 
 * Returns the client data associated with the callback function of
 * a token.
 */
void *
Tok_GetArgument(Tok_Token t) {
    return t->clientData;
}

/*  Heartbeat(svc) -- sends ALIVE message periodically  */

static void *
Heartbeat(void *arg) {
    Tok_Service svc = arg;
    ClientData cd = svc->clientData;
    while (cd->active) {
	TokSendVals(svc, TSVC_ALIVE, -1, "");
	sleep(HBT_INTERVAL);
    }
    return NULL;
}

/*  Listener(svc) -- handles incoming messages  */

static void *
Listener(void *arg) {
    Tok_Service svc = arg;
    ClientData cd = svc->clientData;
    TokPickle p;
    int type, from, to;

    while (cd->active) {
	p = TokRecvPickle(svc);
	if (p == NULL) {
	    if (cd->active)
		TokAbort("Listener lost port");
	    else
		break;
	}
	type = TokUnpickleInt(p);
	from = TokUnpickleInt(p);
	to = TokUnpickleInt(p);
	/*sig*/ TokUnpickleInt(p);
	switch (type) {
	    case TSVC_CONFIG:	RcvConfig(svc, p, from, to);		break;
	    case TSVC_GRANT:	RcvGrant(svc, p);			break;
	    case TSVC_REVOKE:	RcvRevoke(svc, from,
					TokUnpickleString(p));		break;
	    case TSVC_CONFIRM:	RcvConfirm(svc, TokUnpickleInt(p));	break;
	    default:		TokAbort("unexpected message");
	}
	TokFreePickle(p);
    }
    return NULL;
}

/*  RcvConfig(svc, p, from, to) -- handle config message, respond w/ catalog  */

static void
RcvConfig(Tok_Service svc, TokPickle p, int from, int to) {

    ClientData cd = svc->clientData;

    if (cd->tracing) {
	char buf[100];
	int leader = TokUnpickleInt(p);
	int n = TokUnpickleBytes(p, buf, sizeof(buf));
	int i;
	fprintf(stderr, "CONFIG(%d,%d)(%d,", from, to, leader);
	for (i = 0; i < n; i++) 
	    fprintf(stderr, "%c", "-=+"[(int)buf[i]]);
	fprintf(stderr, ")\n");
    }

    if (to != 0)
	cd->ready = TRUE;
    TokSendCatalog(svc, from);
}

/*  RcvGrant(svc, p) -- handle GRANT message  */

static void
RcvGrant(Tok_Service svc, TokPickle p) {

    ClientData cd = svc->clientData;
    Tok_Token t;
    long msgnum = TokUnpickleInt(p);
    char *tkname = TokUnpickleString(p);
    int nbytes = TokUnpickleInt(p);

    if (cd->tracing)
	fprintf(stderr, "GRANT(%ld, %s, %dB)\n", msgnum, tkname, nbytes);

    TokForget(svc, msgnum);			/* stop reissuing request */
    t = TokLookup(svc, tkname, FALSE);
    TokFree(tkname);

    if (t != NULL) {
	TokLockMutex(&t->mutex);
	t->datalen = nbytes;
	if (t->state == TOK_REQUESTED && t->msgnum == msgnum) {
	    if (t->datalen > 0) {
		t->data = TokAlloc(nbytes);
		TokExtractBytes(p, t->data, nbytes);
	    } else {
		t->data = NULL;
	    }
	    t->dirty = FALSE;
	    t->state = TOK_RECEIVED;
	}
	TokSignalCond(&t->cv);
	TokUnlockMutex(&t->mutex);
    }
}

/*  RcvRevoke(svc, from, name) -- handle REVOKE message by notifying callback */

static void
RcvRevoke(Tok_Service svc, int from, char *tkname) {
    Tok_Token t;

    if (svc->clientData->tracing)
	fprintf(stderr, "REVOKE(%s)\n", tkname);

    t = TokLookup(svc, tkname, FALSE);
    if (t != NULL) {
	/*
	 *  If token still held, invoke user callback, but only once.
	 *  (On subsequent attempts, t->state will be wrong.)
	 */
	TokLockMutex(&t->mutex);
	if (t->state == TOK_FREE || t->state == TOK_REQUESTED) {
	    TokSendVals(svc, TSVC_RETURN, from, "lssi",
		0, tkname, "", FLAG_DEASSIGN);
	    TokUnlockMutex(&t->mutex);
	    return;
	}
	if (t->state != TOK_HELD && t->state != TOK_REVISED) {
	    TokUnlockMutex(&t->mutex);
	    return;				/* nothing to do */
	}
	t->state = TOK_LEAVING;
	TokSignalCond(&t->cv);
	TokUnlockMutex(&t->mutex);		/* first, release lock */
	if (t->callback != NULL) {
	    TokSpawn(DoCallback, t);		/* then, call callback func */
	}
    } else {
	/*
	 * We don't have this token; respond by deassigning.
	 */
	TokSendVals(svc, TSVC_RETURN, from, "lssi",
	    0, tkname, "", FLAG_DEASSIGN);
    }
    TokFree(tkname);
}

/*  DoCallback(token) -- invoke user callback.  Run in its own thread.  */

static void *
DoCallback(void *arg) {
    Tok_Token t = arg;
    t->callback(t, t->clientData);
    return NULL;
}

/*  RcvConfirm(svc, n) -- handle CONFIRM message acknowledging RETURN msg n  */

static void
RcvConfirm(Tok_Service svc, long msgnum) {
    Tok_Token t;

    if (svc->clientData->tracing)
	fprintf(stderr, "CONFIRM(%ld)\n", msgnum);
    
    t = TokForget(svc, msgnum);			/* stop reissuing request */
    if (t == NULL || t->msgnum != msgnum) {
	return;					/* if other action since */
    }

    TokLockMutex(&t->mutex);
    switch (t->state) {
	case TOK_REVISED:
	    t->dirty = FALSE;
	    t->state = TOK_HELD;
	    TokSignalCond(&t->cv);
	    TokUnlockMutex(&t->mutex);
	    break;
	case TOK_LEAVING:
	    /* no change in status */
	    /* release remains pending */
	    TokUnlockMutex(&t->mutex);
	    break;
	case TOK_RELEASED:	
	    /* token has already been unlinked; now free it */
	    TokUnlockMutex(&t->mutex);
	    TokDestroyMutex(&t->mutex);
	    TokDestroyCond(&t->cv);
	    TokFree(t->name);
	    TokFree(t->data);
	    TokFree(t);
	    break;
	default:
	    TokUnlockMutex(&t->mutex);
	    TokAbort("unexpected state in RcvConfirm");
    }
}

/*
 *  NewMsgNum(t) -- assign new message number to t->msgnum;
 */
static void NewMsgNum(Tok_Token t) {
    ClientData cd = t->service->clientData;
    TokLockMutex(&cd->mnMutex);
    t->msgnum = cd->nextmsg++;
    TokUnlockMutex(&cd->mnMutex);
}
