/*
 * config.c --
 *
 *	This file implements functions related to the server configuration.
 */

static char rcsid[] =
"$Id: config.c,v 1.6 2001/09/25 17:40:15 gmt Exp $";

#include "tokInt.h"
#include <stdarg.h>

/*
 * Private (per-service) data structure:
 */
typedef struct TokConfigData {
    int nservers;	/* number of servers */
    int signature;	/* host list signature */
    int leader;		/* leading server */
    int myid;		/* ID of this client */
    char **hosts;	/* array of server host names */
    int *ports;		/* array of server port numbers */
    char *states;	/* array of server states */
} *ConfigData;

/*
 * Configuration parameters and manifest constants:
 */

#define MAX_SERVERS 100

#define SERVER_DOWN 0
#define SERVER_BOOTING 1
#define SERVER_READY 2

/*
 * TokSetConfig(svc, servers) -- initialize for subsequent operations.
 *
 * Returns network handle.
 */

TokNetwork
TokSetConfig(Tok_Service svc, char *servers[]) {
    ConfigData cd;
    char *s, *t, **p;
    int i, n;

    cd = svc->configData = TokZalloc(sizeof(struct TokConfigData));
    cd->nservers = 0;
    for (p = servers; *p != NULL; p++)
	cd->nservers++;
    cd->hosts = TokAlloc(cd->nservers * sizeof(char *));
    cd->ports = TokAlloc(cd->nservers * sizeof(int));
    cd->states = TokZalloc(cd->nservers);

    if (cd->nservers > MAX_SERVERS)
	TokAbort("too many servers -- increase MAX_SERVERS");

    for (i = 0; i < cd->nservers; i++) {
	s = servers[i];
	t = strchr(s, ':');
	if (t == NULL) {
	    cd->hosts[i] = TokSalloc(s);
	    cd->ports[i] = 0;
	} else {
	    n = t - s;
	    cd->hosts[i] = strncpy(TokZalloc(n + 1), s, n);
	    cd->ports[i] = atoi(t + 1);
	}
    }

    cd->signature = TokHashList(servers) & 0x1FFF;

    return TokInitNetw(cd->hosts, cd->ports, cd->nservers);
}

/*  TokSetStates(svc, a) -- update server state array  */

void
TokSetStates(Tok_Service svc, char *a) {
    ConfigData cd = svc->configData;
    memcpy(cd->states, a, cd->nservers);
}

/*  TokServer(svc, tkname) -- return index of current server for given token  */

int
TokServer(Tok_Service svc, char *tkname) {
    ConfigData cd = svc->configData;
    int i, j, t, tmp[MAX_SERVERS];
    int h = TokHash(tkname);
    int n = h % cd->nservers;

    if (cd->states[n] != SERVER_DOWN) {
	return n;
    }

    for (j = 1; j < cd->nservers; j++)
	tmp[j] = j;
    tmp[n] = 0;

    for (i = 1; i < cd->nservers - 1; i++) {
	h = TokRehash(h);
	j = i + h % (cd->nservers - i);
	t = tmp[j];
	if (cd->states[t] != SERVER_DOWN)
	    return t;
	tmp[j] = tmp[i];
    }

    return -1;
}

/*
 *  TokSendVals(svc, typecode, dest, fmt, args) -- send values as a message
 *
 *  typecode: the message classification code
 *  destination: server index, or -1 to send to leader
 *  fmt: characters indicating types of subsequent arguments:
 *	i  integer value		passed as int
 *	l  long integer value		long
 *	s  string value			char *
 *	b  byte array			void *, int length
 *	t  token msgnum+name+data	Tok_Token
 *	p  raw pickle contents		TokPickle
 *
 *  A 't' argument (only one is allowed) records the t->msgnum field
 *  and causes the message to be re-sent until it is acknowledged.
 */

void
TokSendVals(Tok_Service svc, int tc, int dest, char *fmt, ...) {
    va_list ap;
    TokPickle p, q;
    Tok_Token t = NULL;
    char c;
    ConfigData cd = svc->configData;

    if (dest < 0)
	dest = cd->leader;

    p = TokEmptyPickle();
    TokPickleInt(p, tc);
    TokPickleInt(p, cd->myid);
    TokPickleInt(p, dest);
    TokPickleInt(p, cd->signature);

    va_start(ap, fmt);
    while ((c = *fmt++) != '\0')
	switch (c)  {
	    default:	TokAbort("unrecognized msend format char");
	    case 'i':	TokPickleInt(p, va_arg(ap, int));  break;
	    case 'l':	TokPickleInt(p, va_arg(ap, long));  break;
	    case 's':	TokPickleString(p, va_arg(ap, char *));  break;
	    case 'b': {
	        void *buf = va_arg(ap, void *);
		int len = va_arg(ap, int);
		TokPickleBytes(p, buf, len);
		break;
	    }
	    case 't': {
		t = va_arg(ap, Tok_Token);
		TokPickleInt(p, t->msgnum);
		TokPickleString(p, t->name);
		TokPickleBytes(p, t->data, t->datalen);
		break;
	    }
	    case 'p': {
		q = va_arg(ap, TokPickle);
		TokAppendBytes(p, TokPickleData(q), TokPickleSize(q));
		break;
	    }
	}
    va_end(ap);

    TokSend(svc->net, dest, TokPickleData(p), TokPickleSize(p));
    if (t == NULL) {
	TokFreePickle(p);
    } else {
	TokPersist(t, p);
    }
}

/*
 *  rcvpickle(svc) -- receive next message (returned as an allocated TokPickle)
 *
 *  Returns NULL if the port has been closed.
 *
 *  The contents of any CONFIG messages are noted here before return.
 */

TokPickle
TokRecvPickle(Tok_Service svc) {
    char buf[MAX_MSGLEN + 1];
    int n, len;
    TokPickle p;
    ConfigData cd = svc->configData;

    len = TokRecv(svc->net, buf, sizeof(buf));
    if (len < 0)
	return NULL;
    if (len == sizeof(buf)) {
    	TokAbort("Received message exceeds MAX_MSGLEN");
    }
    p = TokFullPickle(buf, len);
    if (buf[0] == TSVC_CONFIG) {
	TokUnpickleInt(p);			/* type == TSVC_CONFIG */
	TokUnpickleInt(p);			/* from */
	cd->myid = TokUnpickleInt(p);		/* to */
	TokUnpickleInt(p);			/* signature */
	cd->leader = TokUnpickleInt(p);		/* leader */
	n = TokUnpickleInt(p);			/* length of state array */
	TokExtractBytes(p, cd->states, n);	/* server states */
	TokResetPickle(p);			/* reset for rereading */
    }
    return p;
}
