/*
 * netw.c --
 *
 *	This file implements functions for communicating via UDP.
 *
 *	The environment variable TSVC_LOSE may be set to specify
 *	a percentage of incoming message to deliberately "lose".
 */

static char rcsid[] =
"$Id: netw.c,v 1.9 2001/08/31 23:36:40 gmt Exp $";

#include "tokInt.h"

#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

/*
 * Opaque network information structure:
 */
struct TokNetwork {
    int insock;			/* input socket fd */
    int outsock;		/* output socket fd */
    char *hostport;		/* host:port string */
    struct sockaddr_in *addrs;	/* server addresses */
    int lose;			/* drop incoming message if rand() < this */
};

/*  TokInitNetw(names, portnums, n) -- init interface */

TokNetwork
TokInitNetw(char *names[], int ports[], int n) {
    TokNetwork net;
    struct hostent *hp;
    struct sockaddr_in myaddr;
    int i, alen, port;
    char hostbuf[200], *s;

    net = TokAlloc(sizeof(struct TokNetwork));
    net->insock = socket(AF_INET, SOCK_DGRAM, 0);
    net->outsock = socket(AF_INET, SOCK_DGRAM, 0);
    if (net->insock < 0 || net->outsock < 0)
	TokPerror("can't create socket");

    myaddr.sin_family = AF_INET;
    myaddr.sin_addr.s_addr = INADDR_ANY;
    myaddr.sin_port = 0;
    if (bind(net->insock, (struct sockaddr *) &myaddr, sizeof(myaddr)) < 0)
	TokPerror("can't bind socket");

    alen = sizeof(myaddr);
    if (getsockname(net->insock, (struct sockaddr *) &myaddr, &alen) < 0)
	TokPerror("can't get port number");
    port = ntohs(myaddr.sin_port);
    sprintf(hostbuf, ":%d", port);
    net->hostport = TokSalloc(hostbuf);

    net->addrs = TokZalloc(n * sizeof(struct sockaddr_in));
    for (i = 0; i < n; i++) {
	hp = gethostbyname(names[i]);
	if (!hp) {
	    sprintf(hostbuf, "unknown host: %s", names[i]);
	    TokAbort(hostbuf);
	}
	memcpy(&net->addrs[i].sin_addr, hp->h_addr, hp->h_length);
	net->addrs[i].sin_family = AF_INET;
	net->addrs[i].sin_port = htons(ports[i]);
    }

    if ((s = getenv("TSVC_LOSE")) != NULL) 
	net->lose = (atoi(s) / 100.0) * RAND_MAX;
    else
	net->lose = 0;

    return net;
}

/*
 * TokTermNetw(net) -- shut down network and free storage.
 */
void
TokTermNetw(TokNetwork net) {
    close(net->insock);
    close(net->outsock);
    TokFree(net->addrs);
    TokFree(net->hostport);
    TokFree(net);
}

/*
 * TokHostPort(net) -- return host:port string for network interface.
 */
char *TokHostPort(TokNetwork net) {
    return net->hostport;
}

/*
 * TokSend(net, dest, buf, n) -- send n-byte message to dest.
 */
void
TokSend(TokNetwork net, int dest, void *buf, int n) {
    struct sockaddr *a = (struct sockaddr *) &net->addrs[dest];
    if (sendto(net->outsock, buf, n, 0, a, sizeof(*a)) >= 0)
	return;
    if (errno != ECONNREFUSED)		/* allow "no receiver" error */
	TokPerror("TokSend");
}

/*
 * TokRecv(net, buf, len) -- receive message, returning length.
 *
 * Returns -1 if the file descriptor has been closed.
 */
int
TokRecv(TokNetwork net, void *buf, int len) {
    for (;;) {
	len = read(net->insock, buf, len);	/* read port */
	if (rand() < net->lose)
	    continue;				/* simulate packet loss */
	if (len > 0)
	    return len;				/* if read successful */
	if (len < 0 && errno != EBADF)
	    TokPerror("TokRecv");		/* if not EOF / closed */
	return -1;				/* report this one failure */
    }
}
