/*
 * tkbroker.c --
 *
 *	This file implements a simple command-based interactive program
 *	for testing the token service.
 *
 *	usage:  tkbroker configfile
 *
 *	Commands are read from from standard input and the results
 *	are printed on standard output.
 *
 *	Commands are:
 *		# comment
 *		request name [x]
 *		update name [value]
 *		release name [value]
 *		list
 *		clear
 *		close
 */

static char rcsid[] =
"$Id: tkbroker.c,v 1.5 2003/07/18 02:28:01 gmt Exp $";
 
#include "tok.h"			/* use only the public interface */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static Tok_Token lookup(char *name);
static void enter(Tok_Token t);
static void delete(Tok_Token t);

static void revoker(Tok_Token t, void *arg);
static char* data(Tok_Token t);

static void list(void);
static void clear(void);
static void request(char *name, char *flag);
static void update(char *name, char *value);
static void release(char *name, char *value);

static void kaboom(char *msg);
static int words(char *s, char *w[], int n);

#define MAXTOKENS 25			/* maximum simultaneously held tokens */
static Tok_Token tokens[MAXTOKENS];	/* current token collection */

static Tok_Service svc;			/* token service handle */

int
main(int argc, char *argv[]) {
    char **servers;
    char line[100], *w[3];
    int n;

    if (argc != 2)
	kaboom("usage: broker configfile");
    servers = Tok_Config(argv[1]);
    if (!servers)
	kaboom("invalid configuration");

    printf("[connecting...]\n");
    svc = Tok_Open(servers);
    printf("[ready]\n");

    for (;;) {
	
	printf("> ");
	if (fgets(line, sizeof(line), stdin) == 0)
	    break;
	if (! isatty(fileno(stdin)))
	    printf("%s", line);
	n = words(line, w, 3);
	if (n == 0 || w[0][0] == '#')
	    continue;

	if (strcmp(w[0], "close") == 0)
	    break;
	else if (strcmp(w[0], "list") == 0)
	    list();
	else if (strcmp(w[0], "request") == 0)
	    request(w[1], w[2]);
	else if (strcmp(w[0], "update") == 0)
	    update(w[1], w[2]);
	else if (strcmp(w[0], "release") == 0)
	    release(w[1], w[2]);
	else if (strcmp(w[0], "clear") == 0)
	    clear();
	else
	    printf("\tUNKNOWN COMMAND: %s\n", w[0]);
    }

    printf("\n[closing]\n");
    Tok_Close(svc);
    return 0;
}

/*
 *  lookup(name) -- find token in local collection
 */
static Tok_Token
lookup(char *name) {
    int i;
    for (i = 0; i < MAXTOKENS; i++)
	if (tokens[i] != NULL && strcmp(Tok_GetName(tokens[i]), name) == 0)
	    return tokens[i];
    return NULL;
}

/*
 *  enter(token) -- add token to collection
 */
static void
enter(Tok_Token t) {
    int i;
    for (i = 0; i < MAXTOKENS; i++)
	if (tokens[i] == NULL) {
	    tokens[i] = t;
	    return;
	}
    kaboom("no room for another token");
}

/*
 *  delete(token) -- remove token from collection
 */
static void
delete(Tok_Token t) {
    int i;
    for (i = 0; i < MAXTOKENS; i++)
	if (tokens[i] == t) {
	    tokens[i] = NULL;
	    return;
	}
}

/*
 *  revoker(token, arg) -- process REVOKE request from server
 */
static void
revoker(Tok_Token t, void *arg) {
    printf("\n\tREVOKE %s\n", Tok_GetName(t));
}

/*
 *  data(token) -- return data string in C form using temporary buffer.
 */
static char *
data(Tok_Token t) {
    static char *buf;
    static int bufl;
    int len;

    len = Tok_GetLength(t);
    if (bufl < len + 1) {
	free(buf);
	bufl = len + 100;
	buf = malloc(bufl);
    }
    memcpy(buf, Tok_GetData(t), len);
    buf[len] = '\0';
    return buf;
}

/*
 *  list() -- implement "list" command
 */
static void
list(void) {
    int i, n;
    Tok_Token t;

    for (i = n = 0; i < MAXTOKENS; i++)
	if ((t = tokens[i]) != NULL) {
	    printf("\t%s  %s\n", Tok_GetName(t), data(t));
	    n++;
	}
    if (n == 0)
	printf("\tNONE HELD\n");
}

/*
 *  clear() -- implement "clear" command
 */
static void
clear(void) {
    int i, n;
    Tok_Token t;
    for (i = n = 0; i < MAXTOKENS; i++)
	if ((t = tokens[i]) != NULL) {
	    Tok_Release(t);
	    tokens[i] = NULL;
	    n++;
	}
    if (n == 0)
	printf("\tNONE HELD\n");
}

/*
 *  request(name, flag) -- implement "request" command
 */
static void
request(char *name, char *flag) {
    int f;
    Tok_Token t;

    if (!*name) {
	printf("\tREQUEST WHAT?\n");
	return;
    }
    if (lookup(name) != NULL) {
	printf("\tALREADY HELD\n");
	return;
    }
    if (*flag == 'x' || *flag == 'w')
	f = TOK_EXCLUSIVE;
    else
	f = TOK_SHARED;
    t = Tok_Request(svc, name, f, revoker, NULL);
    enter(t);
    printf("\t%s  %s\n", Tok_GetName(t), data(t));
}

/*
 *  update(name, value) -- implement "update" command
 */
static void
update(char *name, char *value) {
    Tok_Token t;
    
    if (!*name) {
	printf("\tUPDATE WHAT?\n");
	return;
    }
    t = lookup(name);
    if (t == NULL) {
	printf("\tNOT HELD\n");
	return;
    }
    if (*value) {
	Tok_SetData(t, value, strlen(value));
    } else {
	Tok_SetData(t, NULL, 0);
    }
    Tok_Update(t);
    printf("\tOK\n");
}

/*
 *  release(name, value) -- implement "release" command
 */
static void
release(char *name, char *value) {
    Tok_Token t;
    
    if (!*name) {
	printf("\tRELEASE WHAT?\n");
	return;
    }
    t = lookup(name);
    if (t == NULL) {
	printf("\tNOT HELD\n");
	return;
    }
    if (*value)
	Tok_SetData(t, value, strlen(value));
    Tok_Release(t);
    delete(t);
    printf("\tOK\n");
}

/*
 *  kaboom(message) -- abort program
 */
static void kaboom(char *s) {
    fprintf(stderr, "abort: %s\n", s);
    exit(1);
}

/*  words(s,w,n) - break string into words
 *	s is modified in place with \0 replacing "isspace()" separators
 *	w is an array of n char pointers to receive addresses of words
 *	return value gives number of words found
 */
static int
words(char *s, char *w[], int n) {
    int i;

    for (i = 0; i < n; ) {
	while (isspace(*s))
	    s++;
	if (!*s)
	    break;
	w[i++] = s;
	while (*s && !isspace(*s))
	    s++;
	if (!*s)
	    break;
	*s++ = '\0';
    }
    while (i < n)
	w[--n] = s;
    return i;
}
