/*
 * tokInt.h --
 *
 *	This header files describes the facilities internal to the
 *	token client.
 *
 * $Id: tokInt.h,v 1.6 2001/09/25 17:40:15 gmt Exp $
 */

#ifndef _TOKINT
#define _TOKINT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "tok.h"
#include "tokPort.h"

#define FALSE 0
#define TRUE 1



/*
 * Token service message type codes.
 */
#define TSVC_LOGIN	11
#define TSVC_CONFIG	12
#define TSVC_CATALOG	13
#define TSVC_ALIVE	14
#define TSVC_LOGOUT	15

#define TSVC_REQUEST	21
#define TSVC_GRANT	22
#define TSVC_REVOKE	23
#define TSVC_RETURN	24
#define TSVC_CONFIRM	25


/*
 * Message flag values.
 */
#define FLAG_EXCLUSIVE	-1
#define FLAG_SHARED	+1

#define FLAG_UPDATE	1
#define FLAG_DEASSIGN	2


/*
 *  Token state values.
 *  States advance in sequence, and also a few other ways, e.g.:
 *	from HELD to LEAVING on receipt of RevokeMessage
 *	from HELD to RELEASED by release() call
 *	from REVISED to HELD on receiving confirmation of update
 */
#define TOK_FREE	0	/* initial state */
#define TOK_REQUESTED	1	/* requested from server */
#define TOK_RECEIVED	2	/* GrantMessage received */
#define TOK_HELD	3	/* requesting thread awakened */
#define TOK_REVISED	4	/* updated, not yet confirmed */
#define TOK_LEAVING	5	/* RevokeMessage received */
#define TOK_RELEASED	6	/* released, not yet confirmed */



/*
 * Other constants.
 */
#define HBT_INTERVAL 3		/* heartbeat interval, in seconds */
#define LOGIN_INTERVAL 1	/* interval between login attempts */
#define RESEND_INTERVAL 1	/* interval between reissuing messages */
#define RECHECK_INTERVAL 1	/* interval between rechecking unack'd msgs */
#define ACK_NBINS 17		/* number of hash bins for holding unack msgs*/
#define TOK_NBINS 97		/* number of hash bins for held tokens */
#define MAX_MSGLEN 10000	/* maximum message handled */



/* Token Service structure */

struct TokService {
    TokNetwork net;			/* private data for network interface */
    struct TokClientData *clientData;		/* private data for client.c */
    struct TokConfigData *configData;		/* private data for config.c */
    struct TokHoldingsData *holdingsData;	/* private data for holdings.c*/
    struct TokReissueData *reissueData;		/* private data for reissue.c */
};



/* Token structure */

struct TokToken {
    struct TokToken *next;		/* next token in this hash chain */
    Tok_Service service;		/* token service assoc w/ this token */
    char *name;				/* token name */
    int hashcode;			/* hash value derived from name */
    unsigned char state;		/* current state of this token */
    unsigned char dirty;		/* is data value unknown to server? */
    signed char access;			/* TOK_SHARED or TOK_EXCLUSIVE */
    char *data;				/* pointer to data */
    int datalen;			/* length of data */
    void (*callback)(Tok_Token,void*);	/* callback on revoke */
    void *clientData;			/* data value passed as revoke arg */
    long msgnum;			/* message number awaiting ack */
    TokMutex mutex;			/* access interlock */
    TokCond cv;				/* condition variable */
};


/* Pickle structure */

typedef struct TokPickle *TokPickle;

void Tok_init(char *cfgfile);



/*
 * Prototypes:
 */

/* config.c */

TokNetwork	TokSetConfig(Tok_Service svc, char *servers[]);
void		TokSetStates(Tok_Service svc, char *a);
int		TokServer(Tok_Service svc, char *tkname);
void		TokSendVals(Tok_Service svc,int type,int dest,char *fmt,...);
TokPickle	TokRecvPickle(Tok_Service svc);

/* hash.c */

int		TokHash(char *s);
int		TokRehash(int n);
int		TokHashList(char *s[]);

/* holdings.c */

void		TokInitHoldings(Tok_Service svc);
void		TokTermHoldings(Tok_Service svc);
Tok_Token	TokLookup(Tok_Service svc, char *name, int create);
void		TokDelete(Tok_Token t);
void		TokSendCatalog(Tok_Service svc, int dest);

/* pickle.c */

TokPickle	TokEmptyPickle(void);
TokPickle	TokFullPickle(void *buf, int len);
void		TokFreePickle(TokPickle p);

int		TokPickleSize(TokPickle p);
void *		TokPickleData(TokPickle p);
void		TokResetPickle(TokPickle p);

void		TokPickleInt(TokPickle p, long v);
void		TokPickleString(TokPickle p, char *s);
void		TokPickleBytes(TokPickle p, void *buf, int n);
void		TokAppendBytes(TokPickle p, void *buf, int n);

long		TokUnpickleInt(TokPickle p);
char *		TokUnpickleString(TokPickle p);
int		TokUnpickleBytes(TokPickle p, void *buf, int n);
int		TokExtractBytes(TokPickle p, void *buf, int n);

/* reissue.c */

void		TokInitReissue(Tok_Service svc);
void		TokTermReissue(Tok_Service svc);
void		TokPersist(Tok_Token t, TokPickle p);
Tok_Token	TokForget(Tok_Service svc, long seqnum);


#endif /* _TOKINT */
