/*  decode.c -- code for interpreting the TopoVista tile format  */

#include "topo.h"

#define WEST 0
#define EAST 1
#define NORTH 8
#define SOUTH 9



static void linkTile(Tile *newt);



/**
 *  Interpolates two elevation values, handling MISSING_DATA.
 */
static int interp(Tile *t1, int x1, int y1, Tile *t2, int x2, int y2) {
  int e1 = EMDATA(t1, x1, y1);
  int e2 = EMDATA(t2, x2, y2);
  if( e1 == MISSING_DATA )
    return e2;
  else if( e2 == MISSING_DATA )
    return e1;
  else 
    return (e1 + e2 + 1) / 2;
}



/**
 *  Removes gaps along tile boundaries by forcing shared elevations
 *  to the values from the finer-resolution tile.
 *  A tie is broken in favor of the first tile.
 */
static void stitch(Tile *t0, Tile *t1, int o) {
  int i, n, x0, x1, y0, y1, dx, dy;
  Tile *t2;

  if( t1->dlevels > t0->dlevels ) {
    t2 = t0;
    t0 = t1;
    t1 = t2;
    o ^= 1;
  }

  n = 1 << tileOrder;
  x0 = x1 = y0 = y1 = dx = dy = 0;

  switch( o ) {
    case WEST:   dy = 1;  x1 = n;  break;
    case EAST:   dy = 1;  x0 = n;  break;
    case NORTH:  dx = 1;  y0 = n;  break;
    case SOUTH:  dx = 1;  y1 = n;  break;
  }
  for( i = 0; i < n; i++ ) {
    EMDATA(t1, x1, y1) = EMDATA(t0, x0, y0);
    x0 += dx;  y0 += dy;
    x1 += dx;  y1 += dy;
  }
}



/* this is magic to get the correct neighbor pointers between tiles. */
/* the caller must hold the wrapLock mutex. */

static int linkType[16] = {1,2,3,0,2,1,0,3,3,0,2,1,0,3,1,2};

static void linkNeighbors(Tile *a, Tile *b, int o, int d)
{
  RATTLE(wrapLock);
  stitch(a, b, d);
  switch(linkType[o+d]) {
  case 0:
    a->L2 = b;
    b->L1 = a;
    break;
  case 1:
    a->L1 = b;
    b->L2 = a;
    break;
  case 2:
    a->R1 = b;
    b->R2 = a;
    break;
  case 3:
    a->R2 = b;
    b->R1 = a;
    break;
  }
}



/**
 *  Interpolates missing data values if some trailing elevation records
 *  were lost from the input stream.
 */
static void interpMissing(Tile *t, int side, int quantum, int d) {
  int i, j, xq, yq;
  int dd = 2 * d;

  for( i = 0; i < side; i += dd ) {
    for( j = d; j < side; j += dd ) {
      EMDATA(t, i, j) = interp(t, i, j-d, t, i, j+d);
      EMDATA(t, j, i) = interp(t, j-d, i, t, j+d, i);
    }
  }

  for( i = d; i < side; i += dd ) {
    xq = i / quantum;
    for( j = d; j < side; j += dd ) {
        yq = j / quantum;
        if( (xq ^ yq) & 1 )
          EMDATA(t, i, j) = interp(t, i-d, j+d, t, i+d, j-d);
        else
          EMDATA(t, i, j) = interp(t, i-d, j-d, t, i+d, j+d);
    }
  }
}



/**
 *  Reads a tile corner value, handling scaling and MISSING_DATA.
 */
static void readCorner(inp f, Tile *newt, int x, int y) {
  long v;

  readInteger(f, &v);
  if( v != MISSING_DATA )
    v *= newt->scaleZ;
  EMDATA(newt, x, y) = v + ELEVATION_BIAS;
}



/**
 *  Reads a regular subgrid using differences from interpolated values.
 *  (If dx < dy, interpolation is vertical, else horizontal.)
 */
static void readInterp(inp f, Tile *newt, int x0,int y0,int dx,int dy, int edge)
{
  int e, e1, e2, x, y, xoff, yoff, xlim, ylim;
  long d;
  float scale = newt->scaleZ;
  float unscale = 1.0 / scale;

  if( dx < dy ) {
    xoff = 0;
    yoff = dy / 2;		/* interpolate vertically */
  } else {
    xoff = dx / 2;		/* interpolate horizontally */
    yoff = 0;
  }
  xlim = x0 + edge - xoff;
  ylim = y0 + edge - yoff;

  for( y = y0; y <= ylim; y += dy ) {
    for( x = x0; x <= xlim; x += dx ) {
      readInteger(f, &d);
      e1 = EMDATA(newt, x - xoff, y - yoff) - ELEVATION_BIAS;
      e2 = EMDATA(newt, x + xoff, y + yoff) - ELEVATION_BIAS;
      if( e1 != MISSING_DATA )  e1 = e1 * unscale + (e1 >= 0 ? 0.5 : -0.5);
      if( e2 != MISSING_DATA )  e2 = e2 * unscale + (e2 >= 0 ? 0.5 : -0.5);
      e = (e1 + e2) / 2 + d;
      if( e > MISSING_DATA + 2 )
        e *= scale;
      else
        e = MISSING_DATA;
      EMDATA(newt,x,y) = e + ELEVATION_BIAS;
    }
  }
}


/**
 * Reads data from a TopoVista tile header record.
 */
static void addHeader(Tile *newt, inp rec, int flag) {
  char title[100], version[20], comment[1000];
  long r, zone, datum, k, ht, xxx, yyy, g, l, h;
  int o, treeSize;

  readString(rec, title, sizeof(title));	/* 1. title */
  readString(rec, version, sizeof(version));	/* 2. version */
  readString(rec, comment, sizeof(comment));	/* 3. comment */
  readInteger(rec, &r);				/* 4. resolution */
  readInteger(rec, &zone);			/* 5. zone */
  readInteger(rec, &datum);			/* 6. datum */
  readInteger(rec, &k);				/* 7. order */
  readInteger(rec, &ht);			/* 8. error levels */
  readInteger(rec, &xxx);			/* 9. x */
  readInteger(rec, &yyy);			/* 10. y */
  readInteger(rec, &g);				/* 11. elevation granularity */
  readInteger(rec, &l);				/* 12. minimum elevation */
  readInteger(rec, &h);				/* 13. maximum elevation */

  newt->scaleZ = g / 100.0;
  newt->h = (h * newt->scaleZ) + ELEVATION_BIAS;
  newt->l = (l * newt->scaleZ) + ELEVATION_BIAS;

  if( r != mPerCell )
    fatal("addHeader: incompatible tile resolution");
  if( zone != utmZone )
    fatal("addHeader: incompatible UTM zone");
  if( k != tileOrder )
    fatal("addHeader: incompatible tile order");
  if( newt->xx != XXX2UTM(xxx) || newt->yy != YYY2UTM(yyy) )
    fatal("addHeader: xx/yy mismatch");

  tileSizeM = (rows - 1) * r;

  newt->L1 = newt->R1 = newt->L2 = newt->R2 = NULL;
  if( xxx & 1 ) {
    o = 2;
    newt->ly1 = rows-1; newt->ry1 = 0;
    newt->ly2 = 0;	newt->ry2 = rows-1;
    newt->lx3 = cols-1; newt->rx3 = 0;
  } else {
    o = 0;
    newt->ly1 = 0;	newt->ry1 = rows-1;
    newt->ly2 = rows-1; newt->ry2 = 0;
    newt->lx3 = 0;	newt->rx3 = cols-1;
  }
  if( yyy & 1 ) {
    o += 1;
    newt->lx1 = cols-1; newt->rx1 = 0;
    newt->lx2 = 0;	newt->rx2 = cols-1;
    newt->ly3 = 0;	newt->ry3 = rows-1;
  } else {
    newt->lx1 = 0;	newt->rx1 = cols-1;
    newt->lx2 = cols-1; newt->rx2 = 0;
    newt->ly3 = rows-1; newt->ry3 = 0;
  }

  treeSize = (1 << (2 + 2*k)) - 1;
  newt->data = alloc(rows * cols * sizeof(short));
  newt->err = alloc(treeSize * sizeof(short));

  newt->info = zalloc(treeSize * sizeof(char));
  newt->info[LCHILD(0)] = SURFACE;		/* ensure valid state */
  newt->info[RCHILD(0)] = SURFACE;
}



/**
 * Reads data from a TopoVista tile elevation record.
 * Assumes that no previous elevation records were missing.
 */
static void addElev(Tile *newt, inp rec, int flag) {
  int i, dx;
      
  i = flag - TV_ELEV;
  if( i == 0 ) {
    readCorner(rec, newt, 0, 0);
    readCorner(rec, newt, cols-1, 0);
    readCorner(rec, newt, 0, rows-1);
    readCorner(rec, newt, cols-1, rows-1);
  } else {
    dx = 1 << (tileOrder - i);
    readInterp(rec, newt, dx, 0, 2 * dx, 2 * dx, rows - 1); 
    readInterp(rec, newt, 0, dx, dx, 2 * dx, rows - 1); 
  }
}



/**
 * Reads data from a TopoVista tile error record.
 * Assumes that no previous error records were missing.
 */
static void addErrs(Tile *newt, inp rec, int flag) {
  int e, enext, i, j, n;
  long v;
  float z;

  i = flag - TV_ERRS;
  n = 1 << i;
  enext = n - 1;
  z = newt->scaleZ;

  for( j = 0; j < n; j++ ) {
    e = newt->err[PARENT(enext)];
    readInteger(rec, &v);
    newt->err[enext] = (unsigned short)(e + z * v);
    enext++;
  }

  if( i == 0 ) {	/* fix up first generation which has no parent */
    newt->err[0] = (unsigned short)(z * v);
  }
}



/**
 *  Builds a partial or complete tile and links it to its neighbors,
 *  given a tile for which some data records have been collected.
 *
 *  Returns 1 if new data has been stored, or 0 if not.
 */
int buildTile(Tile *newt)
{
  int c, dx, i, len, o, q;
  int dbase, dlast, ebase, elast, hgt, unknown, xxx, yyy;
  unsigned char *data;
  inp in;
  Triangle t;
  TileRec *r;
  Wrap *wrap = newt->wrap;

  dbase = 0;
  dlast = dbase + newt->dlevels;
  ebase = tileOrder + 1;
  elast = ebase + newt->elevels;

  /* need header before we can start */
  if( newt->data == 0 && HEADREC(newt).tr_data == NULL )
    return 0;				/* if header never seen */

  /* process the data records */
  for( i = 0; i < RECCOUNT; i++ ) {
    r = &newt->trec[i];
    data = r->tr_data;
    if( data == NULL )
      continue;				/* if this record is absent */

    len = r->tr_length;
    newt->nbytes += len + TV_RHLEN;
    in = openMem(data, len);
    if( r->tr_flags & TV_GZIP )
      in = unzip(in);

    /* dispatch to proper processor, if this record can be processed now */
    if( i == 0 )
      addHeader(newt, in, r->tr_flags);		/* header */
    else if( i == elast + 1 ) {
      addErrs(newt, in, r->tr_flags);		/* error data */
      elast = i;
      newt->elevels = elast - ebase;
    } else if( i > elast ) {
      continue;				/* don't delete, will need later */
    } else if( i == dlast + 1 ) {
      addElev(newt, in, r->tr_flags);		/* elevation data */
      dlast = i;
      newt->dlevels = dlast - dbase;
    } else if( i > dlast && i < ebase ) {
      continue;				/* don't delete, will need later */
    }

    in->close(in);
    free(r->tr_data);
    r->tr_data = NULL;
    r->tr_length = 0;
    r->tr_done = 1;
  }

  #ifdef LOG_TILES
    c = 'a' + (dlast - dbase) - 2;
    if( c < 'a' )
      c = 'z';
    fprintf(stderr, "  tile %s  ID %08X  read %c+%02d %6d bytes\n",
      newt->tname, newt->tid, c, elast - ebase, newt->nbytes);
  #endif

  xxx = TIDXXX(newt->tid);
  yyy = TIDYYY(newt->tid);
  o = "\06\00\04\02" [2 * (xxx & 1) + (yyy & 1)];  /* LCHILD orientation */

  /* interpolate missing elevation values */
  q = dx = 1 << (tileOrder - (dlast - dbase - 1));
  while( dx > 1 )
    interpMissing(newt, rows, q, dx >>= 1);

  /* complete the error tree */
  /* all triangles with tree height <= unknown need error calculated */
  /* (leaf triangles have tree height 0) */

  /* (We could do something clever here like do errDFS from the
      triangles at height ht-1 but that would involve calculating the
      orientation of these triangles which is a ruler-function-like
      thing.) */

  hgt = 2 * newt->dlevels;		/* total height of error tree */
  unknown = hgt - newt->elevels;	/* height of first known level */
  t.tile = newt;
  t.i = 0;
  ERR(t) = 0;
  if( unknown > 0 ) {
    t.i = LCHILD(0);
    errDFS(t, unknown, hgt - 1, o,
      newt->lx1, newt->ly1, newt->lx2, newt->ly2, newt->lx3, newt->ly3);
    t.i = RCHILD(0);
    errDFS(t, unknown, hgt - 1, o + 4,
      newt->rx1, newt->ry1, newt->rx2, newt->ry2, newt->rx3, newt->ry3);
  }

  /* compute surface normals */
  surfaceNormal(newt);
  colorTile(newt);

  /* make this a tile, and link it in */
  wrap->type = TYPE_TILE;
  linkTile(newt);
  return 1;
}



/**
 *  Links a wrapped Tile into the hierarchy.
 */
static void linkTile(Tile *newt)
{
  int eh, xxx, yyy, o;
  Wrap *w;
  WrapGen *wg;
  Tile *tile;

  xxx = UTM2XXX(newt->xx);
  yyy = UTM2YYY(newt->yy);
  o = "\06\00\04\02" [2 * (xxx & 1) + (yyy & 1)];  /* LCHILD orientation */

  /* update globals */
  LOCK(wrapLock);
  eh = eyeZAbsolute - zzmax;
  xxmin = MIN(xxmin, newt->xx);
  xxmax = MAX(xxmax, newt->xx + tileSizeM);
  yymin = MIN(yymin, newt->yy);
  yymax = MAX(yymax, newt->yy + tileSizeM);
  if( zzmin == 0 && zzmax == 0 ) {	/* if very first tile */
    zzmin = newt->l;
    zzmax = newt->h;
    eyeZAbsolute = zzmax + eh; 
    setEyeRefHeight();
  } else {
    zzmin = MIN(zzmin, newt->l);
    zzmax = MAX(zzmax, newt->h);
    eyeZAbsolute = zzmax + eh; 
  }
  postRedisplays(RC);			/* because eyeZAbsolute updated */

  /* link newt into L1,R1,L2,R2 pointer network */
  wg = genWraps(TYPE_TILE);
  while( (w = nextWrap(wg)) != NULL ) {
    tile = w->data;
    if( tile->xx == newt->xx + tileSizeM && tile->yy == newt->yy ) {
      linkNeighbors(newt, tile, o, EAST);
    } else if( tile->xx == newt->xx - tileSizeM && tile->yy == newt->yy ) {
      linkNeighbors(newt, tile, o, WEST);
    } else if( tile->xx == newt->xx && tile->yy == newt->yy + tileSizeM ) {
      linkNeighbors(newt, tile, o, NORTH);
    } else if( tile->xx == newt->xx && tile->yy == newt->yy - tileSizeM ) {
      linkNeighbors(newt, tile, o, SOUTH);
    }
  }

  UNLOCK(wrapLock);
}
