/*  ingzip.c -- functions for reading from a possibly zipped file  */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "topo.h"



#ifdef NO_ZLIB

    inp unzip(inp src) {
        if (src->peek(src) != 0x1F)	/* if input is not gzipped */
	    return src;
	else {
	    fprintf(stderr, "  no ZLIB -- can't handle compressed input\n");
	    exit(1);
	}
    }

#else 



#include <zlib.h>

#define IBUFL 4096		/* input buffer length */
#define OBUFL 4096		/* output buffer length */



typedef struct {
    struct instruct in;
    inp src;
    z_streamp strm;
    unsigned char *ibuf;
    unsigned char *obuf, *onext;
} ins_zip;

static int hskip(inp);
static int peekZip(inp);
static int charZip(inp);
static void closeZip(inp);
static void fill(ins_zip *);



/*  unzip(in) -- interpose unzipper ahead of input.  */

inp unzip(inp src) {
    ins_zip *inz;

    if (src->peek(src) != 0x1F)		/* if input is not gzipped */
	return src;

    if (hskip(src) == 0)		/* skip gzip header */
	return NULL;

    inz = alloc(sizeof(ins_zip));
    inz->in.peek = peekZip;
    inz->in.rchar = charZip;
    inz->in.rbytes = readBytes;
    inz->in.rline = readLine;
    inz->in.close = closeZip;
    inz->src = src;

    inz->strm = alloc(sizeof(z_stream));
    inz->strm->next_in = inz->ibuf = alloc(IBUFL);
    inz->strm->next_out = inz->obuf = inz->onext = alloc(OBUFL);
    inz->strm->avail_in = 0;
    inz->strm->avail_out = 0;
    inz->strm->opaque = NULL;
    inz->strm->zalloc = NULL;
    inz->strm->zfree = NULL;

    inflateInit2(inz->strm, -MAX_WBITS);  /* magic flag means no zlib header */
    return &inz->in;
}



/*  peekZip(in) -- peek at next character.  */

static int peekZip(inp in) {
    ins_zip *inz = (ins_zip *) in;
    if (inz->onext == inz->strm->next_out)
	fill(inz);
    if (inz->onext == inz->strm->next_out)
	return EOF;
    return *inz->onext;
}



/*  charZip(in) -- read next character.  */

static int charZip(inp in) {
    ins_zip *inz = (ins_zip *) in;
    if (inz->onext == inz->strm->next_out)
	fill(inz);
    if (inz->onext == inz->strm->next_out)
	return EOF;
    return *inz->onext++;
}



/*  closeZip(in) -- close file and invalidate structure.  */

static void closeZip(inp in) {
    ins_zip *inz = (ins_zip *) in;
    inz->src->close(inz->src);
    inflateEnd(inz->strm);
    free(inz->strm);
    free(inz->ibuf);
    free(inz->obuf);
    memset(in, 0, sizeof(ins_zip));
    free(in);
}



/*  fill(inz) -- refill zstream output buffer, which is known empty.  */

static void fill(ins_zip *inz) {
    z_streamp z = inz->strm;
    if (z->avail_in == 0) {
	z->next_in = inz->ibuf;
	z->avail_in = inz->src->rbytes(inz->src, z->next_in, IBUFL);
    }
    z->next_out = inz->onext = inz->obuf;
    z->avail_out = OBUFL;
    inflate(z, Z_SYNC_FLUSH);
}



/*
 *  hskip(in) -- skip gzip header
 *
 *  Returns 0 if unsuccessful.
 *
 *  Based on "gzio.c" from Zlib version 1.1.3,
 *  Copyright (C) 1995-1998 Jean-loup Gailly,
 *  http://www.gzip.org/zlib/.
 */

/* gzip flag byte */
#define ASCII_FLAG   0x01 /* bit 0 set: file probably ascii text */
#define HEAD_CRC     0x02 /* bit 1 set: header CRC present */
#define EXTRA_FIELD  0x04 /* bit 2 set: extra field present */
#define ORIG_NAME    0x08 /* bit 3 set: original file name present */
#define COMMENT      0x10 /* bit 4 set: file comment present */
#define RESERVED     0xE0 /* bits 5..7: reserved */

static int hskip(inp in) {
    int method, flags, c, len;

    if (in->rchar(in) != 0x1F)		/* first magic byte */
	return 0;
    if (in->rchar(in) != 0x8b)		/* second magic byte */
	return 0;

    method = in->rchar(in);
    flags = in->rchar(in);
    if (method != Z_DEFLATED || (flags & RESERVED) != 0) {
	return 0;			/* unrecognized format */
    }

    for (len = 0; len < 6; len++)
	c = in->rchar(in);		/* discard time, xflags and OS code */

    if ((flags & EXTRA_FIELD) != 0) {	/* skip extra field */
	len = in->rchar(in);
	len += in->rchar(in) << 8;
	while (len-- != 0 && (c = in->rchar(in)) != EOF)
	    ;
    }

    if ((flags & ORIG_NAME) != 0) {	 /* skip original file name */
	while ((c = in->rchar(in)) != 0 && c != EOF)
	    ;
    }
    if ((flags & COMMENT) != 0) {	/* skip .gz file comment */
	while ((c = in->rchar(in)) != 0 && c != EOF)
	    ;
    }
    if ((flags & HEAD_CRC) != 0) {	/* skip header CRC */
	c = in->rchar(in);
	c = in->rchar(in);
    }

    return c != EOF;
}



#endif
