/*  input.h -- function declarations for TopoVista input system  */
 
/*
 *  This is an object-oriented input system written in C.
 *
 *  uopen(s) opens a stream, given a URL-type string in one of these forms:
 *	file:///path	file (standard URL; host must be empty)
 *	file:path	file (shortcut form; allows current directory)
 *	http://...	HTTP socket (standard URL)
 *	tvtp://...	TVTP socket (TopoVista tile server)
 *	literal:data	data contained within string (our invention)
 *  If the argument s is not of URL form, it is tried as a filename.
 *  If s is null, standard input is used.
 *
 *  uopen returns an "inp" pointer to an allocated object through which
 *  all subsequent calls are made.  uopen() returns NULL if unsuccessful.
 *
 *  bopen(s, len) handles the more general case where the argument
 *  s may contain embedded zero bytes, typically with "literal:".
 *
 *  unzip(in) returns an "inp" that can read from either a gzipped or
 *  normal source.  If the underlying source is not gzipped, or if the
 *  zlib library is not present, unzip() may just return its argument.
 *  If it returns a new object, it takes responsibility for closing the
 *  original source; in all cases the caller should subsequently use
 *  only the returned object.  
 *
 *  An inp object provides the following functions:
 *
 *  peek(in) and rchar(in) return the next character (or EOF) from the
 *  input.  rchar() advances the input stream while peek() does not.
 *
 *  rline(in, buf, len) reads the next line of input into a buffer.
 *  The input line may be terminated by CR, LF, or CRLF, and is stored
 *  in the buffer with just a zero-byte terminator.  rline returns
 *  buf for success or NULL for EOF.
 *  
 *  rbytes(in, buf, len) reads a fixed number of bytes into a buffer.
 *  It returns the number of bytes actually stored, which indicates EOF
 *  if less than len.
 *
 *  close(in) releases an input stream.  It closes the underlying file,
 *  if any, and releases all associated storage.  The argument object
 *  is invalidated by this call.
 */

#include <stdio.h>
#include <stdlib.h>

#define MAX_URL_LEN 1000

typedef struct instruct *inp;
struct instruct {
  int (*peek)(inp);			/* view char w/o consuming */
  int (*rchar)(inp);			/* read one character */
  int (*rbytes)(inp, void *, size_t);	/* read bytes into buffer */
  char *(*rline)(inp, char *, size_t);	/* read line into buffer */
  void (*close)(inp);			/* close file, invalidate obj */
};

typedef struct url {
  char scheme[10];			/* file, http, ftp, tvtp, etc. */
  char host[MAX_URL_LEN];		/* host name */
  int port;				/* port number, if specified */
  char path[MAX_URL_LEN];		/* file path */
} url, *urlp;



/* public functions */

inp uopen(char *URL);
inp bopen(char *URL, size_t len);
inp unzip(inp in);
int parseURL(char *string, url *result);



/* tstream.c functions */

int	readInteger (inp f, long *value);
int	readString  (inp f, char *buf, size_t len);



/* private functions */

inp openHTTP(char *URL);
inp openTVTP(char *URL);
inp openFile(char *filename);
inp openFP(FILE *fp);
inp openMem(void *addr, size_t len);
int readBytes(inp, void *, size_t);
char *readLine(inp, char *, size_t);
