/*  mqueue.c -- message queue functions  */

/*
 *  These functions implement message queues for process synchronization.
 *  The message queue structure is opaque and is manipulated only in here.
 *  Each "message" is a simple void* value whose interpretation is left
 *  to the user.
 */

#include "topo.h"

#ifndef UNITHREAD


typedef struct Message {	/* individual message structure */
  void *value;			/* message value */
  struct Message *prev;		/* backward link */
  struct Message *next;		/* forward link */
} Message;

struct MsgQueue {		/* message queue structure */
  pthread_mutex_t lock;		/* modification lock */
  pthread_cond_t cv;		/* condition variable for blocking */
  Message list;			/* header for linked list of messages */
};


/*
 * messageQueue() -- create new message queue and return opaque pointer.
 */
MsgQueue *messageQueue(void) {
  MsgQueue *mq = zalloc(sizeof(MsgQueue));	/* allocate queue structure */
  pthread_mutex_init(&mq->lock, NULL);		/* initialize lock */
  pthread_cond_init(&mq->cv, NULL);		/* init condition variable */
  mq->list.prev = mq->list.next = &mq->list;	/* init linked list */
  return mq;					/* return pointer */
}


/*
 * sendMessage(mq, msg) -- add message to end of queue.
 */
void sendMessage(MsgQueue *mq, void *msg) {
  Message *m = zalloc(sizeof(Message));
  m->value = msg;			/* initialize the message struct */
  pthread_mutex_lock(&mq->lock);	/* lock the linked list */
  m->next = &mq->list;			/* add the message to the list */
  m->prev = m->next->prev;
  m->prev->next = m->next->prev = m;
  pthread_cond_broadcast(&mq->cv);	/* signal any blocked readers */
  pthread_mutex_unlock(&mq->lock);	/* unlock the list */
}


/*
 * rcvMessage(mq) -- get next message from queue.
 * If no message is available, blocks until one is sent by another process.
 */
void *rcvMessage(MsgQueue *mq) {
  Message *m;
  void *val;

  pthread_mutex_lock(&mq->lock);		/* lock the list */
  while( mq->list.prev == &mq->list ) {		/* while queue is empty */
    pthread_cond_wait(&mq->cv, &mq->lock);	/* block on CV */
  }
  m = mq->list.next;				/* get entry from non-empty q */
  m->next->prev = m->prev;			/* remove from linked list */
  m->prev->next = m->next;
  pthread_mutex_unlock(&mq->lock);		/* unlock the list */

  val = m->value;				/* extract the message value */
  free(m);					/* free the entry */
  return val;					/* return the value */
}


#endif /*UNITHREAD*/
