/*  pnmops.c -- input operations on PNM files  */

#include "topo.h"

#define MASK8 0x80



static int getBit(inp in)
{
  static int c;
  static int bits = 0;
  int b;

  if( bits == 0 ) {
    c  = in->rchar(in);
    bits = 8;
  }
  b = c & MASK8;
  c <<= 1;
  bits--;
  return(b);
}



static void discardToCR(inp in)
{
  int c;

  while ((c = in->rchar(in)) != EOF && c != '\n')
    ;
}



static void getPNMToken(inp in, char *tok, int n)
{
  int c, i;

  while ((c = in->rchar(in)) != EOF && (isspace(c) || c == '#')) {
    if (c == '#') {
      discardToCR(in);
    }
  }
  tok[0] = c;
  i = 1;
  while (i < n && (c = in->peek(in)) != EOF && ! isspace(c) && c != '#') {
    tok[i++] = in->rchar(in);
  }
  tok[i] = '\0';
}



static void getPNMComment(inp in, char *tok, int n)
{
  int c, i;

  i = 0;
  while (isspace(in->peek(in)))
    in->rchar(in);
  if (in->peek(in) == '#') {
    tok[i++] = in->rchar(in);
    while (i < n && (c = in->rchar(in)) != EOF && c != '\n') {
      tok[i++] = c;
    }
    if (c == '\n' && i < n)
      tok[i++] = '\n';
    else if (c != '\n')
      discardToCR(in);		/* throw away the rest of the comment */
  }
  tok[i] = '\0';
}



static void getPNMLabel(inp in, char *tok, int n)
{
  int c, i;

  while ((c = in->rchar(in)) != EOF && isspace(c))
    ;
  i = 0;
  if (c == '#') {
    while (i < n && (c = in->rchar(in)) != EOF && c != '#') {
      tok[i++] = c;
    }
  }
  tok[i] = '\0';
}



int findLegend(Legend l,unsigned char r,unsigned char g,unsigned char b)
{
  LegendEntry *lep;
  int i;

  lep = l.entry;
  for(i=0;i<l.size;++i) {
    if(lep->red == r && lep->green == g && lep->blue == b)
      break;
    lep++;
  }
  return(i);
}



static void mergeLegend(Legend *lp, char *label,
			unsigned char r, unsigned char g, unsigned char b)
{
  int s,t;
  LegendEntry *lep;

  t = findLegend(*lp,r,g,b);
  if (t == lp->size) {
    if( t ) {
      lp->entry = (LegendEntry *)realloc(lp->entry,(t+1)*sizeof(LegendEntry));
    } else {
      lp->entry = (LegendEntry *)malloc(sizeof(LegendEntry));
    }
    if( lp->entry == NULL ) nospace();
    ++lp->size;
    lep = &(lp->entry[t]);
    lep->label = (char *)malloc((strlen(label)+1)*sizeof(char));
    if( lep->label == NULL ) nospace();
    strcpy(lep->label,label);
    lep->red = r;
    lep->green = g;
    lep->blue = b;
  } else {
    lep = &(lp->entry[t]);
    s = strlen(lep->label);
    lep->label = (char *)
      realloc(lep->label,(s+strlen(label)+2)*sizeof(char));
    lep->label[s] = ',';
    lep->label[s+1] = 0;
    strcat(lep->label,label);
  }
}



static void readPNMLegend(inp in, Legend *lp)
{
  int n,i;
  unsigned char r,g,b;
  char buf[72];
  char label[72];

  getPNMComment(in,buf,71);
  sscanf(buf,"# %d\n",&n);

  for(i=0;i<n;++i) {
    getPNMLabel(in,label,71);
    getPNMToken(in,buf,71);
    r = (unsigned char)(atof(buf)*255.);
    getPNMToken(in,buf,71);
    g = (unsigned char)(atof(buf)*255.);
    getPNMToken(in,buf,71);
    b = (unsigned char)(atof(buf)*255.);

    mergeLegend(lp,label,r,g,b);
  }
}



static Pnm *tryReadPNM(inp in)
{
  Pnm *pnm;
  int max=255,ptype;
  char buf[72];
  int row, col;
  unsigned char *pr, *pg, *pb;

  getPNMToken(in,buf,71);
  ptype = buf[1] - '0';
  if( buf[0] != 'P' || ptype < 1 || ptype > 6 ) {
    /*    fprintf(stderr, "Can't read pnm file (wrong format).\n");*/
    return(NULL);
  }

  pnm = (Pnm *)malloc(sizeof(Pnm));
  if (pnm == NULL) {
    fprintf(stderr, "Can't allocate space for Pnm structure.\n");
    return(NULL);
  }
  pnm->ll2xy = NULL;
  pnm->legend.size = 0;
  pnm->legend.entry = NULL;
  getPNMComment(in,buf,71);

  if( ! strcmp(buf,TOPOFLAG) ) {	/* 4 lat,lon,x,y reference points */
    getPNMComment(in,buf,71);
    if( ! strcmp(buf,LEGENDFLAG) ) {
      readPNMLegend(in,&(pnm->legend));
      getPNMComment(in,buf,71);
    }
    if( ! strcmp(buf,REFPTHEADER) )
      getPNMComment(in,buf,71);
    sscanf(buf,REFPTFORMATIN,
	   &(pnm->swUV.u),&(pnm->swUV.v),&(pnm->swXY.u),&(pnm->swXY.v));
    getPNMComment(in,buf,71);
    sscanf(buf,"# %lf %lf %lf %lf\n",
	   &(pnm->nwUV.u),&(pnm->nwUV.v),&(pnm->nwXY.u),&(pnm->nwXY.v));
    getPNMComment(in,buf,71);
    sscanf(buf,"# %lf %lf %lf %lf\n",
	   &(pnm->neUV.u),&(pnm->neUV.v),&(pnm->neXY.u),&(pnm->neXY.v));
    getPNMComment(in,buf,71);
    sscanf(buf,"# %lf %lf %lf %lf\n",
	   &(pnm->seUV.u),&(pnm->seUV.v),&(pnm->seXY.u),&(pnm->seXY.v));

    pnm->swUV.u *= 3600.;	/* Convert to arcseconds */
    pnm->swUV.v *= 3600.;
    pnm->nwUV.u *= 3600.;
    pnm->nwUV.v *= 3600.;
    pnm->neUV.u *= 3600.;
    pnm->neUV.v *= 3600.;
    pnm->seUV.u *= 3600.;
    pnm->seUV.v *= 3600.;

    pnm->ll2xy = createTransform(pnm->swUV, pnm->nwUV, pnm->neUV, pnm->seUV,
				 pnm->swXY, pnm->nwXY, pnm->neXY, pnm->seXY);
  }

  getPNMToken(in, buf, 71);
  pnm->width = atoi(buf);
  getPNMToken(in, buf, 71);
  pnm->height = atoi(buf);
  if( ptype != 1 && ptype != 4 ) {
    getPNMToken(in, buf, 71);
    max = atoi(buf);
  }

  if( ptype > 3 )
    in->rchar(in);	/* discard single character after maxval (i.e. \n) */

  pnm->red = malloc(pnm->width * pnm->height);
  pnm->green = malloc(pnm->width * pnm->height);
  pnm->blue = malloc(pnm->width * pnm->height);
  if (pnm->red == NULL || pnm->green == NULL || pnm->blue == NULL) {
    fprintf(stderr,"Can't allocate space for RGB array.\n");
    free(pnm->red);
    free(pnm->green);
    free(pnm->blue);
    free(pnm);
    return NULL;
  }


  /*
   * PNM files are stored starting at NW corner while DEM files
   * (and all structures in this program) start at SW corner.
   * So we have to flip the pnm file over.
   */
  for(row = pnm->height - 1; row >= 0; row--) {
    pr = pnm->red + row * pnm->width;
    pg = pnm->green + row * pnm->width;
    pb = pnm->blue + row * pnm->width;
    for(col = 0; col < pnm->width; col++) {
      switch(ptype) {
      case 1:
	getPNMToken(in,buf,71);
	if( buf[0] == '1' )
	  *pr = *pg = *pb = 0;
	else
	  *pr = *pg = *pb = 255;
	break;
      case 2:
	getPNMToken(in,buf,71);
	*pr = *pg = *pb = (unsigned char)(255.*(double)atoi(buf)/(double)max);
	break;
      case 3:
	getPNMToken(in,buf,71);
	*pr = (unsigned char)(255. * (double)atoi(buf) / (double)max);
	getPNMToken(in,buf,71);
	*pg = (unsigned char)(255. * (double)atoi(buf) / (double)max);
	getPNMToken(in,buf,71);
	*pb = (unsigned char)(255. * (double)atoi(buf) / (double)max);
	break;
      case 4:
	if( getBit(in) )
	  *pr = *pg = *pb = 0;
        else
	  *pr = *pg = *pb = 255;
	break;
      case 5:
	*pr = *pg = *pb = (unsigned char)(255.*in->rchar(in)/(double)max);
	break;
      case 6:
	*pr = (unsigned char)(255. * in->rchar(in) / (double)max);
	*pg = (unsigned char)(255. * in->rchar(in) / (double)max);
	*pb = (unsigned char)(255. * in->rchar(in) / (double)max);
	break;
      }
      pr++;
      pg++;
      pb++;
    }
  }
  return pnm;
}



Rgb *initRGB(int w, int h)
{
  Rgb *rgb;
  int x, y;
  unsigned char *red, *green, *blue;

  rgb = (Rgb *)malloc(sizeof(Rgb));
  if (rgb) {
    rgb->red = malloc(w * h);
    rgb->green = malloc(w * h);
    rgb->blue = malloc(w * h);
  }
  if (!rgb || !rgb->red || !rgb->green || !rgb->blue) {
    fprintf(stderr, "Can't allocate space for RGB array.\n");
    free(rgb);
    return NULL;
  }
  rgb->width = w;
  rgb->height = h;
  rgb->legend.size = 0;
  rgb->legend.entry = NULL;

  red = rgb->red;
  green = rgb->green;
  blue = rgb->blue;
  for (y = 0; y < h; ++y) {
    for(x = 0; x < w; ++x) {
      *red++ = 102;
      *green++ = 102;
      *blue++ = 102;
    }
  }

  return rgb;
}



Pnm *readPNM(char *filename)
{
  inp f;
  Pnm *pnm;
  Wrap *wrap;

  wrap = fetch(filename, TYPE_PNM);
  f = wrap->file;
  if( f == NULL )
    return destroy(wrap);
  f = unzip(f);
  pnm = tryReadPNM(f);
  f->close(f);
  wrap->file = NULL;
  if( pnm == NULL )
    return destroy(wrap);
  addWrap(wrap);
  return pnm;
}
