/*  surface.c -- compute surface based on error criteria  */


#include "topo.h"

static double xi, yi;	/* eye position in grid units */
static int height;
static double gridDistTh, gridErrTh;



static int tooCoarse(Triangle t, int x3, int y3)
{
  int h, dx, dy;
  double d;

  if( !INFOV(t) )	/* if not in field of view */
    return 0;

  /* Check distance from eye */
  dx = (int)(x3 - xi);
  dy = (int)(y3 - yi);
  d = (double)dx * (double)dx + (double)dy * (double)dy;
  if( d == 0 )				/* if right below eye */
    return 1;

  d = sqrt(d);				/* d in grid units */
  if( d  < gridDistTh * height )
    return 1;


  /* Check error threshold */
  h = ERR(t);

  if( h > d * gridErrTh )
    return 1;

  return 0;
}



static void errSplit(Triangle t,
		     int x1, int y1, int x2, int y2, int x3, int y3)
     /* splits triangle if it is too coarse. */
{
  int xm, ym;

  --height;
  if( !ISLEAF(t) && tooCoarse(t, x3, y3) ) {
    if( ONSURFACE(t) ) {
      split(t);
    }
    xm = (x1 + x2) / 2;
    ym = (y1 + y2) / 2;
    t.i = LCHILD(t.i);
    errSplit(t, x3, y3, x1, y1, xm, ym);
    t.i += 1;
    errSplit(t, x2, y2, x3, y3, xm, ym);
  }
  ++height;
}



void errSurface( void )
{
  Tile *tile;
  Triangle l,r;
  Wrap *wrap;
  WrapGen *wg;
  double starttime=0.0,endtime;

  if( tflag ) {
    starttime=getTime();
  }

  LOCK(wrapLock);
  numSurfTri = 0;
  wg = genWraps(TYPE_TILE);
  while( (wrap = nextWrap(wg)) != NULL ) {
    tile = wrap->data;
    l.tile = tile;
    r.tile = tile;
    /* SURFACE starts as triangles LCHILD(0) and RCHILD(0) */
    l.i = LCHILD(0);
    PUTONSURFACE(l);
    INFO(l) |= triInFOV(l);
    r.i = RCHILD(0);
    PUTONSURFACE(r);
    INFO(r) |= triInFOV(r);
    numSurfTri += 2;
  }

  gridErrTh = errThresh * mPerCell;
  gridDistTh = distThresh / mPerCell;
  wg = genWraps(TYPE_TILE);
  while( (wrap = nextWrap(wg)) != NULL ) {
    tile = wrap->data;
    xi = (eye[0] - tile->xx) / mPerCell;
    yi = (eye[1] - tile->yy) / mPerCell;
    l.tile = tile;
    r.tile = tile;
    l.i = LCHILD(0);
    r.i = RCHILD(0);

    height = 1 + 2 * tileOrder;
    errSplit(l, tile->lx1,tile->ly1,tile->lx2,tile->ly2,tile->lx3,tile->ly3);
    errSplit(r, tile->rx1,tile->ry1,tile->rx2,tile->ry2,tile->rx3,tile->ry3);
  }
  UNLOCK(wrapLock);

  if (tflag) {
    endtime=getTime();
    fprintf(stderr,"Surface built in %f seconds.\n",endtime-starttime);
  }
}
