/*  tstream.c -- typed stream input  */

/*
 *  See TypedOutputStream.java for a description of the file format.
 *
 *  20-feb-2001/gmt  initial version
 *  11-may-2001/gmt  fix missing includes, potential portability problem
 *  25-may-2001/wse  use GETC instead of getc
 *  11-aug-2001/gmt  reorder codes to group all int formats together
 *  19-sep-2001/gmt  use TopoVista input routines
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "input.h"



/*
 *  readInteger(file, &value) -- read integral value and store in *v.
 *
 *  returns 0 for success, -1 for EOF or failure.
 */
int readInteger(inp f, long *value) {
    int b, n;
    long v, diff;
    static long prevInt = 0;

    b = f->rchar(f);			/* read header byte */
    if (b == EOF)
	return -1;

    if (b < 0x40) {
	*value = prevInt = b;		/* 6-bit positive value */
	return 0;
    }
    if (b < 0x80) {
	*value = prevInt = b - 0x80;	/* 6-bit negative value */
	return 0;
    }
    if (b < 0xC0) {
	diff = -(b & 0x20) | (b & 0x1F);
	*value = prevInt += diff;	/* 6-bit signed difference */
	return 0;
    }

    if (b == 0xFE) {			/* null */
	*value = 0;
	return 0;
    }

    if ((b & 0xF8) == 0xC0) {		/* 11-bit signed value */
	v = ((- (b & 4)) | (b & 3)) << 8;
	b = f->rchar(f);
	if (b == EOF)
	    return -1;
	*value = prevInt = v | b;
	return 0;
    }

    if ((b & 0xF8) == 0xC8) {		/* 11-bit signed difference */
	v = ((- (b & 4)) | (b & 3)) << 8;
	b = f->rchar(f);
	if (b == EOF)
	    return -1;
	*value = prevInt += v | b;
	return 0;
    }

    if ((b & 0xF0) != 0xD0)
	return -1;			/* if not an integer */
    if ((b & 8) == 0)
	prevInt = 0;			/* if value follows, not difference */

    n = b & 7;				/* get bytecount-1 */
    b = f->rchar(f);
    if (b == EOF)
	return -1;
    v = (signed char) b;		/* sign extend first byte */

    while (n-- > 0) {			/* while more bits */
	b = f->rchar(f);
	if (b == EOF)
	    return -1;
	v = (v << 8) | b;		/* squeeze them in */
    }
    *value = prevInt += v;		/* combine with previous value */
    return 0;
}



/*
 *  readString(file, buf, len) -- read string into buffer.
 *
 *  returns actual length for success, -1 for EOF or failure.
 */
int readString(inp f, char *buf, size_t len) {
    size_t i;
    long n, tsp;

    int h = f->rchar(f);		/* read header byte */
    if (h == EOF)
	return -1;
    else if (h == 0xFE)			/* treat null as empty string */
	h = 0XE0;
    else if ((h & 0xF0) != 0xE0)
	return -1;			/* invalid header byte */

    n = h & 0x0F;			/* extract length */
    tsp = 0;

    if (n == 15) {
	if (readInteger(f, &n) < 0)	/* read first length value */
	    return -1;
	if (n < 0) {			/* if negative, counts trailg spaces */
	    tsp = -n;
	    if (readInteger(f, &n) < 0 || n < 0)      /* read trimmed strlen */
		return -1;
	}
    }

    for (i = 0; i < n; i++) {		/* read characters into buffer */
	int c = f->rchar(f);
	if (c == EOF)
	    return -1;
	if (i < len)
	    buf[i] = c;
    }

    n += tsp;				/* set total len including trailers */

    while (i < len && tsp-- > 0)	/* add trailing spaces that fit */
	buf[i++] = ' ';

    if (n < len)
	buf[n] = '\0';			/* terminate string */
    else
	buf[len - 1] = '\0';

    return n;				/* return total length */
}
