/*  tvtp.c -- asynchronous interface to TVTP protocol  */

#ifndef UNITHREAD

#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include "topo.h"

static void issue(void);
static void init(char *name);
static void *listener(void *dummy);
static void *timeout(void *dummy);
static void *newTile(void *tile);


static int udp = -1;		/* file descriptor for UDP port */



/*
 *  tvfetch(name) -- fetch file asynchronously using TVTP protocol.
 */
void tvfetch(char *name) {
  Wrap *wrap;

  if( udp < 0 )			/* if first call, start listener thread */
    init(name);

  LOCK(wrapLock);

  /* check for existing or in-progress tile */
  wrap = findWrap(name, TYPE_WANTED | TYPE_TILE | TYPE_UNAV);
  if( wrap != NULL ) {
    if( wrap->stale ) {
      freeTile(wrap->data);	/* delete old, stale tile */
      delWrap(wrap);
    } else {
      UNLOCK(wrapLock);		/* otherwise return if already have tile */
      return;
    }
  }

  /* add to list and send request */
  wrap = wantTile(name);
  issue();

  UNLOCK(wrapLock);

}



/*
 *  init(name) -- initialize port and threads.
 */
static void init(char *name) {
  url u;
  struct hostent *hostaddr;
  struct sockaddr_in sockaddr;
  pthread_t lthread;

  /* parse name */
  if( !parseURL(name, &u) )
    fatal("tvtp: unparseable");
  if( u.port == 0 )
    u.port = TV_PORT;

  /* look up host */
  hostaddr = gethostbyname(u.host);
  if( !hostaddr )
    fatal("tvtp: no hostaddr");

  /* create socket */
  if( (udp = socket(AF_INET, SOCK_DGRAM, 0)) < 0 )
    pfatal("tvtp socket");

  /* bind to port */
  memset(&sockaddr, 0, sizeof(sockaddr));
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_addr.s_addr = INADDR_ANY;
  sockaddr.sin_port = 0;
  if( bind(udp, (struct sockaddr *) &sockaddr, sizeof(sockaddr)) < 0 )
    pfatal("tvtp bind");

  /* connect remote port */
  sockaddr.sin_port = htons(u.port);
  memcpy(&sockaddr.sin_addr, hostaddr->h_addr, hostaddr->h_length);
  if( connect(udp, (struct sockaddr *) &sockaddr, sizeof(sockaddr)) < 0 )
    pfatal("tvtp connect");

  /* spawn timeout thread */
  pthread_create(&lthread, NULL, timeout, NULL);

  /* spawn listener thread */
  pthread_create(&lthread, NULL, listener, NULL);
}



/*
 *  issue() -- send request to server.
 *
 *  Scans the Wrap list looking for TYPE_WANTED entries, and sends
 *  up to MAX_ISSUE of them depending on the number of tiles already
 *  marked as "waiting".
 */
static void issue(void) {
  url u;
  Tile *tile;
  Wrap *wrap;
  WrapGen *wg;
  char request[MAX_URL_LEN];
  int incount = 0;
  
  LOCK(wrapLock);
  wg = genWraps(TYPE_WANTED | TYPE_TILE);
  while( (wrap = nextWrap(wg)) != NULL ) {
    tile = wrap->data;
    if( tile->waiting > 0 )
      incount++;			/* count outstanding requests */
  }
  for( ; incount < MAX_ISSUE; incount++ ) {
    tile = mostNeeded();
    if( tile == NULL )
      break;
    tile->waiting = 1;
    parseURL(tile->wrap->name, &u);
    sprintf(request, "GET %08X %s", tileID(u.path), u.path);
    if( send(udp, request, strlen(request), 0) < 0 )
      pfatal("issue");
    incount++;
  }
  UNLOCK(wrapLock);
}



/*
 *  timeout(dummy) -- asynchronous timeout handler.
 *
 *  This function runs asynchronously in its own thread and checks unanswered
 *  requests once a second.  The "waiting" field of the tile structure is
 *  incremented, and if it has exceeded TV_TIMEOUT seconds, an EOF is simulated.
 */
static void *timeout(void *dummy) {
  int eofcount;
  Tile *tile;
  Wrap *wrap;
  WrapGen *wg;
  pthread_t pth;

  for( ;; ) {
    sleep(1);
    eofcount = 0;
    LOCK(wrapLock);
    wg = genWraps(TYPE_WANTED | TYPE_TILE);
    while( (wrap = nextWrap(wg)) != NULL ) {
      tile = wrap->data;
      if( tile->waiting <= 0 )
        continue;
      if( ++tile->waiting > TV_TIMEOUT ) {
        fprintf(stderr, "TIMEOUT %s\n", wrap->name);
        eofcount++;
        tile->waiting = -1;		/* give up on this tile */
        /* interpret and link data in a separate thread */
        pthread_create(&pth, NULL, newTile, tile);
        pthread_detach(pth);
      }
    }
    UNLOCK(wrapLock);
    if( eofcount > 0 ) {
      issue();	 /* check to see if it's time to issue another request */
    }
  }
}



/*
 *  listener(dummy) -- asynchronous UDP port listener.
 *
 *  This function runs asynchronously in its own thread and reads TVTP
 *  datagrams from the UDP port.  When it receives an EOF message, it
 *  spawns a transient thread to register the collected data.
 */

static void *listener(void *dummy) {
  int id, k, n, seq, flag, xxx, yyy, zone;
  Tile *tile;
  Wrap *wrap;
  pthread_t pth;
  char name[MAX_TILE_NAME];
  unsigned char buf[TV_MAXLEN];

  for( ;; ) {

    /* read next datagram */
    n = recv(udp, buf, TV_MAXLEN, 0);
    if( n < 0 )
      pfatal("tvtp recv");
    if( n < TV_PHLEN || memcmp(buf, "TVTP\02", 5) != 0 )
      fatal("tvtp: corrupt message");
    seq = buf[6];
    if( seq == TV_EOF )
      flag = -1;
    else
      flag = buf[TV_PHLEN + 4] & ~TV_GZIP;
    if( flag > TV_NFLAGS )
      fatal("tvtp: invalid flag value");

    /* identify the tile */
    id = unpack4(buf + 8);
    k = TIDORDER(id);
    zone = TIDZONE(id);
    xxx = TIDXXX(id);
    yyy = TIDYYY(id);
    synthPath(name, zone, xxx, yyy, k);

    /* find wrap entry, or make one if never before seen */
    LOCK(wrapLock);
    wrap = findWrap(name, TYPE_WANTED | TYPE_TILE | TYPE_UNAV);
    if( wrap == NULL || wrap->type == TYPE_UNAV ) {
      if( seq == TV_EOF ) {
        UNLOCK(wrapLock);
        continue;			/* ignore EOF if no Wrap for it */
      }
      wrap = wantTile(name);		/* make new wrap list entry */
    }
    tile = wrap->data;
    tile->waiting = 1;			/* reset timeout counter */

    /* handle EOF message */
    if( seq == TV_EOF ) {		/* if EOF message */
      tile->waiting = -1;		/* mark as complete */
      UNLOCK(wrapLock);

      /* interpret and link data in a separate thread */
      pthread_create(&pth, NULL, newTile, tile);
      pthread_detach(pth);

      /* check to see if it's time to issue another request */
      issue();
      continue;
    }

    /* save the data */
    addRec(wrap->data, buf + TV_PHLEN, n - TV_PHLEN);
    UNLOCK(wrapLock);
    postRedisplays(RO);			/* update status discs */
  }
}



/*
 *  Adds a new tile and rebuilds the relief map.
 *  Parameters are compatible with pthread_create.
 */
void *newTile(void *arg) {

  Tile *tile = arg;
  Wrap *wrap = tile->wrap;
  LOCK(wrapLock);
  if( buildTile(tile) ) {
    postRedisplays(NR + NS + RO + RP);
  } else {
    if( tile->dlevels == 0 )
      wrap->type = TYPE_UNAV;
    postRedisplays(RO);
  }
  tile->waiting = 0;
  UNLOCK(wrapLock);
  return NULL;
}



#endif  /* UNITHREAD */
