package topovista.explorer;

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;

import topovista.grid.*;
import topovista.toksvc.*;
import topovista.util.*;



/**
 *  Survey is the main class of the TopoVista Explorer application.
 */
public class Survey {



private static String DEFAULT_DIR = "../../data/supai";
public static String TVG_EXTENSION = ".tvg";
public static String IMG_PREFIX = "topovista/explorer";


static String classname = "Survey";		// name of this class
static String command = "java " + classname;	// usage message prefix

static String dir = DEFAULT_DIR;	// data directory
static String start = null;		// start location

static String tsconfig;			// token service config, if any
static Client tsclient;			// corresponding toksvc.Client

static String uprefix = null;		// URL prefix to use in tokens
static int litmax = 0;			// max size file to register as literal


static int interval;			// time interval between steps

static JFrame frame;			// top level application frame
static Plat plat;			// plat map panel


static Explorer[] algs = {		// explorer algorithms, in menu order
    new BoundedWalk(),
    new ChaoticHop(),
    new EdgeGrowth(),
    new RandomWalk(),
    new SpiralWalk(),
    new WeightedRandom(),
};

static char pathtype;			// path algorithm selection character
static Explorer ex;			// currently chosen explorer


static World curWorld;			// current world
static Cell curCell;			// current location within that world



static java.util.Timer timer = new java.util.Timer(true);	// master timer
static TimerTask task;						// current task

class AdvanceTask extends TimerTask {				// task code
    public void run() {
	Cell locn = ex.nextLoc(curWorld, curCell);
	if (locn != null) {
	    visit(locn);
	}
    }
};



/**
 *  Main program.  Initializes the system, then hands control
 *  to the chosen explorer algorithm.
 */
public static void main(String[] args) throws Exception {

    ckargs(args);
    window(dir);
    if (tsconfig != null) {
	tsconnect(tsconfig);
    }

    restart();
}



/**
 *  Reinitializes a New World.
 */
static void restart() throws IOException {

    if (tsconfig != null && curWorld != null) {
	Cell[] all = curWorld.all();
	if (all.length > 0) {
	    for (int i = 0; i < all.length; i++) {
		Cell ce = all[i];
		if (ce.visited && ce.hasdata) {
		    Token tk = tsclient.request(
			ce.name, Token.EXCLUSIVE, null, null);
		    tk.release("");
		}
	    }
	}
    }

    curWorld = new World(new File(dir));
    curCell = (start == null) ? curWorld.choose() : curWorld.find(start);
    plat.configure(curWorld);
    plat.repaint();

    visit(curCell);
    ex.init();
    reschedule();
}



/**
 *  Schedules the advancement task.
 */

static void reschedule() {
    if (task != null) {
	task.cancel();
    }
    task = new TimerTask() {
	public void run() {
	    Cell locn = ex.nextLoc(curWorld, curCell);
	    if (locn != null) {
		visit(locn);
	    }
	}
    };
    timer.schedule(task, interval, interval);
}



/**
 *  Records a visit to a cell.
 */
static void visit(Cell ce) {

    curCell = ce;

    // if new discovery, register cell contents or path with token service
    if (!ce.visited && ce.hasdata && tsclient != null) {
	register(ce);
    }

    ce.visited = true;			// mark as visited
    plat.setCurrent(ce);		// circle on plat map
    plat.repaint();			// repaint map window
}



/**
 *  Registers a cell with the token service.
 */
static void register(Cell ce) {

    Token tk = tsclient.request(ce.name, Token.EXCLUSIVE, null, null);

    if (litmax > 0) try {
	File f = new File(dir + '/' + ce.name + TVG_EXTENSION);
	long n = f.length();
	if (n > 0 && n <= litmax) {
	    tk.release(loadlit(f));
	    return;
	}
    } catch (Throwable t) {
    }

    if (uprefix != null) {
	tk.release(uprefix + '/' + ce.name);
    } else {
	tk.release(dir + '/' + ce.name);
    }
}



/**
 *  Loads a file to make a literal value.
 */
public static byte[] loadlit(File f) throws IOException {
    int n;
    final int BUFSIZE = 8192;
    byte[] buf = new byte[BUFSIZE];
    InputStream d = new BufferedInputStream(new FileInputStream(f));
    ByteArrayOutputStream bs = new ByteArrayOutputStream(BUFSIZE);
    bs.write("literal:".getBytes());
    while ((n = d.read(buf)) >= 0) {
	bs.write(buf, 0, n);
    }
    d.close();
    return bs.toByteArray();
}



/**
 *  Cracks the command arguments.
 */
static void ckargs(String [] args) throws Exception {

    String s;
    Options o = new Options(new String[] {
	"-p c w",	"path type",
	"-i n 200",	"interval, in milliseconds",
	"-d dir",	"directory of data files",
	"-t cfg",	"token service configuration",
	"-u url",	"URL prefix to use in token values",
	"-l size 0",	"register literal if no larger than this",
	"[start]",	"initial cell designator",
    }, command, args);

    s = o.getValue('p');
    pathtype = (s.length() > 0) ? s.charAt(0) : '\0';
    for (int i = 0; i < algs.length; i++) {	// find selected explorer
	if (pathtype == algs[i].ch) {
	    ex = algs[i];
	    break;
	}
    }
    if (ex == null) {
	System.err.println("unrecognized -p value");
	System.exit(1);
    }

    tsconfig = o.getValue('t');
    uprefix = o.getValue('u');
    litmax = Integer.parseInt(o.getValue('l'));

    s = o.getValue('d');
    if (s != null) {
	dir = s;
    }

    interval = Integer.parseInt(o.getValue('i'));
    if (interval <= 0) {
	interval = 1;
    }

    int i = o.getIndex();
    if (i < args.length) {
	start = args[i++];
    }
}



/**
 *  Connects to the token service, displaying a progress box if delayed.
 *  (The progress box advances slowly to 90%, then stalls, because we
 *  can't really measure how close to completion we are.)
 */
static void tsconnect(String cfgfile) {

    final ProgressMonitor p = new ProgressMonitor(frame,
	"Connecting to token server", "", 0, 100);

    java.util.Timer t = new java.util.Timer();
    t.schedule(new TimerTask() {
	int i = 0;
	public void run() {
	    if (p.isCanceled()) {
		System.exit(0);
	    }
	    if (++i < 90) {
		p.setProgress(i);
	    }
	}
    }, 200, 200);

    try {
	tsclient = new Client(new Configuration(tsconfig));
    } catch (IOException e) {
	System.err.println(e);
	System.exit(1);
    }

    t.cancel();
    p.close();
}



/**
 *  Creates a standard 28x28 JButton using a 24x24 image from our Jar file,
 *  and registers a key mnemonic for the button.
 */
static JButton imageButton(String filename, String description, int mnemonic) {
    URL u = ClassLoader.getSystemResource(IMG_PREFIX + '/' + filename);
    Icon i = new ImageIcon(u, description);
    JButton b = new JButton(i);
    b.setPreferredSize(new Dimension(28, 28));	// why is this needed??
    b.setMnemonic(mnemonic);
    return b;
}



/**
 *  Initializes the graphics window and returns the Plat it contains.
 */
static Plat window(String label) {

    final int BDR = 8;				// common border size

    String[] elabels = new String[algs.length];
    for (int i = 0; i < algs.length; i++) {
	elabels[i] = algs[i].label;
    }
    final JComboBox abox = new JComboBox(elabels);
    abox.setSelectedItem(ex.label);
    abox.addItemListener(new ItemListener() {
	public void itemStateChanged(ItemEvent e) {
	    ex = algs[abox.getSelectedIndex()];
	    ex.init();
	}
    });

    JButton cbutton = imageButton("Center24.gif", "Recenter", KeyEvent.VK_C);
    cbutton.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	    plat.configure(curWorld);
	}
    });

    JButton rbutton = imageButton("Rewind24.gif", "Rewind", KeyEvent.VK_R);
    rbutton.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	    try {
		restart();
	    } catch (IOException x) {
		x.printStackTrace();
		System.exit(0);
	    }
	}
    });

    JButton qbutton = imageButton("Zap24.gif", "Quit", KeyEvent.VK_Q);
    qbutton.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	    System.exit(0);
	}
    });

    JPanel header = new JPanel();
    header.setLayout(new BoxLayout(header, BoxLayout.X_AXIS));
    header.add(cbutton);
    header.add(Box.createRigidArea(new Dimension(BDR, 0)));
    header.add(abox);
    header.add(Box.createRigidArea(new Dimension(BDR, 0)));
    header.add(rbutton);
    header.add(Box.createHorizontalGlue());
    header.add(Box.createRigidArea(new Dimension(BDR, 0)));
    header.add(qbutton);

    // speed slider is logarithmic: interval = 10 ** (sliderval / 100) msec
    int sval = 600 - 100 * (int)(Math.log(interval) / Math.log(10) + 0.5);
    JSlider slider = new JSlider(JSlider.VERTICAL, 0, 600, sval);
    slider.setMajorTickSpacing(100);
    slider.setPaintTicks(true);
    slider.addChangeListener(new ChangeListener() {
	public void stateChanged(ChangeEvent e) {
	    JSlider s = (JSlider) e.getSource();
	    interval = (int) Math.pow(10.0, 6 - 0.01 * s.getValue());
	    reschedule();
	};
    });

    plat = new Plat();
    plat.setPreferredSize(new Dimension(360, 360));
    plat.setBackground(Color.white);

    JPanel center = new JPanel();
    center.setLayout(new BoxLayout(center, BoxLayout.X_AXIS));
    center.add(slider);
    center.add(Box.createRigidArea(new Dimension(BDR, 0)));
    center.add(plat);

    JPanel footer = new JPanel();
    footer.setLayout(new BoxLayout(footer, BoxLayout.X_AXIS));
    footer.add(legend(Plat.HASDATA, "unexplored"));
    footer.add(Box.createRigidArea(new Dimension(BDR, 0)));
    footer.add(legend(Plat.EXPLORED, "explored"));
    footer.add(Box.createHorizontalGlue());
    footer.add(Box.createRigidArea(new Dimension(BDR, 0)));
    footer.add(legend(Plat.VISITED, "visited"));
    footer.add(Box.createRigidArea(new Dimension(BDR, 0)));
    footer.add(legend(Plat.SCOUTED, "scouted"));

    JPanel pane = new JPanel();
    pane.setBorder(BorderFactory.createEmptyBorder(BDR, BDR, BDR, BDR));
    pane.setLayout(new BoxLayout(pane, BoxLayout.Y_AXIS));
    pane.add(header);
    pane.add(Box.createRigidArea(new Dimension(0, BDR)));
    pane.add(center);
    pane.add(Box.createRigidArea(new Dimension(0, BDR)));
    pane.add(footer);

    frame = new JFrame(label);
    frame.getContentPane().add(pane, BorderLayout.CENTER);
    frame.addWindowListener(new WindowAdapter() {
	public void windowClosing(WindowEvent e) {
	    System.exit(0);
	}
    });

    frame.pack();
    frame.setVisible(true);

    return plat;
}



/**
 *  Constructs a legend for one color.
 */
static JLabel legend(final Color co, String text) {

    Icon ic = new Icon() {
	final static int SIZE = 8;
	public int getIconHeight() { return SIZE; }
	public int getIconWidth()  { return SIZE; }
	public void paintIcon(Component c, Graphics g, int x, int y) {
	    g.setColor(co);
	    g.fillRect(x, y, SIZE, SIZE);
	}
    };

    JLabel lb = new JLabel(text, ic, SwingConstants.RIGHT);
    lb.setBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2));
    lb.setBackground(Color.white);
    lb.setOpaque(true);
    return lb;
}



} // class Survey
