package topovista.grid;

import java.io.*;
import java.text.*;
import java.util.*;
import java.util.zip.*;

import topovista.util.*;



/**
 *  A Grid object represents a regular, rectangular grid of elevation data.
 *
 *  Most of the useful constructors are for specific subclasses.
 *  Constructors in this superclass are for combining existing Grids.
 *
 *  @author Gregg Townsend, University of Arizona
 *  @author originally based on dem2pgm.c from TopoVista 3.0
 *  @version August, 2001
 */



public class Grid {



public final static int RADIANS = 0;
public final static int FEET   = 1;
public final static int METERS = 2;
public final static int ARCSEC = 3;

public final static int NAD27 = 1;
public final static int WGS72 = 2;
public final static int WGS84 = 3;
public final static int NAD83 = 4;
public final static int OLD_HAWAII = 5;
public final static int PUERTO_RICO = 6;

public final static int GEO_SYSTEM = 0;
public final static int UTM_SYSTEM = 1;

public final static int UTM_FALSE_EASTING = 500000;



    String filename;		// input file name
    PushbackReader inputFile;	// input file handle

    int treelevels;		// from TopoVista input grid

    // public data fields
    public int cols, rows;

    // information from "A" record
    public String header;
    public String freeform;
    public int refsys;
    public int zone;
    public int datum;		// one of the above datums (default=NAD27)
    public int xyUnit;
    public int zUnit;

    // quadrangle corner locations are not saved, only x/y/z extremes
    public double xRes, xMin, xMax;
    public double yRes, yMin, yMax;
    public double zRes, zMin, zMax;

    public double xLL, yLL;	// coords of lower left (southwest) data cell

    // Elevations are biased by 1000 and stored as type char (unsigned short).
    // This accommodates Death Valley to Denali in either feet or decimeters.
    // A stored value of zero indicates a missing value.
    public static int ELEVATION_BIAS = 1000;
    public static char MISSING_ELEV = 0;
    public char[][] elev;



/**
 *  Loads and returns a grid file, aborting if an error occurs.
 */
public static Grid checkLoad(String filename, boolean partial) {
    String ename = (filename == null) ? "stdin" : filename;
    try {
	return load(filename, partial);
    } catch (NumberFormatException e) {
	System.err.println("\n" + ename + ": unacceptable grid file");
    } catch (IOException e) {
	System.err.println("\n" + ename + ": " + e.toString());
    } catch (Error e) {
	System.err.println("\n" + ename + ": " + e.toString());
    } catch (Throwable t) {
	System.err.println("\n" + ename + ": ");
	t.printStackTrace();
    }
    System.exit(1);
    return null;	// not reached
}



/**
 *  Loads and returns a partial or full grid file, throwing an I/O
 *  exception in case of problems.
 *
 *  A partial grid contains only the header ("A record") information.
 *  If the filename is null or "-", standard input is read.
 */
public static Grid load(String filename, boolean partial) throws IOException {

    Grid g;

    // open input file
    InputStream ins;
    if (filename == null || filename.equals("-")) {
	filename = "[stdin]";
	ins = System.in;
    } else {
	ins = new FileInputStream(filename);
    }

    ins = new BufferedInputStream(new GZInputStream(ins));
    ins.mark(2);
    int b1 = ins.read();
    int b2 = ins.read();
    ins.reset();

    if (b1 < 0) {
	throw new EOFException(filename);
    } else if (b1 == 0xE4 && b2 == 'B') {
	g = new BinaryGrid(new TypedInputStream(ins), partial);
    } else if (b1 == 'T' && b2 == 'V') {
	g = new TileGrid(ins, partial);
    } else {
	g = new NativeGrid(ins, partial);
    }

    ins.close();
    g.filename = filename;
    return g;
}



/**
 *  Ensures that a particular row of the grid is available for access.
 *  (Always implemented, but needed only with LazyGrid objects.)
 */
public void loadRow(int y) {
    char[] e = elev[y];		// validate y value, do nothing else
}



/**
 *  Does nothing here.
 *  (In LazyGrid, frees the row.)
 */
public void discardRow(int y) {
    char[] e = elev[y];		// validate y value, do nothing else
}



/**
 *  Writes human-readable statistics about this Grid object.
 */
public void writeStats(PrintStream p) {
    p.println();
    p.println(filename + ":");
    p.println("   " + header);
    if (freeform.length() > 0) {
	p.println("   " + freeform);
    }
    p.print("   grid size "
	+ cols + " columns by " + rows + " rows");
    if (treelevels > 0) {
	int n = 1;
	while ((1 << n) < cols) {
	    n++;
	}
	p.print(", error tree " + treelevels + "/" + (2 * n) + " deep");
    }
    p.println();
    p.println("   " +
	(refsys == UTM_SYSTEM ? ("UTM zone " + zone) : "geographic") +
	" coordinates, " + datumName(datum) + " datum, resolution " +
	xRes + " by " + yRes + units(xyUnit));
    p.println("   XLL = " + xLL + ";  W-E range " + f(xMin) + " to " + f(xMax));
    p.println("   YLL = " + yLL + ";  S-N range " + f(yMin) + " to " + f(yMax));
    p.println("   elevation range " +
	f(zMin) + " to " + f(zMax) + units(zUnit) +
	", resolution " + zRes + units(zUnit));
}

private static DecimalFormat df = new DecimalFormat("#0.0");

private static String f(double v) {
    return df.format(v);
}

public static String datumName(int i) {
    switch (i) {
	case NAD27:		return "NAD27";
	case WGS72:		return "WGS72";
	case WGS84:		return "WGS84";
	case NAD83:		return "NAD83";
	case OLD_HAWAII:	return "OldHawaii";
	case PUERTO_RICO:	return "PuertoRico";
	default:		return "unknown";
    }
}

public static String units(int i) {
    switch(i) {
	case RADIANS:	return " radians";
	case FEET:	return " feet";
	case METERS:	return " meters";
	case ARCSEC:	return " arc-seconds";
	default:	return " units";
    }
}



/**
 *  Constructs an empty Grid object containing no data.
 */
public Grid() {}



/**
 *  Constructs a Grid by overlaying an existing Grid on
 *  a larger data array.
 */
public Grid(Grid g, double xLL, double yLL, double xLim, double yLim) {
    initFrom(g, g);
    this.xLL = xLL;
    this.yLL = yLL;
    cols = (int) ((xLim - xLL) / g.xRes);
    rows = (int) ((yLim - yLL) / g.yRes);
    alloc();
    overlay(g);
}



/**
 *  Constructs a new Grid object by merging two existing Grid objects.
 *  The objects must be compatible as indicated in the code
 *  (same UTM zone, units, resolution, etc.).
 */
public Grid(Grid g1, Grid g2) {
    initFrom(g1, g2);
    alloc();
    overlay(g1);
    overlay(g2);
}



/**
 *  Initializes header information in this Grid by copying from
 *  two mutually compatible Grids.  The two arguments may
 *  be the same to initialize from a single Grid.
 *  One or both arguments may be this Grid.
 *
 *  @return	vertical offset of g2 with respect to g1
 */
int initFrom(Grid g1, Grid g2) {

    String s = g1.incompatible(g2);
    if (s != null) {
	throw new Error("incompatible grids: " + s);
    }

    int dx = (int) ((g2.xLL - g1.xLL) / xRes);
    int dy = (int) ((g2.yLL - g1.yLL) / yRes);

    header = "Merged Grid";
    freeform = "";
    refsys = g1.refsys;
    zone = g1.zone;
    datum = g1.datum;
    xyUnit = g1.xyUnit;
    zUnit = g1.zUnit;
    xRes = g1.xRes;
    yRes = g1.yRes;
    zRes = g1.zRes;
    xMin = Math.min(g1.xMin, g2.xMin);
    yMin = Math.min(g1.yMin, g2.yMin);
    xMax = Math.max(g1.xMax, g2.xMax);
    yMax = Math.max(g1.yMax, g2.yMax);
    zMin = Math.min(g1.zMin, g2.zMin);
    zMax = Math.max(g1.zMax, g2.zMax);
    xLL = Math.min(g1.xLL, g2.xLL);
    yLL = Math.min(g1.yLL, g2.yLL);

    if (dx > 0) {
	cols = Math.max(g1.cols, g2.cols + dx);
    } else {
	cols = Math.max(g2.cols, g1.cols - dx);
    }

    if (dy > 0) {
	rows = Math.max(g1.rows, g2.rows + dy);
    } else {
	rows = Math.max(g2.rows, g1.rows - dy);
    }

    return dy;
}



/**
 *  Checks whether another Grid is incompatible with this one
 *  for purposes of overlaying data.
 *  Returns the description of the first incompatible field detected,
 *  or null if compatible.
 */
public String incompatible(Grid g) {
    if (refsys != g.refsys)	return "reference system";
    if (zone != g.zone)		return "UTM zone";
    if (datum != g.datum)	return "datum";
    if (xyUnit != g.xyUnit)	return "ground unit";
    if (xRes != g.xRes)		return "X resolution";
    if (yRes != g.yRes)		return "Y resolution";
    return null;
}



/**
 *  Overlays data from grid g onto this grid.
 *  It is assumed that this grid is compatible and configured.
 *  Missing individual rows are allocated if not yet present.
 *
 *  <P> If the two grids have differing zRes or zUnit values,
 *  both are rescaled to decimeters before overlaying.
 */
public void overlay(Grid g) {

    if (zRes != g.zRes || zUnit != g.zUnit) {
	this.scaleTo(0.1, METERS);
	g.scaleTo(0.1, METERS);
    }

    xMin = Math.min(xMin, g.xMin);
    yMin = Math.min(yMin, g.yMin);
    xMax = Math.max(xMax, g.xMax);
    yMax = Math.max(yMax, g.yMax);
    zMin = Math.min(zMin, g.zMin);
    zMax = Math.max(zMax, g.zMax);
    int x = (int) ((g.xLL - xLL) / xRes);
    int y = (int) ((g.yLL - yLL) / yRes);

    for (int i = 0; i < g.rows; i++) {
	if (elev[y + i] == null) {
	    alcRow(y + i);
	}
	for (int j = 0; j < g.cols; j++) {
	    int o = this.elev[y + i][x + j];
	    if (o == MISSING_ELEV) {
		this.elev[y + i][x + j] = g.elev[i][j];
	    }
	}
    }
}



/**
 *  Rescales this grid to have the specified zRes and zUnit values.
 */
public void scaleTo(double res, int units) {

    double f = this.zRes / res;
    switch(units - this.zUnit) {
	case 0:					break;
	case FEET - METERS:	f /= .3048;	break;
	case METERS - FEET:	f *= .3048;	break;
	default:	throw new Error("can't convert these vertical units");
    }
    if (f == 1.0) {
	return;
    }
    for (int i = 0; i < rows; i++) {
	for (int j = 0; j < cols; j++) {
	    int o = elev[i][j];
	    if (o != MISSING_ELEV) {
		o = (int) (f * (o - ELEVATION_BIAS) + 0.5);
		elev[i][j] = (char) (o + ELEVATION_BIAS);
	    }
	}
    }

    zMin *= f * res / zRes;
    zMax *= f * res / zRes;
    zRes = res;
    zUnit = units;
}



/**
 *  Allocates arrays, given that rows and cols have been set.
 */
void alloc() {
    elev = new char[rows][];
    for (int i = 0; i < rows; i++) {
	alcRow(i);
    }
}



/**
 *  Allocates row n, if not already present.
 */
void alcRow(int n) {
    elev[n] = new char[cols];
    if (MISSING_ELEV != 0) {
	for (int i = 0; i < cols; i++) {
	    elev[n][i] = MISSING_ELEV;
	}
    }
}



/**
 *  Returns a String representation of this Grid object.
 */
public String toString() {
    String s = this.getClass().getName();
    s = s.substring(s.lastIndexOf('.') + 1);
    return s + "(" + filename + "," + cols + "x" + rows + ")";
}



} // class Grid
