package topovista.grid;

import java.util.*;



/**
 *  A LazyGrid is a Grid object that is more frugal of memory,
 *  but requires more discipline in its use.
 *  A LazyGrid is built by constructing an empty LazyGrid
 *  and then adding minimal Grid objects atop it.
 *  loadRow(x) must be called before accessing row x of the grid;
 *  discardRow(x) can be called to allow freeing of memory.
 */

class LazyGrid extends Grid {

    // input files needed for each row, indexed by row number
    private BitSet[] needed = new BitSet[0];

    // input data currently loaded in each row, indexed by row number
    private BitSet[] loaded = new BitSet[0];

    // input file names, indexed by bit number
    private ArrayList filenames = new ArrayList();

    // trace output to stderr, if any
    private String trace = null;



/**
 *  Constructs an empty LazyGrid.
 */
public LazyGrid() {
}



/**
 *  Sets the trace message written to stderr when a file is loaded.
 */
public void setTrace(String t) {
    trace = t;
}



/**
 *  Adds the specified partial grid (lazily).
 *  If the grid has no filename, it supplies only boundaries, not data.
 */
public void add(Grid g) {

    int dy;
    if (header == null) {
	dy = initFrom(g, g);
    } else {
	String s = this.incompatible(g);
	if (s != null) {
	    throw new Error("incompatible grids: " + s);
	}
	dy = initFrom(this, g);
	if (zRes != g.zRes || zUnit != g.zUnit) {
	    // set to scale all input to decimeters
	    this.zRes = 0.1;
	    this.zUnit = Grid.METERS;
	}
    }
    needed = aadjust(needed, dy);
    loaded = aadjust(loaded, dy);

    if (g.filename != null) {
	int bitn = filenames.size();
	filenames.add(g.filename);
	if (dy < 0) {
	    dy = 0;
	}
	for (int i = 0; i < g.rows; i++) {
	    needed[dy+i].set(bitn);
	}
    }
}



/**
 *  Returns a BitSet array adjusted by dy.
 */
private BitSet[] aadjust(BitSet[] a, int dy) {

    BitSet[] bs = new BitSet[rows];
    if (dy >= 0) {
	System.arraycopy(a, 0, bs, 0, a.length);
    } else {
	System.arraycopy(a, 0, bs, -dy, a.length);
    }
    for (int i = 0; i < bs.length && bs[i] == null; i++) {
	bs[i] = new BitSet();
    }
    for (int i = bs.length - 1; i >= 0 && bs[i] == null; i--) {
	bs[i] = new BitSet();
    }
    return bs;
}



/**
 *  Loads data as necessary for use of the specified row.
 */
public void loadRow(int y) {
    if (elev == null) {
	elev = new char[rows][];
    }
    if (elev[y] == null) {
	alcRow(y);
    }
    BitSet have = loaded[y];
    BitSet need = needed[y];
    if (! need.equals(have)) {
	for (int i = 0; i < need.length(); i++) {
	    if (need.get(i) && ! have.get(i)) {
		loadFile(i);
	    }
	}
    }
}



/**
 *  Frees the specified row.
 */
public void discardRow(int y) {
    elev[y] = null;
}



/**
 *  Loads file n.
 */
private void loadFile(int n) {
    String fname = (String) filenames.get(n);
    Grid g = Grid.checkLoad(fname, false);
    int dy = initFrom(this, g);
    this.overlay(g);
    for (int i = 0; i < g.rows; i++) {
	loaded[i+dy].set(n);
    }
    if (trace != null) {
	System.err.print(trace);
	System.err.flush();
    }
}



} // class LazyGrid
