package topovista.grid;

import java.io.*;

import topovista.util.*;



/**
 *  The Merge class contains the all-purpose command-level utility
 *  program for inspecting, merging, and transforming grid files.
 *
 *  @author Gregg Townsend, University of Arizona
 *  @version June, 2002
 */



public class Merge {



private final static String COMMAND_NAME = "java egrid.Merge";

private final static int DEFAULT_PRUNE = 3;	// default error tree pruning



/**
 *  Reads, inspects, combines, and reformats elevation grid files.
 *
 *  <P><PRE>
 *  Usage: java Merge [-{prcbt}] [filename ...]
 *
 *	-p	produce raw data in PGM format
 *	-r	produce relief map in PGM format
 *	-c	produce contour map in PGM format
 *	-b	produce binary (gzipped TypedOutputStream) output
 *	-t cnu	produce TopoVista grids of order c, pruning n error levels
 *		(default n=3), unzipped if suffixed with 'u'
 *
 *  With no option character, writes statistics about the input file(s).
 *
 *  Output is normally to standard output.  With the -t option, one
 *  or more input files must be a named explicitly, and output is to
 *  multiple files named "xxxUyyyc.tvg" in the current directory.
 *  </PRE>
 */

public static void main(String args[]) throws IOException {

    // parse command options
    Options o = new Options(new String[] {
	"-p", "produce raw data in PGM format",
	"-r", "produce relief map in PGM format",
	"-c", "produce contour map in PGM format",
	"-b", "produce binary (gzipped TypedOutputStream) output",
	"-t c[n][u]", "write TopoVista grid files of order c, pruning n",
	"[file...]",  "input files",
    }, COMMAND_NAME, args);

    // check output format specs; can only have one
    char outformat = 0;
    String s = o.getWhich();
    for (int i = 0; i < s.length(); i++) {
	char c = s.charAt(i);
	if ("prcbt".indexOf(c) >= 0) {
	    if (outformat == 0) {
		outformat = c;
	    } else {
		o.setHeader("only one output format can be specified");
		o.usage(System.err, COMMAND_NAME);
		System.exit(1);
	    }
	}
    }

    // if no output format given, produce statistics
    if (outformat == 0) {
	if (args.length == 0) {
	    Grid.checkLoad(null, true).writeStats(System.out);
	} else {
	    for (int i = 0; i < args.length; i++) {
		Grid.checkLoad(args[i], true).writeStats(System.out);
	    }
	}
	return;
    }

    // crack TileGrid option string
    TVGoptions tvgo = getTVGopts(o);

    // buffer standard output!
    System.setOut(new PrintStream(new BufferedOutputStream(System.out)));

    // read and combine grid files
    Grid g;
    int first = o.getIndex();
    int n = args.length - first;
    if (n == 0) {
	if (outformat == 't') {
	    System.err.println("input file name required with -t");
	    System.exit(1);
	}
	g = Grid.checkLoad(null, false);
    } else if (n == 1 && outformat != 't') {
	g = Grid.checkLoad(args[first], false);
    } else {
	g = loadMulti(args, first, tvgo);
	if (outformat != 't') {
	    System.err.print("   loading  ");
	}
    }

    // write output in selected format
    switch (outformat) {
	case 'p':  PGMGrid.writePGM(System.out, g);	break;
	case 'r':  PGMGrid.writeRelief(System.out, g);	break;
	case 'c':  PGMGrid.writeContour(System.out, g);	break;
	case 'b':  BinaryGrid.write(g);			break;
	case 't':
	    TileGrid.write(g, tvgo.order, tvgo.nprune, tvgo.unzip);
	    break;
    }

    System.err.println();
    System.out.flush();		// empirically found to be necessary
				// (IBM Linux JRE 1.3.0-20001124)
}



//  loadMulti() -- load and merge multiple grid files using a LazyGrid

private static Grid loadMulti(String[] args, int first, TVGoptions tvgo) {

    System.err.print("   scanning ");
    LazyGrid g = new LazyGrid();

    for (int i = first; i < args.length; i++) {
	Grid h = Grid.checkLoad(args[i], true);
	try {
	    g.add(h);
	} catch (Error e) {
	    System.err.println("\n" + e);
	    System.err.println("   " + g);
	    System.err.println("   " + h);
	    System.exit(1);
	}
	System.err.print("*");  System.err.flush();
    }
    System.err.println();  System.err.flush();
    g.setTrace("*");

    if (tvgo != null) {
	tvgExpand(g, tvgo.order);
    }
    return g;
}



//  tvgExpand() -- expand Grid to TopoVista tile boundaries

private static void tvgExpand(LazyGrid g, int k) {

    int res = (int) g.xRes;
    int edge = (1 << k) * res;
    int xmod = TileGrid.X_ORIGIN % edge;

    double xLL = g.xLL;
    double yLL = g.yLL;
    double xLim = xLL + g.cols * res;
    double yLim = yLL + g.rows * res;

    xLL -= (xLL - xmod) % edge;
    yLL -= yLL % edge;
    xLim += edge - 2 * res;
    yLim += edge - 2 * res;
    xLim += res - (xLim - xmod) % edge;
    yLim += res - yLim % edge;

    Grid h = new Grid();
    h.initFrom(g, g);
    h.filename = null;
    h.xLL = xLL;
    h.yLL = yLL;
    h.cols = (int) ((xLim - xLL) / g.xRes);
    h.rows = (int) ((yLim - yLL) / g.yRes);
    g.add(h);
}



//  getTVGopts(opts) -- get TVGoptions object, or null if no TVG output

private static class TVGoptions {	// TileGrid output options
    int order;				// grid order
    int nprune;				// number of levels to prune
    boolean unzip;			// unzip flag
};

private static TVGoptions getTVGopts(Options o) {

    // get option value; return null if not used
    String arg = o.getValue('t');
    if (arg == null) {
	return null;
    }

    // get grid order character, the value of the -t argument
    char och = 0;
    if (arg.length() >= 1) {
	och = arg.charAt(0);
    }

    // get unzip flag, if present
    boolean unzip = arg.length() >= 2 && arg.endsWith("u");
    if (unzip) {
	arg = arg.substring(0, arg.length() - 1);
    }

    // get pruning value, if given
    int nprune = DEFAULT_PRUNE;
    if (arg.length() > 1) {
	try {
	    nprune = Integer.parseInt(arg.substring(1));
	} catch (Exception e) {
	    nprune = -1;
	}
    }

    // diagnose errors
    if (och < 'a' || och > 'z' || nprune < 0) {
	o.setHeader("bad argument to -t");
	o.usage(System.err, COMMAND_NAME);
	System.exit(1);
    }

    TVGoptions tvgo = new TVGoptions();
    tvgo.order = och - 'a' + 1;
    tvgo.nprune = nprune;
    tvgo.unzip = unzip;
    return tvgo;
}



} // class Merge
