package topovista.grid;

import java.io.*;
import java.util.zip.*;

import topovista.util.*;



/**
 *  The NativeGrid class reads USGS "Native" mode DEM files
 *  (an older format, ASCII blocked 1024, now written by sdts2dem).
 */

class NativeGrid extends Grid {



private static int MISSING_VALUE = -9999;	// sdts2dem missing-value flag



/**
 *  Constructs a NativeGrid by reading from a file.
 *  If the file format is not correct, a NumberFormatException is likely.
 */
NativeGrid(InputStream ins, boolean partial) throws IOException {

    Reader r = new InputStreamReader(new GZInputStream(ins));
    inputFile = new PushbackReader(new BufferedReader(r));

    // A records

    header = readString(144);			// A1, A2
    int s2d = header.indexOf("(SDTS2DEM");	// ck for new SDTS2DEM comment
    if (s2d <= 0) {
	s2d = header.indexOf("SDTS2DEM");	// or for older version
    }
    if (s2d > 0) {
	freeform = header.substring(s2d).trim();
	header = header.substring(0, s2d).trim();
    } else {
	freeform = header.substring(40, 80).trim();
	header = header.substring(0, 40).trim();
    }

    readInteger(6);				// A3  (level)
    readInteger(6);				// A4  (elevation pattern)
    refsys = readInteger(6);			// A5
    zone = readInteger(6);			// A6

    for (int i = 0; i < 15; i++) {
	readReal(24);				// A7: projection parameters
    }

    xyUnit = readInteger(6);			// A8
    zUnit = readInteger(6);			// A9
    readInteger(6);				// A10  (polygon sides)

    xMin = xMax = readReal(24);			// A11: four corners
    yMin = yMax = readReal(24);
    for (int i = 0; i < 3; i++) {
	double cx = readReal(24);
	double cy = readReal(24);
	xMin = Math.min(xMin, cx);
	xMax = Math.max(xMax, cx);
	yMin = Math.min(yMin, cy);
	yMax = Math.max(yMax, cy);
    }

    zMin = readReal(24);			// A12
    zMax = readReal(24);
    readReal(24);				// A13  (axis angle)
    readInteger(6);				// A14  (accuracy)

    xRes = readReal(12);			// A15
    yRes = readReal(12);
    zRes = readReal(12);

    // LL values are truncated instead of raised to next multiple
    // in order to handle erroneous DEM files
    xLL = xRes * (int) (xMin / xRes);
    yLL = yRes * (int) (yMin / yRes);
    // likewise rows is too big by one to handle both bad and good files
    rows = (int)(2 + (yMax - yMin) / yRes);

    int colsInProfiles = readInteger(6);	// A16
    if (colsInProfiles != 1) {
	throw new Error("Columns-per-profile field is not 1");
    }
    cols = readInteger(6);

    //  "Note: Old format stops here" --DLG Spec, p. 2-25
    //
    //	All NIMA DEM's and DTED's lacking datum descriptors are WGS 72.
    //	All DEM's having the new type A record elements have datums as
    //	indicated in type A record, element 27.
    //
    //	1=North American Datum 1927 (NAD 27)
    //	2=World Geodetic System 1972 (WGS 72)
    //	3=WGS 84
    //	4=NAD 83
    //	5=Old Hawaii Datum
    //	6=Puerto Rico Datum
    //
    //	Assume NAD 27 (for UTM) and WGS 72 (for GEO) unless told otherwise.

    int d = -1;
    if (skipData(891 - 865)) {			// A17 - A26
	d = readInteger(2);			// A27
	skipData(1025 - 893);			// A28 - A31, and block padding
    }
    if (d > 0) {
	datum = d;
    } else if (refsys == UTM_SYSTEM) {
	datum = NAD27;
    } else {
	datum = WGS72;
    }

    if (partial) {			// bail out here on partial read
	return;
    }

    // B records

    alloc();		// allocate arrays now that cols and rows are known

    for (int i = 0; i < cols; ++i) {

	// B1
	int row_id = readInteger();
	int col_id = readInteger();
	if (col_id != i + 1 || row_id != 1) {
	    throw new Error("Profile " + (i + 1) + " claims to be " + col_id);
	}

	// B2
	int rowsInProfile = readInteger(6);
	int colsInProfile = readInteger(6);

	// B3
	double firstXInProfile = readReal(24);
	double firstYInProfile = readReal(24);
	int firstIndex = (int) ((firstYInProfile - yLL) / yRes);

	// B4
	int ldatum = (int) (readReal(24) / zRes);

	// B5
	readReal(24);			// minimum elevation in profile
	readReal(24);			// maximum elevation in profile

	// B6				// finally, the data
	for (int j = firstIndex; j < firstIndex + rowsInProfile; j++) {
	    int e = ldatum + readInteger();
	    if (e == MISSING_VALUE) {
		elev[j][i] = MISSING_ELEV;
	    } else {
		elev[j][i] = (char) (e + ELEVATION_BIAS);
	    }
	}

    }

    // we don't even read the C record

}



/**
 *  Skips a fixed amount of data, but stops at CR or LF.
 *  Returns true if successful (not terminated early).
 */
private boolean skipData(int len) throws IOException {
    while (len-- > 0) {
	int c = inputFile.read();
	if (c < 0) {
	    throw new EOFException(filename);
	} else if (c == '\n' || c == '\r') {
	    return false;
	}
    }
    return true;
}


/**
 *  Reads a fixed-length input string.
 */
private String readString(int len) throws IOException {
    char[] buf = new char[len];
    int n = 0;
    while (n < len) {
	int c = inputFile.read();
	if (c < 0) {
	    throw new EOFException(filename);
	}
	if (c != '\n' && c != '\r') {		 // ignore spurious CR / LF
	    buf[n++] = (char) c;
	}
    }
    return new String(buf);
}


/**
 *  Reads an integer of unknown length.
 */
private int readInteger() throws IOException {
    StringBuffer b = new StringBuffer();
    int c = inputFile.read();
    while (c == ' ' || c == '\n' || c == '\r') {
	c = inputFile.read();
    }
    if (c < 0) {
	throw new EOFException(filename);
    }
    b.append((char) c);
    for (;;) {
	c = inputFile.read();
	if (c == '\n' || c == '\r') {
	    continue;				// ignore spurious CR / LF
	}
	if (c < '0' || c > '9') {
	    break;
	}
	b.append((char) c);
    }
    if (c >= 0) {
	inputFile.unread(c);
    }
    return Integer.parseInt(b.toString());
}


/**
 *  Reads an integer from a fixed-length input field.
 */
private int readInteger(int len) throws IOException {
    String s = readString(len);
    int i = s.lastIndexOf(' ');
    if (i == s.length() - 1) {
	return 0;
    } else {
	return Integer.parseInt(s.substring(i + 1));
    }
}



/**
 *  Reads a double value from a fixed-length input field.
 *  Accepts Fortran D notation in addition to the usual Java forms.
 */
private double readReal(int len) throws IOException {
    return Double.parseDouble(readString(len).replace('D', 'E'));
}



} // class NativeGrid
