package topovista.grid;

import java.io.*;



/**
 *  The PGMGrid class writes elevation grids as PGM files.
 */

class PGMGrid extends Grid {



private final static int PGM_MAX = 255;
private final static int N_CONTOURS = 16;



/**
 *  Writes a Grid object as a grid of elevations in plain PGM format.
 *  Negative elevations are turned positive.
 */
public static void writePGM(PrintStream p, Grid g) {

    int min = (int) ((g.zMin - 1) / g.zRes);
    int max = (int) ((g.zMax + 1) / g.zRes);
    if (max < -min) {
	max = -min;
    }
    String[] istr = new String[max + 1];
    for (int i = 0; i <= max; i++) {
	istr[i] = " " + i;
    }
    initPGM(p, "P2", g.cols, g.rows, max, g.header);
    for (int i = g.rows - 1; i >= 0; i--) {
	g.loadRow(i);
	for (int j = 0; j < g.cols; j++) {
	    int v = g.elev[i][j];
	    if (v > 0) {
		v -= Grid.ELEVATION_BIAS;
	    }
	    if (v < 0) {
		v = -v;
	    }
	    p.print(istr[v]);
	}
	p.println();
	g.discardRow(i);
    }
}



/**
 *  Writes a contour-map raw PGM file of a Grid object.
 */
public static void writeContour(PrintStream p, Grid g) {
    initPGM(p, "P5", g.cols, g.rows, N_CONTOURS - 1, g.header);
    double b = Grid.ELEVATION_BIAS + (g.zMin / g.zRes);
    double m = g.zRes * N_CONTOURS / (g.zMax - g.zMin + 1);
    for (int i = g.rows - 1; i >= 0; i--) {
	g.loadRow(i);
	for (int j = 0; j < g.cols; j++) {
	    int e = g.elev[i][j];
	    if (e == Grid.MISSING_ELEV) {
		p.write(N_CONTOURS - 1);
	    } else {
		double x = e - b;
		double y = m * x;
		p.write((int) y);
	    }
	}
	g.discardRow(i);
    }
}



/**
 *  Writes a shaded-relief raw PGM file of a Grid object.
 */
public static void writeRelief(PrintStream p, Grid g) {
    initPGM(p, "P5", g.cols, g.rows - 1, PGM_MAX, g.header);
    double m = 150 * g.zRes / g.yRes;		// contrast multiplier
    if (g.zUnit == FEET) {
	m *= 0.3048;				// meters per foot
    }
    if (g.xyUnit == ARCSEC) {
	m /=  30.8642;				// meters per arc second
    }
    g.loadRow(g.rows - 1);
    for (int i = g.rows - 1; i > 0; i--) {
	g.loadRow(i - 1);
	for (int j = 0; j < g.cols; j++) {
	    int a = g.elev[i-1][j];
	    int b = g.elev[i][j];
	    if (a == Grid.MISSING_ELEV || b == Grid.MISSING_ELEV) {
		p.write(PGM_MAX / 4);		// missing data
	    } else if (b == Grid.ELEVATION_BIAS) {
		p.write(3 * PGM_MAX / 4);	// sea level
	    } else {
		int o = (int) (PGM_MAX / 2 + m * (a - b));
		if (o < 0) {
		    p.write(0);
		} else if (o < PGM_MAX) {
		    p.write(o);
		} else {
		    p.write(PGM_MAX);
		}
	    }
	}
	g.discardRow(i);
    }
    g.discardRow(0);
}



/**
 *  Initializes PGM output and writes the header.
 */
private static void initPGM(
PrintStream p, String p2p5, int cols, int rows, int max, String comment) {
    p.println(p2p5);
    if (comment != null) {
	p.println("# " + comment);
    }
    p.println(cols + " " + rows);
    p.println(max);
}



} // class PGMGrid
