package topovista.grid;

/**
 *  A Plane object represents a geometric plane in 3-space,
 *  with provision for evaluating error with respect to a
 *  grid of surface values.
 *
 *  @author Gregg Townsend, University of Arizona
 *  @version March, 2001
 */

public class Plane {

    float a, b, c;		// coefficients: z = a x + b y + c
    float error;		// maximum error seen
    char[][] elev;		// grid of (biased) elevation values



/**
 *  Constructs a Plane based on a biased elevation grid and three x-y
 *  coordinates.  The coordinates specify three points, with the
 *  z values are obtained from the elevation grid.  The constructed
 *  plane passes through these three points.  If the points are
 *  collinear, the results are undefined.  A reference to the elevation
 *  grid is retained for future use.
 */
public Plane(char[][] zdata, int x1, int y1, int x2, int y2, int x3, int y3) {

    elev = zdata;
    int z1 = elev[y1][x1] - Grid.ELEVATION_BIAS;
    int z2 = elev[y2][x2] - Grid.ELEVATION_BIAS;
    int z3 = elev[y3][x3] - Grid.ELEVATION_BIAS;
    float det = x1*y2 + x2*y3 + x3*y1 - x3*y2 - x2*y1 - x1*y3;
    a = (z1*y2 + z2*y3 + z3*y1 - z3*y2 - z2*y1 - z1*y3) / det;
    b = (x1*z2 + x2*z3 + x3*z1 - x3*z2 - x2*z1 - x1*z3) / det;
    c = (x1*y2*z3 + x3*y1*z2 + x2*y3*z1 - x3*y2*z1 - x1*y3*z2 - x2*y1*z3) / det;
}



/**
 *  Computes the z value corresponding to an x and y value,
 *  remembering the maximum error seen with respect to the
 *  grid of z data.
 */
public float eval(int x, int y) {
    float z = a * x + b * y + c;
    float v = elev[y][x] - Grid.ELEVATION_BIAS;
    error = Math.max(error, Math.abs(z - v));
    return z;
}



/**
 *  Returns a String representation of this plane.
 */
public String toString() {
    return "Plane(z=" + a + "x+" + b + "y+" + c + ")";
}



} // class Plane
