package topovista.rpserver;

import java.util.*;

import topovista.grid.*;



/**
 *  Represents one of the individual requests inside a TVRP request message.
 */
class Request implements TVRP {



    Message msg;	// containing message
    int tid;		// tile ID
    int bits;		// bitmap of desired records
    int pri;		// priority (interest level)
    long answered;	// timestamp when answered



/**
 *  Constructs a request object from eight bytes of a TVRP message.
 */
Request(Message msg, byte[] data, int offset) {
    this.msg = msg;
    tid = int4(data, offset);
    int w = int4(data, offset + 4);
    bits = (w & RECS_MASK) >>> RECS_SHIFT;
    pri = (w & PRI_MASK) >>> PRI_SHIFT;
}



/**
 *  Timestamps this request and suppresses duplicate requests.
 *  Any request bit that matches an answer in sentq is cleared
 *  if the response was sent within the last NOREPEAT_MSEC milliseconds.
 *  Also removes from the sent queue any entries older than that,
 *  because they no longer serve any purpose.
 */
public void suppress(LinkedList sentq) {
    this.answered = System.currentTimeMillis();
    synchronized (sentq) {
	Iterator i = sentq.iterator();
	while (i.hasNext()) {
	    Request r = (Request) i.next();
	    if (this.answered - r.answered > NOREPEAT_MSEC) {
		i.remove();
	    } else if (this.tid == r.tid) {
		this.bits &= ~r.bits;
	    }
	}
    }
}



/**
 *  Converts four big-endian data bytes into an integer.
 */
public static int int4(byte[] data, int offset) {
    return
	((data[offset] & 0xFF) << 24) |
	((data[offset + 1] & 0xFF) << 16) |
	((data[offset + 2] & 0xFF) <<  8) |
	((data[offset + 3] & 0xFF) <<  0);
}



/**
 *  Returns a string representation of this request.
 */
public String toString() {
    return "Request(" + TileGrid.hexID(tid) + "," + pri + "," +
	TileGrid.hexID(bits) + ")";
}



} // class Request
