package topovista.rpserver;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;

import topovista.grid.*;


/*
 *  Monitors a set of request queues and processes requests
 *  in priority order.  If the queues go empty, the responder
 *  blocks on the containing array and must be notified when
 *  a new entry is added.
 *
 *  The responder is a Thread object that must be explicitly
 *  started by the code that creates it.
 */
class Responder extends Thread implements TVRP {

    private int delay;			// artificial delay per send, in msec
    private double dropFract;		// fraction of output to drop
    private LinkedList[] rqqueues;	// incoming request queues to monitor
    private DatagramSocket socket;	// socket for sending replies
    private Hashtable dtable;		// maps filenames to their directories

    private byte[] nodata = new byte[INFO_HEADER];


/**
 *  Constructs a responder to process the given set of queues.
 */
Responder(LinkedList[] rqqueues, int delay, double dropFract) throws Exception {
    System.err.println("Initializing Responder:");
    this.rqqueues = rqqueues;
    this.delay = delay;
    this.dropFract = dropFract;
    this.socket = new DatagramSocket();
    this.dtable = new Hashtable();
    scanDirectory(new File("."));	// initialize file lookup table
}



/**
 *  Loops forever processing requests.
 */
public void run() {

    for (;;) try {

	Request r = bestRequest(rqqueues);	// get request
	Message m = r.msg;			// get containing message
	Client c = m.client;			// get client
	if (m != c.latest) {			// if message is old, ignore
	    continue;
	}
	LinkedList done = c.done;		// get list of reqs answered
	r.suppress(done);			// exclude records just sent
	if (r.bits == 0) {			// if nothing left, ignore
	    continue;
	}
	File f = (File) dtable.get(new Integer(r.tid));	// find data file
	if (f != null) {
	    respond(r, f);			// send back data
	} else {
	    r.bits = RECS_MASK;			// remember as ALL records unav
	    reply(r, UNAV_MASK | RECS_MASK, nodata);	// send reply
	    report(r, 0, RECS_MASK);		// log nothing sent
	}
	synchronized (done) {
	    done.add(r);			// remember request as processed
	}
    } catch (Throwable e) {
	e.printStackTrace();
	System.exit(1);
    }
}



/**
 *  Scans a directory, recursively, adding all TopoVista files
 *  to the global Hashtable dtable.
 */
void scanDirectory(File dir) {
    int n = 0;
    String[] flist = dir.list();
    for (int i = 0; i < flist.length; i++) {
	String fname = flist[i];
	File f = new File(dir, fname);
	if (fname.endsWith(".tvg")) {
	    int tid = TileGrid.makeID(fname);
	    dtable.put(new Integer(tid), f);
	    n++;
	} else {
	    if (f.isDirectory()) {
		scanDirectory(f);
	    }
	}
    }
    if (n > 0) {
	System.err.println("   " + dir + ":\t" + n + " files");
    }
}



/**
 *  Removes and returns the best request from the request queues,
 *  waiting if there is no request currently.
 */
Request bestRequest(LinkedList[] rqqueues) throws Exception {
    for (;;) {
	synchronized (rqqueues) {
	    for (int i = PRI_MAX; i >= 0; i--) {
		LinkedList ll = rqqueues[i];
		if (rqqueues[i].size() > 0) {
		    return (Request) rqqueues[i].removeFirst();
		}
	    }
	    rqqueues.wait();
	}
    }
}



/**
 *  Responds to a request for data.
 */
void respond(Request r, File f) throws Exception {

    byte[][] data = load(f);		// load data
    int wanted = r.bits;		// bitmap of requested records
    int avail = 0;			// bitmap of available records

    for (int i = 0; i < RECS_BITS; i++) {
	int j = RECS_SHIFT + RECS_BITS - 1 - i;
	int b = 1 << j;
	if (data[j] != null) {
	    avail |= b;
	    if ((wanted & b) != 0) {
		reply(r, b, data[j]);
		wanted &= ~b;
	    }
	}
    }
    int sent = r.bits & ~wanted;
    int unav = wanted;

    if (wanted != 0) {			// if any wanted record was unavailable
	// report ALL unavailable records,
	// not just those requested with this particular priority
	unav = RECS_MASK & ~avail;
	reply(r, UNAV_MASK | unav, nodata);
	r.bits |= unav;			// add to request, since we answered
    }

    report(r, sent, unav);		// log records sent
}



/**
 *  Reads a TopoVista tile (version 4F) from the specified file,
 *  returning an array of records indexed by the shift count of the
 *  corresponding request bit.  Each record is a byte array holding
 *  one TV record with room for a message header at the front.
 */
byte[][] load(File f) throws Exception {

    DataInputStream ins =
	new DataInputStream(new BufferedInputStream(new FileInputStream(f)));
    byte[][] data = new byte[RECS_BITS][];

    for (;;) {
	try {
	    byte[] hdr = new byte[8];
	    ins.readFully(hdr);
	    if (!(new String(hdr, 0, 4)).equals("TV4F")) {
		return data;
	    }
	    int n = ((hdr[5] & 0xFF) << 16) +
		((hdr[6] & 0xFF) << 8) + (hdr[7] & 0xFF);
	    byte[] rec = new byte[INFO_HEADER + n];
	    System.arraycopy(hdr, 0, rec, INFO_HEADER, hdr.length);
	    ins.readFully(rec, INFO_HEADER + hdr.length, n - hdr.length);
	    int flag = hdr[4] & 0x7F;
	    if (flag == 0x10) {
		data[RECS_BITS - 1] = rec;
	    } else if (flag >= 0x20 && flag <= 0x28) {
		data[RECS_BITS - 2 - (flag - 0x20)] = rec;
	    } else if (flag >= 0x41 && flag <= 0x50) {
		data[RECS_BITS - 2 - 9 - (flag - 0x41)] = rec;
	    }
	} catch (EOFException e) {
	    ins.close();
	    return data;
	}
    }
}



/**
 *  Sends an info message with the given bitmap in the header.
 *  The header is written into the first INFO_HEADER bytes of
 *  the supplied buffer, which must have space reserved
 *  for this purpose.
 */
void reply(Request r, int mask, byte[] buf) throws Exception {

    pack(buf, 0, 0x5456494E);		// "TVIN"
    pack(buf, 4, r.tid);		// tile ID
    pack(buf, 8, mask);			// bitmap

    if (Math.random() >= dropFract) {	// if not chosen for random drop
	socket.send(new DatagramPacket(buf, buf.length, r.msg.client.addr));
    }
    if (delay > 0) {
	Thread.sleep(delay);
    }

}



/**
 *  Packs a 4-byte integer into a buffer in big-endian fashion.
 */
public static void pack(byte[] buf, int offset, int value) {
    buf[offset+3] = (byte) value;  value >>= 8;
    buf[offset+2] = (byte) value;  value >>= 8;
    buf[offset+1] = (byte) value;  value >>= 8;
    buf[offset+0] = (byte) value;
}



/**
 *  Writes a log message to stderr indicating records sent and denied.
 */
void report(Request r, int sent, int unav) {
    System.err.print(
	df.format(new Date(r.answered)) + TileGrid.makeName(r.tid) + " ");
    for (int i = RECS_BITS - 1; i >= 0; i--) {
	int b = 1 << (RECS_SHIFT + i);
	if ((sent & b) != 0) {
	    System.err.print("PONMLKJIHGFEDCBAhgfedcbazH".charAt(i));
	} else if ((unav & b) != 0) {
	    System.err.print('*');
	} else {
	    System.err.print('-');
	}
    }
    System.err.println(" " + r.pri + " " + r.msg.client.addr);
}
private static DateFormat df =
    new SimpleDateFormat("yyyy-MM-dd HH:mm:ss ");



/**
 *  Returns a string representation of this responder.
 */
public String toString() {
    return "Responder(" + rqqueues + "," + socket.getLocalPort() + ")";
}



} // class Responder
