//  Broker.java -- token server interface program


package topovista.toksvc;

import java.io.*;
import java.util.*;



/**
 *  Implements an interactive client for the token service.
 */
public class Broker {


static Client client;
static Revoker revoker;

static BufferedReader reader;
static boolean scripted;

static Hashtable tokens = new Hashtable();



/**
 *  <P> The Broker class provides a command-line program for
 *  interactive or scripted testing of a token server.
 *  The Broker creates a single Client instance and communicates
 *  with an existing token server.
 *  Commands are passed to the Client and results are printed
 *  on standard output.
 *  Asynchronous <STRONG>revoke</STRONG> messages
 *  are also printed on standard output.
 *
 *  <P> Usage: java Broker configfile [scriptfile]
 *
 *  <P> Commands: <UL>
 *	<LI>  <STRONG>#</STRONG> <VAR>comment</VAR>
 *	<LI>  <STRONG>get</STRONG> <VAR>name</VAR>
 *	<LI>  <STRONG>put</STRONG> <VAR>name</VAR> <VAR>datastring</VAR>
 *	<LI>  <STRONG>request</STRONG> <VAR>name</VAR> [<STRONG>x</STRONG>]
 *	<LI>  <STRONG>update</STRONG>  <VAR>name</VAR> [<VAR>datastring</VAR>]
 *	<LI>  <STRONG>release</STRONG> <VAR>name</VAR> [<VAR>datastring</VAR>]
 *	<LI>  <STRONG>list</STRONG>
 *	<LI>  <STRONG>clear</STRONG>
 *	<LI>  <STRONG>close</STRONG>
 *  </UL>
 */

public static void main(String args[]) throws Exception {

    if (args.length < 1 || args.length > 2) {
	System.err.println("usage: java Broker configfile [scriptfile]");
	System.exit(1);
    }
    System.out.println("[connecting...]");
    client = new Client(new Configuration(args[0]));
    client.setTrace(true);
    if (args.length == 2) {
	scripted = true;
	reader = new BufferedReader(new FileReader(args[1]));
    } else {
	reader = new BufferedReader(new InputStreamReader(System.in));
    }
    System.out.println("[ready]");

    revoker = new Revoker() {
	public void revoke(Token t, Object o) {
	    if (scripted) {
		System.out.println("\tREVOKE " + t.name);
	    } else {
		System.out.println("\n\tREVOKE " + t.name + "\n> ");
	    }
	}
    };

    for (;;) {

	System.out.print("> ");
	String line = reader.readLine();
	if (line == null) {
	    client.close();
	    if (scripted) {
		System.out.println();
	    }
	    System.exit(0);
	}
	if (scripted) {
	    System.out.println(line);
	}

	StringTokenizer st = new StringTokenizer(line);
	int n = st.countTokens();
	if (n == 0) {
	    continue;
	}

	String cmd = st.nextToken().intern();
	String arg1 = (n > 1) ? st.nextToken() : null;
	String arg2 = (n > 2) ? st.nextToken() : null;

	if (cmd.charAt(0) == '#') {
	    continue;
	} else if (cmd == "get") {
	    get(arg1);
	} else if (cmd == "put") {
	    put(arg1, arg2);
	} else if (cmd == "request") {
	    request(arg1, arg2);
	} else if (cmd == "update") {
	    update(arg1, arg2);
	} else if (cmd == "release") {
	    release(arg1, arg2);
	} else if (cmd == "list") {
	    list();
	} else if (cmd == "clear") {
	    clear();
	} else if (cmd == "close") {
	    client.close();
	    System.exit(0);
	} else {
	    System.out.println("\tUNKNOWN COMMAND: " + cmd);
	}

    }
}


static void list() {
    int n = 0;
    Enumeration e = tokens.keys();
    while (e.hasMoreElements()) {
	String s = (String) e.nextElement();
	Token t = (Token) tokens.get(s);
	System.out.println("\t" + t.name + "  " + new String(t.getData()));
	n++;
    }
    if (n == 0) {
	System.out.println("\tNONE HELD");
    }
}


static void get(String name) {
    if (name == null) {
	System.out.println("\nGET WHAT?");
	return;
    }
    try {
	System.out.println("\t" + new String(client.get(name)));
    } catch (Throwable e) {
	System.out.println("\t" + e);
    }
}


static void put(String name, String val) {
    if (name == null) {
	System.out.println("\nPUT WHAT?");
	return;
    }
    if (val == null) {
	val = "";
    }
    try {
	client.put(name, val.getBytes());
	System.out.println("\tOK");
    } catch (Throwable e) {
	System.out.println("\t" + e);
    }
}


static void request(String name, String access) {
    if (name == null) {
	System.out.println("\nREQUEST WHAT?");
	return;
    }
    if (tokens.get(name) != null) {
	System.out.println("\tALREADY HELD");
	return;
    }
    char c = ' ';
    if (access != null && access.length() > 0) {
	c = access.charAt(0);
    }
    int a = (c == 'x' || c == 'w') ?  Token.EXCLUSIVE : Token.SHARED;
    try {
	Token t = client.request(name, a, revoker, null);
	tokens.put(name, t);
	System.out.println("\t" + t.name + "  " + new String(t.getData()));
    } catch (Throwable e) {
	System.out.println("\t" + e);
	e.printStackTrace();
    }
}


static void update(String name, String val) {
    if (val == null) {
	val = "";
    }
    try {
	Token t = lookup(name);
	t.update(val.getBytes());
	System.out.println("\tOK");
    } catch (Throwable e) {
	System.out.println("\t" + e);
    }
}


static void release(String name, String val) {
    try {
	Token t = lookup(name);
	if (val != null) {
	    t.release(val.getBytes());
	} else {
	    t.release();
	}
	tokens.remove(name);
	System.out.println("\tOK");
    } catch (Throwable e) {
	System.out.println("\t" + e);
    }
}


static void clear() {
    int n = 0;
    Vector v = new Vector();
    Enumeration e = tokens.keys();
    while (e.hasMoreElements()) {
	v.addElement(e.nextElement());
    }
    for (int i = 0; i < v.size(); i++) {
	String name = (String) v.elementAt(i);
	Token t = lookup(name);
	t.release();
	tokens.remove(name);
    }
    System.out.println("\tOK");
}



static Token lookup(String name) {
    if (name == null) {
	throw new Error("token name missing");
    }
    Token t = (Token) tokens.get(name);
    if (t == null) {
	throw new Error("token not found: " + name);
    }
    return t;
}



} // class Broker
