//  CircularBuffer.java -- for maintaining the last N events of some sort


package topovista.toksvc;

import java.io.*;
import java.util.*;



/**
 *  CircularBuffer provides a form of cyclic storage
 *  where the last N items are available and the newest item
 *  overwrites the oldest one.
 */

public class CircularBuffer {

    private Object[] data;		// item storage
    private int cap;			// capacity (length) of array
    private int next;			// next location to store
    private int used;			// current usage (0 <= used <= cap)
    private int serial;			// serial number of this buffer



private static int nextsn = 1;			// serial number of next buffer



/**
 *  The size of a buffer when not specified.
 */
public final static int DEFAULT_SIZE = 1000;



/**
 *  Creates a new buffer of default capacity.
 */
public CircularBuffer() {
    this(DEFAULT_SIZE);
}



/**
 *  Creates a new buffer of specified capacity.
 */
public CircularBuffer(int n) {
    data = new Object[n];
    cap = n;
    next = 0;
    used = 0;
    serial = nextsn++;
}



/**
 *  Returns the capacity of the buffer.
 */
public int capacity() {
    return cap;
}



/**
 *  Returns the current usage of the buffer.
 */
public int size() {
    return used;
}



/**
 *  Returns a string representation of the buffer,
 *  including serial number, size, and capacity.
 */
public String toString() {
    return "CircularBuffer(#" + serial + "," + used + "/" + cap + ")";
}



/**
 *  Reinitializes the buffer, setting it empty.
 */
public synchronized CircularBuffer clear() {
    next = 0;
    used = 0;
    return this;
}



/**
 *  Adds one item to the buffer, discarding the oldest item if necessary.
 */
public synchronized CircularBuffer add(Object o) {
    data[next++] = o;
    next %= cap;
    if (used < cap) {
	used++;
    }
    return this;
}



/**
 *  Returns an array containing all current items, from oldest to newest.
 */
public Object[] toArray() {
    return toArray(new Object[used]);
}



/**
 *  Copies the current items, from oldest to newest, into the provided
 *  array, and returns the array.  If the array is insufficient, only
 *  the newest items are returned.  If the array is too large, extra
 *  items are filled with nulls.
 */
public synchronized Object[] toArray(Object[] a) {
    int n = a.length < used ? a.length : used;
    int o = (cap + next - n) % cap;
    int i = 0;
    while (i < n) {
	a[i++] = data[o++];
	o %= cap;
    }
    while (i < a.length) {
	a[i++] = null;
    }
    return a;
}



/**
 *  Returns an enumeration of the current buffer contents.
 */
public synchronized Enumeration elements() {
    return new Enumeration() {
	Object[] a = data;
	int c = cap;
	int i = cap + next - used;
	int j = i + used;
	public boolean hasMoreElements() {
	    return i < j;
	}
	public Object nextElement() {
	    if (i < j) {
		return a[i++ % cap];
	    } else {
		throw new NoSuchElementException();
	    }
	}
    };
}



}  // class CircularBuffer
