//  Client.java -- procedures for interfacing to token server



package topovista.toksvc;

import java.io.*;
import java.net.*;
import java.util.*;



/**
 *  A Client provides the programming interface to a token server.
 */

public class Client {

    private Hashtable tokens;		// table of known tokens

    private Timer hbtimer;		// heartbeat timer

    private DatagramSocket rcvport;	// inbound socket
    private int portnum;		// associated port number

    private int sessID;			// our session ID
    private boolean closing;		// set true if closing

    private DatagramSocket[] sendport;	// outbound sockets
    private InetAddress[] shost;	// server host addresses
    private int[] sport;		// server port numbers
    private int msig;			// server list signature

    private Timer atimer;		// message acknowledgment timer
    private Hashtable awaiting;		// messages currently awaiting acks

    private int leader;			// index of token leader
    private byte[] sstates;		// server state array

    private boolean trace;		// trace messages on stdout?
    private String tlabel;		// tracing label with session ID



private static final int HEARTBEAT = 3000;	// heartbeat interval, in msec
private static final int LEADER_WAIT = 1000;	// wait for initial reply, msec

private static final int TIMEOUT = 5000;	// wait for message ack



/**
 *  Constructs a Client instance, returning only after
 *  communication has been established with the token service.
 */
public Client(Configuration g) throws IOException {
    this(g, 0);
}



/**
 *  Constructs a Client instance.  If a nonzero timeout is given, and
 *  the token service has not been contacted after that many seconds,
 *  then an IOException is thrown instead.
 */
public Client(Configuration g, int timeout) throws IOException {

    rcvport = new DatagramSocket();	// open incoming datagram port
    portnum = rcvport.getLocalPort();

    servers(g);				// initialize server data structures

    register(timeout);			// log in to token system

    tokens = new Hashtable();		// init empty token collection


    // start thread to listen for incoming messages
    Thread t = new Thread() {
	public void run() {
	    receiver();
	}
    };
    t.setDaemon(true);
    t.start();

    // define timer handler for message acks
    awaiting = new Hashtable();		// table of pending messages
    atimer = new Timer() {
	public void alarm(Timer tm, Object val) {
	    noack();
	}
    };

    // define timer handler for sending heartbeats
    hbtimer = new Timer() {
	public void alarm(Timer tm, Object val) {
	    heartbeat();
	}
    };
    hbtimer.set(HEARTBEAT, null);
}



/**
 *  Initializes data structures for sending to the servers.
 */
private void servers(Configuration g) throws IOException {

    String[] s = g.getArray("server");
    int n = s.length;
    if (n == 0) {
	throw new Error("no servers configured");
    }
    msig = Util.hash(s) & 0x1FFF;

    sendport = new DatagramSocket[n];
    shost = new InetAddress[n];
    sport = new int[n];

    for (int i = 0; i < s.length; i++) {
	sendport[i] = new DatagramSocket();
	shost[i] = InetAddress.getByName(Util.hostPart(s[i]));
	sport[i] = Util.portNumber(s[i]);
    }
}



/**
 *  Contacts a server and registers this client.
 */
private void register(int timeout) throws IOException {

    Message m;
    long expiration = System.currentTimeMillis() + timeout * 1000L;

    rcvport.setSoTimeout(LEADER_WAIT);

    for (;;) {				// repeat until successful

	send(new LoginMessage(":" + portnum), leader);

	int n = Message.MAX_LENGTH;
	DatagramPacket d = new DatagramPacket(new byte[n], n);
	try {
	    rcvport.receive(d);
	} catch (InterruptedIOException e) {
	    if (timeout > 0 && System.currentTimeMillis() >= expiration) {
		throw new IOException("cannot contact token service");
	    }
	    leader = (leader + 1) % sendport.length;
	    continue;
	}

	m = Message.unpickle(d);
	if (m instanceof ConfigMessage) {
	    leader = ((ConfigMessage) m).leader;
	    if (m.to != 0) {
		break;
	    }
	}
    }

    // have received a message with nonzero session ID; we're logged in.
    sessID = m.to;
    sstates = ((ConfigMessage) m).sstates;
    rcvport.setSoTimeout(0);

    tlabel = "C" + sessID + ": ";
}



/**
 *  Updates the current server configuration.  The client responds with
 *  a CatalogMessage detailing this client's holdings of tokens
 *  currently served by the sender.
 */
private void config(ConfigMessage m) {

    leader = m.leader;			// save possibly new leader
    sstates = m.sstates;		// save server states

    Vector v = new Vector();
    synchronized (tokens) {
	int i = 0;
	for (Enumeration e = tokens.elements(); e.hasMoreElements(); ) {
	    Token t = (Token) e.nextElement();
	    if (t.isAssigned() && t.server(sstates) == m.from) {
		v.addElement(t);
	    }
	}
    }
    Token[] tklist = new Token[v.size()];
    v.copyInto(tklist);
    send(new CatalogMessage(tklist), m.from);
}



/**
 *  Enables or disables the tracing of incoming messages on standard output.
 */
public void setTrace(boolean b) {
    if (b != trace) {
	System.out.println(tlabel + "setTrace(" + b + ")");
	if (b) {
	    System.out.println(tlabel + this + "  config=(" +
		leader + "," + ConfigMessage.simage(sstates) + ")");
	}
    }
    trace = b;
}



/**
 *  Listens for incoming messages, and processes them.
 *  (Runs continuously in a spawned daemon thread.)
 */
private void receiver() {
    for (;;) {
	try {
	    int n = Message.MAX_LENGTH;
	    DatagramPacket d = new DatagramPacket(new byte[n], n);
	    rcvport.receive(d);
	    Message m = Message.unpickle(d);
	    if (trace) {
		System.out.println(tlabel + m);
	    }
	    switch (m.type) {
		case ConfigMessage.MTYPE:  config((ConfigMessage) m);    break;
		case ConfirmMessage.MTYPE: confirm((ConfirmMessage) m);  break;
		case GrantMessage.MTYPE:   grant((GrantMessage) m);      break;
		case RevokeMessage.MTYPE:  revoke((RevokeMessage) m);    break;
	    }
	} catch (Throwable t) {
	    if (!closing) {
		Util.abort("receiver", t);
	    }
	}
    }
}



/**
 *  Sends a heartbeat to the main server when activated by the timer.
 */
private void heartbeat() {
    send(new AliveMessage(), leader);
    hbtimer.set(HEARTBEAT, null);
}



/**
 *  Sends a message to the indicated server.
 */
private void send(Message m, int to) {

    if (closing) {
	throw new Error("attempt to use closed session");
    }

    m.from = sessID;
    m.to = to;
    m.msig = msig;
    byte[] buf = m.pickle(new Pickle()).toByteArray();

    DatagramPacket p = new DatagramPacket(buf,buf.length,shost[to],sport[to]);
    try {
	sendport[to].send(p);
    } catch (IOException e) {
    }
}



/**
 *  Sends a NumberedMessage to the responsible server.
 */
void tsend(NumberedMessage m) {
    send(m, m.token.server(sstates));
}



/**
 *  Sends a NumberedMessage for which acknowledgment is expected.
 *  The message is periodically reissued until properly acknowledged.
 *  This is used by the Token code.
 */
void asend(NumberedMessage m) {
    awaiting.put(new Long(m.msgnum), m);	// register message
    tsend(m);					// send it
    if (!atimer.isArmed()) {
	atimer.set(TIMEOUT, null);		// arm timer, if not active
    }
}



/**
 *  Notes a message acknowledgment, returning the corresponding
 *  token if one can be found.
 */
Token ack(long msgnum) {
    NumberedMessage m;
    synchronized(awaiting) {
	m = (NumberedMessage) awaiting.remove(new Long(msgnum));
						// remove pending message
	if (awaiting.size() == 0) {		// if there are no more,
	    atimer.clear();			// stop the timer
	}
    }
    if (m != null) {				// if outgoing message found
	return (Token) m.token;			// return its token
    } else {
	return null;
    }
}



/**
 *  Handles an alarm indicating it's time to check for
 *  unacknowledged messages.
 */
void noack() {
    synchronized(awaiting) {
	if (awaiting.size() > 0) {		// if any unack'd messages
	    for (Enumeration e = awaiting.elements(); e.hasMoreElements(); ) {
		NumberedMessage m = (NumberedMessage) e.nextElement();
		tsend(m);			// resend message
	    }
	    atimer.set(TIMEOUT, null);		// reset timer
	}
    }
}



/**
 *  Obtains a token for shared or exclusive use.
 */
public Token request(String name, int access, Revoker r, Object o) {

    Token t;
    synchronized (tokens) {
	t = (Token) tokens.get(name);
	if (t == null) {
	    tokens.put(name, t = new Token(this, name, r, o));
	}
	t.requested = true;
    }
    t.request(access);
    t.requested = false;
    return t;
}



/**
 *  Reads the data of an unheld token without retaining a lock.
 *  The token is read using shared access.
 */
public byte[] get(String name) {
    Token t = request(name, Token.SHARED, null, null);
    byte[] value = t.getData();
    t.release();
    return value;
}



/**
 *  Assigns the data of an unheld token without retaining a lock.
 *  The token is modified using exclusive access.
 */
public void put(String name, byte[] value) {
    request(name, Token.EXCLUSIVE, null, null).release(value);
}



/**
 *  Deletes a token from the collection.
 */
void forget(Token t) {
    synchronized (tokens) {
	if (!t.requested) {		// if no new request in progress
	    tokens.remove(t.name);	// remove from table
	}
    }
}



/**
 *  Dispatches an incoming GrantMessage to the proper token.
 */
private void grant(GrantMessage m) {
    Token t = ack(m.msgnum);
    if (t != null) {
	t.grant(m);
    }
}



/**
 *  Dispatches an incoming RevokeMessage to the proper token.
 */
private void revoke(RevokeMessage m) {
    Token t = (Token) tokens.get(m.name);
    if (t != null) {
	t.revoke(m);
    } else {
	// we don't have this token; respond by deassigning
	send(new ReturnMessage(
		0, new Token(m.name, null), ReturnMessage.DEASSIGN),
	    m.from);
    }
}



/**
 *  Dispatches an incoming ConfirmMessage to the proper token.
 */
private void confirm(ConfirmMessage m) {
    Token t = ack(m.msgnum);
    if (t != null) {
	t.confirm(m);
    }
}



/**
 *  Closes all sockets and renders the Client inoperable.
 */
public synchronized void close() {
    if (!closing) {
	send(new LogoutMessage(), leader);
	closing = true;
	rcvport.close();	// interrupt the listener thread
	for (int i = 0; i < sport.length; i++) {
	    sendport[i].close();
	}
    }
}



/**
 *  Finalizes the client by calling close().
 */
public void finalize() {
    close();
}



/**
 *  Returns a string representation of this Client.
 *  The string includes session ID, host:port, and held token count.
 */
public String toString() {
    return Util.classname(this) + "(" + sessID + ",:" + portnum + ")";
}



}  // class Client
