//  Configuration.java -- dynamic configuration support.


package topovista.toksvc;

import java.io.*;
import java.util.*;



/**
 *  Configuration reads parameters from a configuration file and
 *  returns their values on demand.  A configuration file contains lines
 *  of the form
 *  <BLOCKQUOTE>
 *	WS name WS value WS # comment
 *  </BLOCKQUOTE>
 *  where WS is whitespace and every field is optional.
 *  Only lines containing a name have any effect.
 *  Names may repeat to build up a list.
 *  Values may be quoted.
 */

public class Configuration {

    private String[] names;
    private String[] values;



/**
 *  Loads configuration data from the named file.
 */
public Configuration(String filename)
throws IOException {
    this(new FileReader(filename));
}



/**
 *  Loads configuration data from the given file.
 */
public Configuration(Reader r)
throws IOException {

    BufferedReader rdr = new BufferedReader(r);
    StreamTokenizer tk = toker(rdr);
    Vector nmvec = new Vector();
    Vector vavec = new Vector();

    String name = null;
    String val = null;
    int ttype;
    while ((ttype = tk.nextToken()) != StreamTokenizer.TT_EOF) {
	if (ttype == StreamTokenizer.TT_EOL) {
	    if (name != null) {
		nmvec.addElement(name.intern());
		vavec.addElement((val == null) ? "" : val.intern());
		name = null;
		val = null;
	    }
	} else if (name == null) {
	    name = tk.sval;
	} else if (val == null) {
	    val = tk.sval;
	} else {
	    val = val + " " + tk.sval;
	}
    }

    nmvec.copyInto(names = new String[nmvec.size()]);
    vavec.copyInto(values = new String[vavec.size()]);
    rdr.close();
}



/**
 *  Creates and configures a StreamTokenizer for parsing the input.
 */
private static StreamTokenizer toker(Reader r) {
    StreamTokenizer tk = new StreamTokenizer(r);
    tk.resetSyntax();
    tk.whitespaceChars(0, ' ');
    tk.wordChars('!', '~');
    tk.wordChars(0xA0, 0xFF);
    tk.quoteChar('"');
    tk.quoteChar('\'');
    tk.commentChar('#');
    tk.eolIsSignificant(true);
    return tk;
}



/**
 *  Returns an array containing the values specified for the given name.
 */
public String[] getArray(String name) {
    int n = 0;
    for (int i = 0; i < names.length; i++) {
	if (names[i].equals(name)) {
	    n++;
	}
    }
    String[] a = new String[n];
    n = 0;
    for (int i = 0; i < names.length; i++) {
	if (names[i].equals(name)) {
	    a[n++] = values[i];
	}
    }
    return a;
}



/**
 *  Returns the last value specified for the given name,
 *  or null if the name was not specified.
 */
public String getString(String name) {
    int i = names.length;
    while (--i >= 0) {
	if (names[i].equals(name)) {
	    return values[i];
	}
    }
    return null;
}



/**
 *  Returns the last specified value for the given name,
 *  if one exists and it is interpretable as an integer;
 *  otherwise returns Integer.MIN_VALUE.
 */
public int getInt(String name) {
    int i = names.length;
    while (--i >= 0) {
	if (names[i].equals(name)) {
	    break;
	}
    }
    try {
	return Integer.parseInt(values[i]);
    } catch (Exception e) {
	return Integer.MIN_VALUE;
    }
}



}  // class Configuration
