//  Logger.java -- a facility for maintaining debugging logs


package topovista.toksvc;

import java.io.*;
import java.util.*;



/**
 *  Logger provides facilities for maintaining debugging logs.
 *
 *  <P> Log messages are classified by <DFN>category</DFN>, and messages
 *  can be enabled or disabled by category.  Each category is identified
 *  by a single-character code.
 *
 *  <P> If no output file is defined, log messages accumulate in a circular
 *  buffer, with older messages discarded to make room for newer ones.
 *  The buffer is dumped by defining an output file.
 */

public class Logger {

    private CircularBuffer buf = new CircularBuffer();	// circular buffer
    private PrintWriter file = null;			// output file
    private BitSet inhibit = new BitSet(128);		// disabled categories



/**
 *  Creates a new logger with no associated output file.
 */
public Logger() {
}



/**
 *  Creates a new logger with the specified output file.
 */
public Logger(String filename)
throws IOException {
    this();
    setFile(filename);
}



/**
 *  Creates a new logger with the specified output file.
 */
public Logger(PrintWriter logfile) {
    this();
    setFile(logfile);
}



/**
 *  Sets the file to receive subsequent logging output.
 */
public void setFile(String filename)
throws IOException {
    setFile(new PrintWriter(new FileWriter(filename)));
}



/**
 *  Sets the file to receive subsequent logging output.
 */
public void setFile(PrintWriter logfile) {
    if (file != null) {
	file.flush();
    }
    file = logfile;
    for (Enumeration e = buf.elements(); e.hasMoreElements(); ) {
	file.println(e.nextElement());
	}
    buf.clear();
}



/**
 *  Closes the log file, if any, associated with this logger.
 *  Subsequent messages accumulate in the circular buffer
 *  but are not written to any file.
 */

public void close() {
    if (file != null) {
	file.close();
    }
    file = null;
}



/**
 *  Enables log messages for the category identified by
 *  a single character.
 */
public void enable(char category) {
    inhibit.clear(category);
}



/**
 *  Enables log messages for the categories identified
 *  by a string of characters.
 */
public void enable(String s) {
    for (int i = 0; i < s.length(); i++) {
	enable(s.charAt(i));
    }
}



/**
 *  Enables log messages for the category identified
 *  by a single character.
 */
public void disable(char category) {
    inhibit.set(category);
}



/**
 *  Disables log messages for the categories identified
 *  by a string of characters.
 */
public void disable(String s) {
    for (int i = 0; i < s.length(); i++) {
	disable(s.charAt(i));
    }
}



/**
 *  Writes a log message, if its associated category is enabled.
 */
public void write(char category, String msg) {
    if (inhibit.get(category)) {
	return;				// if this category disabled
    }
    String s = String.valueOf((char) category) + "  " + msg;
    if (file != null) {
	file.println(s);		// write to file, if set
    } else {
	buf.add(s);			// otherwise put in circular buffer
    }
}



}  // class Logger
