//  ServerToken.java -- a server token


package topovista.toksvc;

import java.util.*;



/**
 *  ServerToken is a Token maintained by a Server, with assignment
 *  information and links for a least-recently-used list.
 */
class ServerToken extends Tk {

    private int asgcount;	// assignment count; -1 if exclusively assigned
    private Vector sessns;	// sessions assigned

    RequestMessage pending;	// pending request

    private ServerToken LRUprev;  // previous entry in LRU chain
    private ServerToken LRUnext;  // next entry in LRU chain



/**
 *  LRU queue.
 */
private static ServerToken LRUhead = new ServerToken("(LRUhead)");
private static ServerToken LRUtail = new ServerToken("(LRUtail)");

static {
    LRUhead.LRUnext = LRUtail;
    LRUtail.LRUprev = LRUhead;
}



/**
 *  Constructs a ServerToken with empty data.
 */
ServerToken(String name) {
    super(name, null);
    sessns = new Vector();
}



/**
 *  Constructs a ServerToken by reading a Pickle.
 */
ServerToken(Pickle p) {
    super(p);
    sessns = new Vector();
}



/**
 *  Constructs a ServerToken by merging data from two other tokens
 *  of the same name.  Assignments are merged.  Data is from the
 *  first argument unless empty.
 */
ServerToken(ServerToken t1, ServerToken t2) {
    super(t1.name, t1.data);
    if (t1.hcode != t2.hcode) {
	throw new Error("attempt to merge incompatible tokens");
    }
    if (data == null || data.length == 0) {
	data = t2.data;
    }

    if (t2.asgcount == 0) {
	this.asgcount = t1.asgcount;
	this.sessns = t1.sessns;
    } else if (t1.asgcount == 0) {
	this.asgcount = t2.asgcount;
	this.sessns = t2.sessns;
    } else if (t1.sessns.size() == 1 && t2.sessns.size() == 1
	    && t1.sessns.firstElement().equals(t2.sessns.firstElement())) {
	this.sessns = (Vector) t1.sessns.clone();
	this.asgcount = (t1.asgcount > 0 && t2.asgcount > 0) ? 1 : -1;
    } else {
	// must merge shared assignment lists
	this.sessns = (Vector) t1.sessns.clone();
	for (int i = 0; i < t2.sessns.size(); i++) {
	    Session s = (Session) t2.sessns.elementAt(i);
	    if (! this.sessns.contains(s)) {
		this.sessns.addElement(s);
	    }
	}
	this.asgcount = sessns.size();
    }
}



/**
 *  Returns <0, 0, or >0 according to whether the ServerToken is
 *  assigned exclusively, unassigned, or assigned in shared mode.
 */
int getState() {
    return asgcount;
}



/**
 *  Returns a list of client sessions holding the ServerToken.
 *  This list should be considered read-only.
 */
Vector getAssignments() {
    return sessns;
}



/**
 *  Checks this token's assignments against the current client sessions,
 *  and cancels any assignments to clients that have disappeared.
 */
void cksess(SessionList sl) {
    if (asgcount == 0) {
	return;
    }
    synchronized (this) {
	// work from back to avoid index changes due to deassignments
	for (int i = sessns.size() - 1; i >= 0; i--) {
	    Session s = (Session) sessns.elementAt(i);
	    if (sl.get(s) == null) {
		deassign(s);
	    }
	}
    }
}



/**
 *  Attempts to assign this token to a session, returning true if successful.
 *  A request that duplicates an existing assignment is considered a success.
 */
boolean assign(Session s, int access) {
    if (access == Token.SHARED) {
	return assignShared(s);
    } else if (access == Token.EXCLUSIVE) {
	return assignExclusive(s);
    } else {
	return false;				// illegal code
    }
}



/**
 *  Attempts an exclusive assignment of this ServerToken.
 */
private synchronized boolean assignExclusive(Session s) {
    if (asgcount > 0) {
	return false;				// currently being shared
    }
    if (asgcount < 0) {
	return sessns.firstElement().equals(s);	// already assigned; same sessn?
    }

    // token is available for assignment
    sessns.addElement(s);
    asgcount = -1;
    delLRU();
    return true;
}



/**
 *  Attempts a shared assignment of this ServerToken.
 */
private synchronized boolean assignShared(Session s) {
    if (asgcount < 0) {
	return false;			// busy in exclusive mode
    }
    if (sessns.contains(s)) {
	return true;			// already assigned to this session
    }
    sessns.addElement(s);
    asgcount++;
    delLRU();
    return true;
}



/**
 *  Cancels any assignment of this ServerToken to a particular session.
 */
synchronized void deassign(Session s) {
    if (sessns.removeElement(s)) {	// if not already deassigned
	if (--asgcount < 0) {		// decrement assignment count
	    asgcount = 0;		// if was exclusive (-1), reset to 0
	}
	if (asgcount == 0) {		// if no longer assigned anywhere
	    addLRU();			// put back on LRU list
	}
    }
}



/**
 *  Returns the first token of the LRU queue, if any, deleting it from the list.
 */
static ServerToken popLRU() {
    synchronized (LRUhead) {
	ServerToken t = LRUhead.LRUnext;
	return (t == LRUtail) ? null : t.delLRU();
    }
}



/**
 *  Removes this token from the LRU queue.
 */
ServerToken delLRU() {
    synchronized (LRUhead) {
	if (this.LRUprev != null) {
	    this.LRUprev.LRUnext = this.LRUnext;
	    this.LRUnext.LRUprev = this.LRUprev;
	    this.LRUprev = this.LRUnext = null;
	}
    }
    return this;
}



/**
 *  Adds this token to the LRU queue.
 *  If already in the queue, moves it to the end.
 */
ServerToken addLRU() {
    synchronized (LRUhead) {
	delLRU();
	this.LRUnext = LRUtail;
	this.LRUprev = LRUtail.LRUprev;
	LRUtail.LRUprev = this;
	this.LRUprev.LRUnext = this;
    }
    return this;
}



}  // class ServerToken
