//  Timer.java -- a simple asynchronous alarm facility



package topovista.toksvc;



/**
 *  Timer provides provides a simple asynchronous alarm
 *  facility.  Subclasses implement an alarm() method that is called when
 *  the timer expires.
 *
 *  <P> For new applications you will almost certainly want to use
 *  java.util.Timer, which was not available when this class was written.
 *  There are probably bugs in here that don't show up in current usage.
 */

public abstract class Timer {

    private Queue timeout;	// message queue for serializing actions
    private Thread watcher;	// thread that performs actions
    private boolean armed;	// is there an alarm call queued?



/**
 *  Constructor.
 */
public Timer() {
    timeout = new Queue();
    watcher = new Thread() {
	public void run() {
	    try {
		watch();
	    } catch (Throwable t) {
		Util.abort("Timer", t);
	    }
	}
    };
    if (watcher.getPriority() < Thread.MAX_PRIORITY) {
	watcher.setPriority(watcher.getPriority() + 1);
    }
    watcher.setDaemon(true);
    watcher.start();
}



/**
 *  Runs as a background process, doing the actual timing and calling.
 */
private void watch() {
    for (;;) {
	Object msg = timeout.get();	// get next serialized action
	if (msg == null) {		// if cancellation signal
	    continue;
	}
	Object val = timeout.get();	// get value to pass on timeout

	long expiry = ((Long)(msg)).longValue();
	while (timeout.size() == 0) {	// while not superseded
	    long current = System.currentTimeMillis();
	    if (current >= expiry) {	// if time to call
		break;
	    } else {
		try {			// if need to sleep
		    Thread.sleep(expiry - current);
		} catch (InterruptedException e) {
		}
	    }
	}

	if (timeout.size() == 0) {	// if not superseded
	    armed = false;
	    try {
		alarm(this, val);	// sound the alarm
	    } catch (Throwable e) {
		// do nothing (all exceptions and errors are ignored)
	    }
	}
    }
}



/**
 *  Sets the timer to go off after the specified number of milliseconds,
 *  replacing any previous setting.
 */
public void set(long msec, Object val) {
    long expiry = System.currentTimeMillis() + msec;
    clear();
    synchronized (timeout) {
	timeout.put(new Long(expiry));	// pass new interval to watcher
	timeout.put(val);		// pass callback value
    }
    armed = true;
    Thread.yield();			// give it a chance to start sleeping
}



/**
 *  Clears any pending timer and renders the timer inactive.
 */
public void clear() {
    timeout.put(null);			// signal cancellation
    armed = false;
    watcher.interrupt();		// interrupt watcher if sleeping
}



/**
 *  Returns true if a timer alarm is currently scheduled.
 *  (Usually correct, but not 100% reliable.)
 */
public boolean isArmed() {
    return armed;
}



/**
 *  Called when the timer goes off.  All delayed alarm() calls are
 *  serialized in a single timer thread, so they are expected to be
 *  reasonably quick, and they are expected to return (either normally
 *  or by throwing an error or exception, which is ignored).
 */
public abstract void alarm(Timer t, Object val);



}  // class Timer
