//  Tk.java -- the base class for Tokens


package topovista.toksvc;

import java.util.*;



/**
 *  Tk is the abstract base class for Token, used by Clients,
 *  and ServerToken, used by Servers.
 */
abstract class Tk {

    public final String name;	// token name
    protected byte[] data;	// token data
    protected int hcode;	// hashcode


private static final byte[] empty = {};	// all-purpose empty byte array



/**
 *  Constructs a Tk with specified data.
 */
Tk(String name, byte[] data) {
    this.name = name;			// save name
    this.hcode = Util.hash(name);	// precompute hashcode
    this.setData(data);			// init data field
}



/**
 *  Constructs a token by reading a Pickle.
 */

//  could be cast as "this(p.getString(), p.getByteArray())", but that
//  gets an erroneous error diagnostic from some JDK 1.1 compilers

Tk(Pickle p) {
    this.name = p.getString();
    this.hcode = Util.hash(name);
    this.setData(p.getByteArray());
}



/**
 *  Adds the encoding of this token to a Pickle.
 */
synchronized Pickle pickle(Pickle p) {
    p.put(name);
    p.put(data != null ? data : empty);
    return p;
}



/**
 *  Returns the data associated with this token.
 *  The returned array should not be modified.
 */
public synchronized byte[] getData() {
    if (data == null) {
	return empty;
    } else {
	return data;
    }
}


/**
 *  Returns the data of this token as a string.
 */
public String getString() {
    return new String(getData());
}



/**
 *  Sets the data associated with this token.
 */
synchronized void setData(byte[] data) {
    if (data == null || data.length == 0) {
	this.data = null;
    } else {
	this.data = (byte[]) data.clone();
    }
}



/**
 *  Returns the index of the primary server for this token,
 *  regardless of whether it is currently functioning.
 *
 *  @param nservers	the number of servers configured
 */
int primary(int nservers) {
    return hcode % nservers;
}



/**
 *  Returns the index of the current server for this token,
 *  which depends on the states of the available servers.
 *  If all servers are down, -1 is returned.
 *
 *  @param sstates	a server state array, as from a RosterMessage.
 */
int server(byte[] sstates) {
    int nservers = sstates.length;
    int i = hcode % nservers;
    int v = sstates[i];
    if (v == Server.READY || v == Server.BOOTING) {
	return i;
    }
    short[] s = sequence(nservers);
    for (i = 1; i < nservers; i++) {
	int j = s[i];
	v = sstates[j];
	if (v == Server.READY || v == Server.BOOTING) {
	    return j;
	}
    }
    return -1;
}



/**
 *  Returns an list of indexes defining the sequence of servers,
 *  beginning with the primary, for this token.
 *
 *  @param nservers	the number of servers configured
 */
short[] sequence(int nservers) {
    int h = hcode;
    short a[] = new short[nservers];
    for (int i = 0; i < nservers; i++) {
	a[i] = (short) i;
    }
    for (int i = 0; i < nservers - 1; i++) {
	Util.swap(a, i, i + h % (nservers - i));
	h = Util.rehash(h);
    }
    return a;
}



/**
 *  Returns a string representation of this message.
 */
public String toString() {
    return Util.classname(this) + "(" + name + "," + dimage() + ")";
}



/**
 *  Returns a printable image, for debugging, of the data value.
 */
public String dimage() {
    if (data == null || data.length == 0) {
	return "";
    } else {
	return Util.image(new String(data), 12);
    }
}



/**
 *  Returns the hash code value of the token.
 *  The hash code is dependent only on the token name.
 */
public int hashCode() {
    return hcode;
}



}  // class Tk
