//  Vault.java -- database of Tokens for a Server


package topovista.toksvc;

import java.util.*;



/**
 *  A Vault maintains a database of Tokens
 *  for a Server.
 */
class Vault {

    private Hashtable alltokens = new Hashtable();



/**
 *  Defines an upper bound for the number of tokens remembered.
 */
static final int CAPACITY = 20000;



/**
 *  Finds a token in the database.
 */
synchronized ServerToken get(String name) {
    return (ServerToken) alltokens.get(name);
}



/**
 *  Finds or creates a token in the database.
 */
synchronized ServerToken lookup(String name) {
    ServerToken t = (ServerToken) alltokens.get(name);	// look up token
    if (t == null) {					// if not found
	/*
	 *  Need to add a new token.  First make room, if necessary.
	 */
	while (alltokens.size() >= CAPACITY) {		// if already full,
	    t = ServerToken.popLRU();			// get oldest token
	    if (t == null) {
		break;
	    }
	    if (t.getState() == 0) {			// if unassigned,
		alltokens.remove(t.name);		// discard it
	    }
	}
	t = new ServerToken(name);			// make a new token
	alltokens.put(t.name, t);			// add to database
	t.addLRU();					// add to LRU list
    }
    return t;
}



/**
 *  Records the holdings of one client as received in a CatalogMessage.
 */
synchronized void register(CatalogMessage m, Session s) {
    for (int i = 0; i < m.tokens.length; i++) {
	ServerToken tknew = (ServerToken) m.tokens[i];
	tknew.assign(s, Token.EXCLUSIVE);
	String name = tknew.name;
	ServerToken tkold = (ServerToken) alltokens.get(name);
	if (tkold != null) {
	    tkold.delLRU();
	    tknew = new ServerToken(tkold, tknew);
	}
	alltokens.put(name, tknew);
	/* tknew.addLRU() not called, because token is in use */
    }
}



/**
 *  Incorporates tokens received from another server in a TransferMessage.
 */
synchronized void transferIn(TransferMessage m) {
    for (int i = 0; i < m.tokens.length; i++) {
	ServerToken tknew = m.tokens[i];
	String name = tknew.name;
	ServerToken tkold = (ServerToken) alltokens.get(name);
	if (tkold != null) {
	    tkold.delLRU();
	    tknew = new ServerToken(tkold, tknew);
	}
	alltokens.put(name, tknew);
	if (tknew.getState() == 0) {
	    tknew.addLRU();
	}
    }
}



/**
 *  Builds a TransferMessage containing tokens that should be
 *  served by the given server.
 */
synchronized TransferMessage transferTo(int dest, byte[] sstates) {
    Vector v = new Vector();
    for (Enumeration e = alltokens.elements(); e.hasMoreElements(); ) {
	ServerToken t = (ServerToken) e.nextElement();
	if (t.server(sstates) == dest) {
	    v.addElement(t);
	}
    }
    ServerToken[] tlist = new ServerToken[v.size()];
    v.copyInto(tlist);
    return new TransferMessage(tlist);
}



/**
 *  Discards tokens that are no longer the concern of this server.
 */
synchronized void forget(Certifier tg) {
    Vector v = new Vector();
    for (Enumeration e = alltokens.elements(); e.hasMoreElements(); ) {
	ServerToken t = (ServerToken) e.nextElement();
	if (tg.status(t) == Certifier.FOREIGN) {
	    v.addElement(t);		// no more reason to keep this token
	}
    }
    for (Enumeration e = v.elements(); e.hasMoreElements(); ) {
	ServerToken t = (ServerToken) e.nextElement();
	alltokens.remove(t.name);
	t.delLRU();
    }
}



}  // class Vault
