package topovista.tvserver;

import java.io.*;
import java.net.*;
import java.util.*;

import topovista.grid.*;
import topovista.util.*;



/**
 *  Implements a TopoVista tile server using the TVTP protocol.
 */
public class TVServer {



private static int REQLENGTH = 1000;

private static int MSGHDR = 14;
private static int MAXREC = 8192 - MSGHDR;


// configuration values
private static int port;	// UDP port number
private static int delay;	// milliseconds to sleep after each packet send
private static String edrop;	// non-null if to drop error messages
private static float dfract;	// fraction to drop intelligently
private static float rfract;	// fraction to drop randomly



// table mapping file names to containing directories
private static Hashtable dtable = new Hashtable();



/**
 *  Runs until halted, accepting and replying to TVTP requests.
 *  The sole argument is the number of the port from which to read requests.
 *
 *  <P> Data is read from "*.tvg" files located in and below the current
 *  directory.  File names are assumed to accurately reflect contents
 *  in accordance with TopoVista naming conventions.  Directories are
 *  scanned at startup; additional files cannot be added dynamically.
 */
public static void main(String[] args) throws Exception {

    String cmdname = "java TVServer";
    Options o = new Options(new String[] {
	"-e",		"drop all error records",
	"-d pct 0",	"intelligently drop a percentage of output messages",
	"-r pct 0",	"randomly drop a percentage of output messages",
	"-s msec 0",	"sleep msec milliseconds after each packet send",
	"portnum",	"port number",
    }, cmdname, args);

    try {
	edrop = o.getValue('e');
	dfract = Integer.parseInt(o.getValue('d')) / 100.0f;
	rfract = Integer.parseInt(o.getValue('r')) / 100.0f;
	delay = Integer.parseInt(o.getValue('s'));
	port = Integer.parseInt(args[o.getIndex()]);
    } catch (Exception e) {
	o.usage(System.err, cmdname);
	System.exit(1);
    }

    DatagramSocket udp = new DatagramSocket(port);

    System.err.println("Initializing...");
    scanDirectory(new File("."));	// initialize file lookup table
    System.err.println("Ready.");

    for (;;) {
	byte[] buf = new byte[REQLENGTH];
	DatagramPacket in = new DatagramPacket(buf, buf.length);
	udp.receive(in);
	String s = new String(in.getData(), 0, in.getLength());
	if (s.startsWith("GET ")) {
	    int id = Integer.parseInt(s.substring(4, 12), 16);
	    int[] vals = TileGrid.parseID(id);
	    String fname = TileGrid.makeName(
	        vals[0], vals[1], vals[2], vals[3]) + ".tvg";
	    Vector recs = load(fname);
	    exclude(recs);
	    byte[][] msgs = bundle(recs, id);
	    System.err.println(in.getSocketAddress() +
		"  " + msgs.length + "  " + s);
	    for (int i = 0; i < msgs.length; i++) {
		udp.send(new DatagramPacket(
		    msgs[i], msgs[i].length, in.getAddress(), in.getPort()));
	        if (delay > 0) {
		    Thread.sleep(delay);
		}
	    }
	}
    }
}



/**
 *  Scans a directory, recursively, adding all TopoVista files
 *  to the global Hashtable dtable.
 */
private static void scanDirectory(File dir) {
    int n = 0;
    String[] flist = dir.list();
    for (int i = 0; i < flist.length; i++) {
	if (flist[i].endsWith(".tvg")) {
	    dtable.put(flist[i], dir);
	    n++;
	} else {
	    File d = new File(dir, flist[i]);
	    if (d.isDirectory()) {
		scanDirectory(d);
	    }
	}
    }
    if (n > 0) {
	System.err.println("   " + dir + ":\t" + n + " files");
    }
}



/**
 *  Reads a TopoVista tile (version 4F) from the specified file,
 *  returning a vector of byte arrays, each array being one TV record.
 *  Records larger than MAXREC are immediately discarded, as are
 *  error records if the command option "-e" was specified.
 *  If the file is not found, an empty vector is returned.
 */
private static Vector load(String fname) throws Exception {
    DataInputStream ins;

    Vector v = new Vector();

    File dir = (File) dtable.get(fname);
    if (dir == null) {
	return v;
    }
    File f = new File(dir, fname);
    ins = new DataInputStream(new BufferedInputStream(new FileInputStream(f)));

    for (;;) {
	try {
	    byte[] hdr = new byte[8];
	    ins.readFully(hdr);
	    if (!(new String(hdr, 0, 4)).equals("TV4F")) {
		return v;
	    }
	    int n = ((hdr[5] & 0xFF) << 16) +
		((hdr[6] & 0xFF) << 8) + (hdr[7] & 0xFF);
	    byte[] rec = new byte[n];
	    System.arraycopy(hdr, 0, rec, 0, hdr.length);
	    ins.readFully(rec, hdr.length, n - hdr.length);
	    if (rec.length >= MAXREC) {
		continue;		/* too big to transmit via UDP */
	    }
	    if (edrop != null && (hdr[4] & 0x7F) >= 0x40) {
		continue;		/* dropping all error packets */
	    }
	    v.add(rec);			/* add record to list */
	} catch (EOFException e) {
	    return v;
	}
    }
}


/**
 *  Alters the list of TV records based on the "-d" and "-r"
 *  command parameters.
 */
private static void exclude(Vector recs) {

    int nd = howMany(recs.size(), dfract);
    int nr = howMany(recs.size(), rfract);

    // remove nd records from the end
    while (nd-- > 0 && recs.size() > 1) {
	recs.remove(recs.lastElement());
    }

    // make a list of chances for random deletion:
    // larger messages have more chances
    Vector v = new Vector();
    for (int i = 1; i < recs.size(); i++) {
	byte[] b = (byte []) recs.elementAt(i);
	int nchances = 1 + b.length / 1500;	// approx sz of ethernet packet
	for (int j = 0; j < nchances; j++) {
	    v.add(b);
	}
    }

    // remove nr records randomly
    Collections.shuffle(v);
    for (int i = 0; i < nr && i < v.size(); i++) {
	recs.remove(v.elementAt(i));
    }
}



/**
 *  Decides how many of n records to drop if each one is dropped
 *  with probability p.  Results vary pseudorandomly.
 */
private static int howMany(int n, double p) {
    int d = 0;
    while (n-- > 0) {
	if (Math.random() < p) {
	    d++;
	}
    }
    return d;
}



/**
 *  Produces an array of TVTP messages encapsulating the given list of
 *  TopoVista records, with addition of an EOF message.
 */
private static byte[][] bundle(Vector v, int tid) {
    int n = v.size();
    long bitmap = 1;			// init bitmap with just EOF msg
    byte[] hdr = { 'T', 'V', 'T', 'P', 2, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
    byte[][] msgs = new byte[n + 1][];

    // tile ID and version
    for (int i = 0; i < 4; i++) {
	hdr[8 + i] = (byte) (tid >> (8 * (3 - i)));
    }
    hdr[12] = 0;
    hdr[13] = 1;			// tile version number

    // data messages
    for (int i = 0; i < n; i++) {
	byte[] rec = (byte[]) v.elementAt(i);
	byte[] m = msgs[i] = new byte[hdr.length + rec.length];
	System.arraycopy(hdr, 0, m, 0, hdr.length);
	System.arraycopy(rec, 0, m, hdr.length, rec.length);
	int flag = rec[4] & 0x7F;
	int seq = (flag < 32) ? 1 : ((flag < 64) ? (flag - 30) : (flag - 42));
	m[6] = (byte) seq;
	bitmap |= 1L << (63 - seq);
    }

    // EOF message
    byte[] m = msgs[n] = new byte[hdr.length + 8];
    System.arraycopy(hdr, 0, m, 0, hdr.length);
    m[6] = 63;				// EOF seqnum
    for (int i = 0; i < 8; i++) {
	m[hdr.length + i] = (byte) (bitmap >> (8 * (7 - i)));
    }

    // reorder the array to interleave data records with error records
    Arrays.sort(msgs, new Comparator() {
	public int compare(Object o1, Object o2) {
	    int seq1 = ((byte[]) o1) [6];
	    int seq2 = ((byte[]) o2) [6];
	    int rank1 = (seq1 < 22) ? (4 * seq1 + 37) : (2 * seq1);
	    int rank2 = (seq2 < 22) ? (4 * seq2 + 37) : (2 * seq2);
	    return rank1 - rank2;
	}
    });

    // next and prev seqnums
    for (int i = 1; i <= n; i++) {
	msgs[i][5] = msgs[i-1][6];	// prev
    }
    for (int i = 0; i < n; i++) {
	msgs[i][7] = msgs[i+1][6];	// next
    }
    return msgs;
}



} // class TVServer
