package topovista.tvsucker;

/**
 *  A Tile instance records information about one particular tile.
 */

import java.util.*;

import topovista.grid.*;



class Tile {

    public int zone;		// UTM zone
    public int xxx, yyy;	// UTM-based coordinates
    public int k;		// tile order
    public String extn;		// filename extension
    public int[] recl;		// length of each record of the tile

    public Timestamp tReq;	// request time
    public Timestamp tResp;	// first response time
    public Timestamp[] tGot;	// receive time for each of N records
    public Timestamp tDone;	// time when input complete


private final static int MAXINDEX = 127;	// maximum record index



/**
 *  Constructs a new Tile, given its concatenated name.
 */
Tile(String name) throws Exception {
    String extn;
    int i = name.indexOf('.', name.lastIndexOf('/') + 1);
    if (i >= 0) {
	extn = name.substring(i);
	name = name.substring(0, i);
    } else {
	extn = "";
    }
    int[] vals = TileGrid.parseName(name);
    init(vals[0], vals[1], vals[2], vals[3], extn);
}



/**
 *  Constructs a new Tile, given its broken-down name.
 */
Tile(int zone, int xxx, int yyy, int k, String extn) {
    init(zone, xxx, yyy, k, extn);
}



/**
 *  Initializes a new Tile.
 */
private void init(int zone, int xxx, int yyy, int k, String extn) {
    this.zone = zone;
    this.xxx = xxx;
    this.yyy = yyy;
    this.k = k;
    this.extn = extn;
    this.recl = new int[MAXINDEX + 1];
    this.tGot = new Timestamp[MAXINDEX + 1];
}



/**
 *  Logs the initial request for a tile.
 */
void req() {
    tReq = new Timestamp();
}

/**
 *  Logs the completion of the request sending.
 */
void resp() {
    tResp = new Timestamp();
}

/**
 *  Logs receipt of record r and stores its length.
 */
void got(int r, int len) {
    r &= MAXINDEX;	// clear gzip bit
    tGot[r] = new Timestamp();
    recl[r] = len;
}

/**
 *  Logs completion of tile receipt, and notifies anyone waiting.
 */
void done() {
    tDone = new Timestamp();
    synchronized (this) {
	this.notifyAll();
    }
}



/**
 *  Returns the name of the Tile by concatenating xxx, k, and yyy.
 */
public String getName() {
    return TileGrid.makeName(zone, xxx, yyy, k) + extn;
}



/**
 *  Returns a string representation of this Tile.
 */
public String toString() {
    return "Tile(" + getName() + ")";
}



/**
 *  Returns a one-line report for this Tile.
 *  Assumes that tDone is valid.
 */
public String report() {

    int t = 0;
    for (int i = 0; i <= MAXINDEX; i++) {
	t += recl[i];
    }

    StringBuffer b = new StringBuffer();
    b.append(rj(t, 6));
    b.append(' ');
    b.append(getName());
    if (tReq != null) {
	b.append(rj(tReq.value, 5));
	b.append(rj(tDone.value, 5));
	b.append(rj(tDone.value - tReq.value, 4));
    } else {
	b.append("    -");
	b.append(rj(tDone.value, 5));
	if (tResp != null) {
	    b.append(rj(tDone.value - tResp.value, 4));
	} else {
	    b.append("   -");
	}
    }
    b.append("  ");

    for (int i = 0; i <= MAXINDEX; i++) {
	if (i == 0x40) {
	    b.append(' ');
	}
	if (recl[i] > 0) {
	    b.append(rchars.charAt(i));
	}
    }

    return b.toString();
}

private static String rchars =		// record indicator characters
    "????????????????+???????????????abcdefghijklmnopqrstuvwxyz??????" +
    "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ??";



/**
 *  Right-justifies an integer in a field of width w (w <= 40);
 */
private static String rj(long n, int w) {
    String s = Long.toString(n);
    int len = s.length();
    if (len < w) {
	return blanks.substring(0, w - len) + s;
    } else {
	return s;
    }
}

private static String blanks = "                                        ";



} // class Tile
