package topovista.util;

import java.io.*;



/**
 *  A TypedOutputStream writes string, integer, floating, and null values
 *  in compressed form without fear of overflowing fixed-width fields.
 *  Full precision is always retained using varying-length encodings.
 *  The format is understood by the related TypedInputStream class.
 *
 *  <P> A big-endian (MSB first) representation is used.
 *  The first byte of each encoded value identifies its type and format,
 *  and may carry part or all of the value:
 *  <PRE>
 *  hex   binary     interpretation
 *  __   _________   _______________________________________________________
 *
 *  0+   0 snnnnnn   small signed integer
 *
 *  8+   10 snnnnn   small signed integer difference
 *
 *  Cn   1100 0snn   11-bit signed integer, including next byte
 *  Cn   1100 1snn   11-bit signed integer difference, including next byte
 *
 *  Dn   11010 bbb   signed integer follows in b+1 bytes
 *  Dn   11011 bbb   signed integer difference follows in b+1 bytes
 *
 *  En   1110 nnnn   n-byte string follows (n < 15)
 *  EF   1110 1111   encoded string follow: [-tspcount,] length, characters
 *
 *  F0   11110 000   integer follows, but float it
 *  Fn   11110 bbb   zero-trimmed floating value follows in b+1 bytes
 *                   (if b+1<=4, value is IEEE float, otherwise IEEE double)
 *
 *  F8   111110 00   0.0
 *  F9   111110 01   1.0
 *  FA   111110 10   -1.0
 *  FB   111110 11   reserved
 *
 *  FC   111111 00   reserved
 *  FD   111111 01   reserved
 *  FE   111111 10   the null value
 *  FF   111111 11   reserved for future use as an escape code
 *  </PRE>
 *
 *  <P> A single byte can represent
 *  the null value,
 *  the empty string,
 *  an integer between -64 and +64,
 *  an integer that differs from the previous one by less than 32,
 *  or one of the three special floating values -1.0, 0.0, and 1.0.
 *
 *  <P> Two bytes are sufficient to represent
 *  a single-character string,
 *  an integer between -1024 and 1024,
 *  an integer within 1024 of the previous one,
 *  or a floated small integer.
 *
 *  <P>
 *  In typical use, most encoded values are more compact than their
 *  native forms.  The worst-case encoding overhead is:
 *  <UL>
 *  <LI> 1 byte for short, int, or long integers
 *  <LI> 1 byte for float or double values
 *  <LI> 1 byte for strings up to 14 characters
 *  <LI> 2 bytes for strings up to 63 characters
 *  <LI> 3 bytes for strings up to 1023 characters
 *  </UL>
 *
 *  @author Gregg Townsend, University of Arizona
 *  @version August, 2001
 */



public class TypedOutputStream extends FilterOutputStream {

    private long prevInt = 0;		// previous integral value output



/**
 *  Constructs a TypedOutputStream from an underlying OutputStream
 */
public TypedOutputStream(OutputStream out) {
    super(makeBuffered(out));
}



/**
 *  Makes an output stream buffered, if it is not already.
 */
private static OutputStream makeBuffered(OutputStream out) {
    if (out instanceof BufferedOutputStream) {
	return out;
    } else {
	return new BufferedOutputStream(out);
    }
}



/**
 *  Writes an integral value.
 */
public void write(long v) throws IOException {

    long diff = v - prevInt;		// difference from previous value
    prevInt = v;			// save curr value as new "previous"

    if (-0x40 <= v && v < 0x40) {
	super.write((int)v & 0x7F);
    } else if (-0x20 <= diff && diff < 0x20) {
	super.write(0x80 | (int)(diff & 0x3F));
    } else if (-0x400 <= v && v < 0x400) {
	super.write(0xC0 + ((((int)(v)) >> 8) & 0x07));
	super.write((int) v);
    } else if (-0x400 <= diff && diff < 0x400) {
	super.write(0xC8 + ((((int)(diff)) >> 8) & 0x07));
	super.write((int) diff);
    } else {
	int h;				// header byte
	if (Math.abs(v) - 1 <= Math.abs(diff) - 1) {	// -1 handles 0x80...00
	    h = 0xD0;			// encode value itself
	} else {
	    h = 0xD8;			// encode difference from previous
	    v = diff;
	}
	int bbb = 1;
	for (long x = v >> 15; x != 0 && x != -1; x >>= 8) {
	    bbb++;
	}
	super.write(h + bbb);
	while (bbb >= 0) {
	    super.write((byte) (v >> (8 * bbb--)));
	}
    }
}



/**
 *  Overrides FilterOutputStream.write(int) to write an integer
 *  in packed form.
 */
public void write(int v) throws IOException {
    this.write((long) v);
}



/**
 *  Writes a float value in full precision.
 */
public void write(float f) throws IOException {

    int bits = Float.floatToRawIntBits(f);
    if (bits == 0) {			// true for +0.0 but not -0.0
	super.write(0xF8);
    } else if (f == 1.0) {		// special case 1.0
	super.write(0xF9);
    } else if (f == -1.0) {		// special case -1.0
	super.write(0xFA);
    } else {
	int n = (int) f;
	// check if plausibly encoded as integer  (not a perfect test)
	// (must have no fraction, be reasonably small, and not -0)
	if (n == f && n < 1024 && n >= -1024 && n != 0) {
	    super.write(0xF0);
	    this.write(n);			// write integer value
	} else {
	    wtf((long)bits << 32, 2, 4);	// write trimmed float value
	}
    }
}



/**
 *  Writes a double value in full precision.
 */
public void write(double g) throws IOException {

    float f = (float) g;
    if (f == g || f != f) {		// if float equals double value, or NaN,
	this.write(f);			// can represent exactly as a float
    } else {
	wtf(Double.doubleToRawLongBits(g), 5, 8);
    }
}



/**
 *  Writes a floating value, trimming trailing zero bytes.
 */
private void wtf(long bits, int minbytes, int maxbytes) throws IOException {

    long mask = 0xFFL << (8 * (8 - maxbytes));
    while (maxbytes > minbytes && (bits & mask) == 0) {
	maxbytes--;
	mask <<= 8;
    }

    super.write(0xF0 - 1 + maxbytes);
    int n = 56;
    for (int i = 0; i < maxbytes; i++) {
	super.write((byte) (bits >> n));
	n -= 8;
    }
}



/**
 *  Writes a string.
 */
public void write(String s) throws IOException {

    if (s == null) {
	super.write(0xFE);
	return;
    }

    int len = s.length();
    if (len <= 2) {
	super.write(0xE0 + len);
	if (len > 0) {
	    super.write(s.charAt(0));
	}
	if (len == 2) {
	    super.write(s.charAt(1));
	}
	return;
    }

    byte[] buf = s.getBytes();

    int t = len;
    while (t > 0 && buf[t-1] == ' ') {	// count trailing spaces
	t--;
    }

    if (len < 15 && (len - t) < 3) {
	t = len;
	super.write(0xE0 + len);
    } else {
	super.write(0xEF);
	if (t < len) {
	    this.write(t - len);
	}
	this.write(t);
    }

    for (int i = 0; i < t; i++) {
	super.write(buf[i]);
    }
}



} // class TypedOutputStream
